(function ($) {
    "use strict";
    /**
     * The static singleton Schedule controller. Schedule is the public interface of the entire Schedule Manager.
     * @class
     * @name Schedule
     */
    var Schedule = function () {
        var self = this;

        /**
         * Initializes a new ScheduleParm object. This object is used for all subsequent calls to the Schedule class
         * @param dsTasks Tasks datasource
         * @param dsPreds Predecessor datasource
         * @param dsTaskUsers TaskUser datasource
         * @param project project data item
         * @param blockoutDays list of blockoutDays from the server
         * @param prefs preference object (must include ActualDateUpdatesPlan,AutoScheduleProjects,AutoScheduleTaskEndDate)
         * @param assignments list of project team members
         * @param options object that holds any options that need to be set for the ScheduleManager:
         *          {retainer: true}
         * @returns {ScheduleParm}
         */
        function initScheduleParms(dsTasks, dsPreds, dsTaskUsers, project, blockoutDays, prefs, assignments, options) {
            var parms = new ScheduleParm(dsTasks, dsPreds, dsTaskUsers, project, blockoutDays, prefs, assignments, options);

            parms.scheduleMgr.loadPredecessors();
            parms.scheduleMgr.initBlockoutDays();

            return parms;
        }

        /**
         * loads all of the schedule bucket information
         * @param parms
         */
        function initBuckets(parms) {
            if (!parms) return;
            parms.scheduleMgr.initBuckets();
        }

        /**
         * Called when a different template is selected, and sets that template's assignments and predecessors
         * @param parms
         * @param dsTaskUsers TaskUser datasource that is loaded with the taskusers for the selected template
         * @param dsPreds Predecessor datasource that is loaded with the predecessors for the selected template
         */
        function setSelectedTemplate(parms, dsTaskUsers, dsPreds) {
            if (!parms) return;
            parms.setSelectedTemplate(dsTaskUsers, dsPreds);
        }

        /**
         * Format the predecessor key and label values for the SelectList control
         * @param parms
         */
        function formatPredValues(parms) {
            if (!parms) return;
            parms.scheduleMgr.formatPredValues();
        }

        /**
         * Primary fieldChange function that is called from the datasource's fieldchange handler
         * @param parms
         * @param item data item
         * @param field field being changed
         * @param dataSetName optional. specifies whether it's the Task or TaskUser dataset (values: window.Schedule.TASKS (default), window.Schedule.TASK_USERS)
         * @returns {ScheduleResponse}
         */
        function fieldChange(parms, item, field, dataSetName) {

            var response = new ScheduleResponse(item, field);

            if (!parms) return response;
            if (field === 'tagged') return response;

            var validation = new ScheduleValidation(parms, item, response, false, dataSetName);

            if (validation[field]) {
                validation[field]();
            } else {
                if (dataSetName === window.Schedule.TASK_USERS && field.indexOf('Offset') >= 0) {
                    validation.Offset(field);
                }
            }

            //if errors were added during validation, no need to schedule
            if (parms.scheduleMgr.hasGeneralErrors()) {
                response.generalErrors = parms.scheduleMgr.getGeneralErrors();
            } else {
                //validate the schedule
                if (!validation.validateSchedule()) {
                    response.generalErrors = validation.getGeneralErrors();
                }
            }

            response.notifications = response.notifications || [];
            if(parms.scheduleMgr.hasGeneralNotifications() || validation.hasGeneralNotifications()){
                response.notifications = parms.scheduleMgr.getGeneralNotifications().concat(validation.getGeneralNotifications());
            }

            validation = null;

            return response;
        }

        /**
         * This is called to validate the entire row
         * @param parms
         * @param item data item
         * @param dataSetName optional. specifies whether it's the Task or TaskUser dataset (values: Schedule.TASKS (default), Schedule.TASK_USERS)
         * @returns {ScheduleResponse}
         */
        function validateRow(parms, item, dataSetName) {

            if (!parms) return;

            var response = new ScheduleResponse(item);
            var validation = new ScheduleValidation(parms, item, response, true, dataSetName);

            dataSetName = validation.dataSetName;

            //list of field names that need to be validated during a row level validation
            var rowValidationFields;

            //variables to handle either validating children or adding errors to the parents
            var childDataSets = [];
            var parentDS;
            var parentErrorMsg;
            var parentItemProperty;

            switch (dataSetName) {
                case window.Schedule.TASKS:
                    //when tasks are validated, the taskUsers also need to be validated
                    rowValidationFields = ['TaskName', 'TaskID', 'ConstraintDayOfTheWeek', 'ConstraintDate', 'PlanDuration', 'PlanStart', 'PlanComplete', 'CalendarEvent', 'BaseStart', 'BaseComplete'];
                    childDataSets.push({dataSetName: window.Schedule.TASK_USERS, data: item[window.Schedule.TASK_USERS], ds: parms.dsTaskUsers, rowErrorMsg: System.localize('|26625|An assignment has an error|')});
                    break;
                case window.Schedule.TASK_USERS:
                    //when taskUsers are validated, the parent DS will need an error added to it if any children contain errors
                    parentDS = parms.dsTasks; //used to add an error to the parent when the child has an error
                    parentErrorMsg = System.localize('|26625|An assignment has an error|');
                    parentItemProperty = 'task'; //in the response, this property will be the parent task which is used to add the error
                //no break!!!!!
                default:
                    rowValidationFields = []; //no specific fields, just call validateRowCleanup() below
                    break;
            }

            var field;
            for (var i = 0; i < rowValidationFields.length; i++) {
                field = rowValidationFields[i];

                if (validation[field]) {
                    validation[field]();
                }
            }

            //call validateRow for any child datasets
            var childValidationItem;
            var childData;
            var childResp;
            var childDS;
            for (var j = 0; j < childDataSets.length; j++) {
                childValidationItem = childDataSets[j];
                childData = childValidationItem.data; //this is the sub data (i.e. only the child items that apply to the parent item)
                if (childData) {
                    childDS = childValidationItem.ds; //this is the actual datasource object from page (used to add the errors)
                    for (var k = 0; k < childData.length; k++) {
                        childResp = validateRow(parms, childData[k], childValidationItem.dataSetName);
                        if (childResp.hasErrors()) {
                            childResp.addDSErrors(childDS);
                            childDS.fire();
                            response.addError(childValidationItem.rowErrorMsg); //generic error message added to the parent row
                        }
                    }
                }
            }

            validation.validateRowCleanup();
            if (!validation.validateSchedule()) {
                response.generalErrors = validation.getGeneralErrors();
            }

            if (parentDS && parentErrorMsg && response.hasErrors() && !parms.options.retainer) {
                parentDS.addError(parentErrorMsg, '', response[parentItemProperty]);
            }

            response.notifications = validation.getGeneralNotifications();

            validation = null;

            return response;
        }

        /**
         * Sets the task defaults on the item, and links it to the previous task (note: it does not add it to the datasource)
         * @param parms
         * @param prev Optional Task that is previous to the newly added one. If specified, the new task's SummaryTaskKey is set to the same as the prev.SummaryTaskKey.
         *      Also the prev task will be set as the predecessor of the new tak
         * @param item The new task item that was created when it was added to the datasource
         */
        function addTask(parms, prev, item) {
            if (!parms) return;
            parms.scheduleMgr.addTask(prev, item);
        }

        /**
         * Links a task as a predecessor
         * @param parms
         * @param task Task that is about to get linked
         * @param linkToThisTask Task that will be set as a predecessor
         * @param lag Lag days
         * @param delaySchedule Pass in true if you are linking multiple tasks at one time so that the scheduling logic doesn't run after each link
         * @param inWeeks 1 or 0 to indicate whether the predecessor is scheduled in weeks
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function linkToTask(parms, task, linkToThisTask, lag, delaySchedule, inWeeks) {
            if (!parms) return;
            parms.scheduleMgr.linkToTask(task, linkToThisTask, lag, delaySchedule, inWeeks);
            return packageReturnObject(parms)
        }

        /**
         * Adds cross-project predecessor information to the list for use in calculations
         * NOTE: Does not add a cross-project predecessor to a task!
         * @param parms
         * @param predKey
         * @param predPlanComplete
         * @param predPlanDuration
         */
        function addCrossProjectPredInformation(parms, predKey, predPlanComplete, predPlanDuration){
            if(!parms) return;
            parms.scheduleMgr.addCrossProjectPredInformation(predKey, predPlanComplete, predPlanDuration);
        }

        /**
         * Deletes the tasks (and their children). If the tasks are used on any transactions, it will add an error and return.
         * This will mark the rowstate of the tasks 'D', but not remove them from the source of the datasource
         * @param parms
         * @param tasks Array of tasks to delete
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function deleteTasks(parms, tasks) {
            if (!parms) return;
            parms.scheduleMgr.deleteTasks(tasks);
            return packageReturnObject(parms);
        }

        /**
         * Sets the tasks to 100% and marks them dirty
         * @param parms
         * @param tasks Array of tasks to mark complete
         */
        function completeTasks(parms, tasks) {

            var item;
            for (var i = 0; i < tasks.length; i++) {
                item = tasks[i];
                if (item.TaskType !== Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                    item.PercComp = 100;
                    parms.scheduleMgr.markTaskDirty(item);
                    fieldChange(parms, item, 'PercComp');
                }
            }
        }

        /**
         * Rolls through the entire schedule and automatically sets the TaskIDs of the tasks
         * @param parms
         */
        function autoNumberTasks(parms) {
            if (!parms) return;
            parms.scheduleMgr.autoNumberTasks();
        }

        /**
         * Sets the TaskID of an individual task
         * @param parms
         * @param task Task data item
         * @param parentNumber The TaskID of the parent
         * @param currentNumber The DisplayOrder of the item at the current level
         */
        function setTaskNumbers(parms, task, parentNumber, currentNumber) {
            if (!parms) return;
            parms.scheduleMgr.setTaskNumbers(task, parentNumber, currentNumber);
        }

        /**
         * Determines whether the task can be moved to the new location. isAllowed will be true if it's allowed
         * @param parms
         * @param taskBeingMoved The task that is being moved
         * @param referenceTask The task that the user is either dropping onto, or in front of.
         * @param dropAsChild If true, the taskBeingMoved is about to be a child of referenceTask
         * @returns {{generalErrors, generalNotifications, isAllowed}|*}
         */
        function isMoveTaskAllowed(parms, taskBeingMoved, referenceTask, dropAsChild) {
            if (!parms) return {isAllowed: false};

            var isAllowed = parms.scheduleMgr.isMoveTaskAllowed(taskBeingMoved, referenceTask, dropAsChild);
            var rtn = packageReturnObject(parms);
            rtn.isAllowed = isAllowed;
            return rtn;
        }

        /**
         * Checks if the referenceTask has assignments, so that we can ask if it's OK to remove the assignments.
         * Returns false if the referenceTask has children
         * @param parms
         * @param referenceTask Task that is about to become the parent
         * @returns Boolean
         */
        function isDropAsChildAllowed(parms, referenceTask) {
            if (!parms) return false;
            return parms.scheduleMgr.isDropAsChildAllowed(referenceTask);
        }

        /**
         * Removes all assigned users from a task
         * @param parms
         * @param task Task data item
         *
         */
        function removeTaskAssignedUsers(parms, task) {
            if (!parms) return false;
            parms.scheduleMgr.removeTaskAssignedUsers(task);
        }

        /**
         * Performs cleanup logic after a task has been moved
         * @param parms
         * @param taskBeingMoved Task that was moved
         * @param oldParentTask The task's old parent task
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function afterMoveTask(parms, taskBeingMoved, oldParentTask) {
            if (!parms) return {success: false};
            parms.scheduleMgr.afterMoveTask(taskBeingMoved, oldParentTask);
            return packageReturnObject(parms);
        }

        /**
         * Copies a task
         * @param parms
         * @param taskBeingCopied Task being copied
         * @param dropTask Task that the copied task will be placed below
         * @param dropAsChild Boolean. If true, make the new task a child of dropTask, otherwise place it below dropTask on the same level
         */
        function copyTask(parms, taskBeingCopied, dropTask, dropAsChild) {
            if (!parms) return;
            parms.scheduleMgr.copyTask(taskBeingCopied, dropTask, dropAsChild);
        }

        /**
         * Returns the parentTask of the task. Returns null if the task is at the top level
         * @param parms
         * @param task
         * @returns {*} parent task
         */
        function getParentTask(parms, task) {
            if (!parms) return;
            return parms.scheduleMgr.getParentTask(task);
        }

        /**
         * Initializes the assignment buckets for use by the Assignment panel
         * @param parms
         * @param task
         */
        function initAssignments(parms, task) {
            if (!parms) return;

            //might be more things to do here, but for now, just reacalc buckets, if necessary
            parms.scheduleMgr.validateTaskBuckets(task);
        }

        /**
         * Causes the schedule to recalculate the dates and then validates the structure (DisplayOrder, ProjectOrder, TaskLevel, etc).
         * It will also recalculate the schedule and buckets with new blockout Days, if the blockoutDays parm is pased in
         * @param parms
         * @param blockoutDays List of blockout days to use for scheduling (usually passed in after the blockoutDays have been refreshed
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function refreshSchedule(parms, blockoutDays) {

            if (!parms) return;

            if (blockoutDays) {
                parms.refreshBlockoutDays(blockoutDays);
            }
            return scheduleMgrRefresh(parms, true);
        }

        /**
         * Calls a schedule manager refresh. Similar to refreshSchedule above, except it does not update blockout days, and validating the structure is optional
         * @param parms
         * @param validateStructure Optional Boolean. If true, the structure will be validated
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function validateSchedule(parms, validateStructure) {
            if (!parms) return;
            return scheduleMgrRefresh(parms, false, validateStructure);
        }

        /**
         * Causes the schedule to recalculate the dates and optionally validate the structure (DisplayOrder, ProjectOrder, TaskLevel, etc).
         * @param parms
         * @param invalidateFirst Boolean. If true, it will set the schedule to "invalid" first which will cause it to reschedule every task. If false, it will only reschedule if necessary.
         * @param validateStructure Optional Boolean.  If true, the structure will be validated
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function scheduleMgrRefresh(parms, invalidateFirst, validateStructure) {

            if (invalidateFirst) parms.scheduleMgr.invalidateSchedule();

            parms.scheduleMgr.validateSchedule();

            if (validateStructure) parms.scheduleMgr.validateStructure();

            return packageReturnObject(parms);
        }

        /**
         * Sets multiple tasks as predecessors
         * @param parms
         * @param tasks List of tasks to link as predecessors
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function linkMultipleTasks(parms, tasks) {
            if (!parms) return;
            parms.scheduleMgr.linkTasks(tasks, true);
            parms.scheduleMgr.validateSchedule();
            return packageReturnObject(parms);
        }

        /**
         * Removes multiple tasks as predecessors
         * @param parms
         * @param tasks List of tasks to remove the link as predecessors
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function unlinkTasks(parms, tasks) {
            if (!parms) return;
            //unlinkTasks returns a boolean if it was able to unlink any tasks or not. We're ignoring that and instead looking for errors inside the parms object by calling packageReturnObject()
            parms.scheduleMgr.unlinkTasks(tasks);
            return packageReturnObject(parms);
        }

        /**
         * This will rollup PercComp, ActStart, ActComplete from tTaskUser to tTask
         * It will also rollup tTaskUser.Hours to tTask.AllocatedHours
         * @param parms
         * @param taskUser Assignment row
         */
        function rollupActuals(parms, taskUser) {
            if (!parms) return;
            fieldChange(parms, taskUser, 'rollupActuals', window.Schedule.TASK_USERS);
        }

        /**
         * Initializes an assignment data row. A new item object must be passed in, this method does not create a new item.
         * @param parms
         * @param item The data item from the new assignment
         */
        function taskUserDefaults(parms, item) {
            if (!parms) return;
            parms.scheduleMgr.taskUserDefaults(item);
        }

        /**
         * This adds the removed assignment to the delete items array for use by getModifiedItems. It also calls rollupAssignedUsersToTask.
         * This does not mark the item as deleted in the datasource. This should be called in the deleteBatchMode handler for the datasource.
         * @param parms
         * @param item Assignment that is about to be removed
         */
        function removeTaskUser(parms, item) {
            if (!parms) return;
            parms.scheduleMgr.removeTaskUser(item);
        }

        /**
         * Checks to see if any assignments that are about to be reassigned are 100%
         * @param parms
         * @param opt
         * @param selectedTasks
         * @returns {Boolean|*} Returns false if any assignments are 100%
         */
        function reassignCheck(parms, opt, selectedTasks) {
            if (!parms) return;
            return parms.scheduleMgr.reassignCheck(opt, selectedTasks);
        }

        /**
         * Reassigns users on tasks
         * @param parms
         * @param opt Reassign options
         * @param selectedTasks
         * @param sendNotifications
         */
        function reassign(parms, opt, selectedTasks, sendNotifications) {
            if (!parms) return;
            parms.scheduleMgr.reassign(opt, selectedTasks, sendNotifications);
            return packageReturnObject(parms);
        }

        /**
         * Calculates the bucket columns and sets them on the tuhTaskParms object that is on the task
         * @param parms
         * @param task The task to calculate the buckets for
         * @param fromLoad Boolean. If true, the logic that recalculates the bucket hours is skipped because it's assumed that nothing has changed
         */
        function calcColumns(parms, task, fromLoad) {
            if (!parms) return;
            parms.scheduleMgr.calcColumns(task, fromLoad);
        }

        /**
         * This adds a template task to the schedule at the spot where it was dropped
         * @param parms
         * @param templateTask Task being dragged from the template
         * @param scheduleTask Task where the templateTask is being dropped
         * @param dropAsChild Boolean. If true, the newly added task will be a child of scheduleTask
         * @param dropAfterScheduleTask If true, it will drop the template task after the scheduleTask rather than before. This parm is ignored if dropAsChild is true.
         * @returns {{generalErrors, generalNotifications}|*}
         */
        function dropTemplateTask(parms, templateTask, scheduleTask, dropAsChild, dropAfterScheduleTask) {
            if (!parms) return;
            return parms.scheduleMgr.dropTemplateTask(templateTask, scheduleTask, dropAsChild, dropAfterScheduleTask, parms.keyMap);
        }

        /**
         * Packages up the return object. This function is used by many of the methods.
         * @param parms
         * @returns {{generalErrors: *, generalNotifications: *}}
         */
        function packageReturnObject(parms) {
            return {generalErrors: parms.scheduleMgr.getGeneralErrors(), generalNotifications: parms.scheduleMgr.getGeneralNotifications()};
        }

        /**
         * Returns all of the locally stored modifiedItems. Used because of unique scheduling save logic.
         * @param parms
         */
        function getModifiedItems(parms) {
            if (!parms) return;
            return parms.scheduleMgr.getModifiedItems();
        }

        /**
         * Called after save to clear ouut and commit all of the modifiedItems including new predecessor keys
         * @param parms
         * @param taskKeyMap The new task key map that was returned from the server after the save
         */
        function commitModifiedItems(parms, taskKeyMap) {
            if (!parms) return;
            parms.scheduleMgr.commitModifiedItems(taskKeyMap);
        }

        /**
         * Takes an array of selected tasks and returns a hierarchical structure with the parents
         * @param parms
         * @param selectedTasks Array of selected tasks
         * @param tasksOnly Boolean
         * @param taskFilter object {entity: UserKey|ServiceKey, key: key}
         * @param taggedTaskUserKeys Array of TaskKeys that were currently selected by the user. We will try to mark them as tagged again if they're still in the list
         * @returns {Array} Hierarchical structure of tasks
         */
        function getSelectedTaskAssignmentsWithParents(parms, selectedTasks, tasksOnly, taskFilter, taggedTaskUserKeys) {
            if (!parms) return;
            return parms.scheduleMgr.getSelectedTaskAssignmentsWithParents(selectedTasks, tasksOnly, taskFilter, taggedTaskUserKeys);
        }

        /**
         * Returns true is the ScheduleTask field should be enabled on the UI
         * @param parms
         * @param task
         * @returns {boolean}
         */
        function scheduleTaskEnabled(parms, task) {

            if (!parms) return;

            if (task.taskUsers && task.taskUsers.length > 0) {
                var tu = task.taskUsers.filter(function (item) {
                    return !item.rowstate || item.rowstate.indexOf('D') <= 0;
                });

                if (tu.length > 0) return false;
            }
            if (parms.scheduleMgr.isPredecessor(task)) return false;
            if (parms.scheduleMgr.hasPredecessors(task)) return false;

            return true;
        }

        /**
         * Returns the predecessors of a task
         * @param parms
         * @param task
         * @returns {Array} List of predecessors
         */
        function getPreds(parms, task){
            if (!parms) return;
            return parms.scheduleMgr.getPreds(task);
        }

        function refreshPreds(parms){
            if (!parms) return;
            parms.scheduleMgr.loadPredecessors();
        }

        self.initScheduleParms = initScheduleParms;
        self.fieldChange = fieldChange;
        self.validateRow = validateRow;
        self.addTask = addTask;
        self.linkToTask = linkToTask;
        self.addCrossProjectPredInformation = addCrossProjectPredInformation;
        self.deleteTasks = deleteTasks;
        self.autoNumberTasks = autoNumberTasks;
        self.setTaskNumbers = setTaskNumbers;
        self.isMoveTaskAllowed = isMoveTaskAllowed;
        self.isDropAsChildAllowed = isDropAsChildAllowed;
        self.removeTaskAssignedUsers = removeTaskAssignedUsers;
        self.afterMoveTask = afterMoveTask;
        self.copyTask = copyTask;
        self.getParentTask = getParentTask;
        self.initBuckets = initBuckets;
        self.setSelectedTemplate = setSelectedTemplate;
        self.formatPredValues = formatPredValues;
        self.initAssignments = initAssignments;
        self.refreshSchedule = refreshSchedule;
        self.validateSchedule = validateSchedule;
        self.linkMultipleTasks = linkMultipleTasks;
        self.unlinkTasks = unlinkTasks;
        self.commitModifiedItems = commitModifiedItems;
        self.getModifiedItems = getModifiedItems;
        self.completeTasks = completeTasks;
        self.rollupActuals = rollupActuals;
        self.taskUserDefaults = taskUserDefaults;
        self.removeTaskUser = removeTaskUser;
        self.reassign = reassign;
        self.reassignCheck = reassignCheck;
        self.calcColumns = calcColumns;
        self.dropTemplateTask = dropTemplateTask;
        self.getSelectedTaskAssignmentsWithParents = getSelectedTaskAssignmentsWithParents;
        self.scheduleTaskEnabled = scheduleTaskEnabled;
        self.getPreds = getPreds;
        self.refreshPreds = refreshPreds;

        //dataSetName constants
        self.TASKS = 'tasks';
        self.TASK_USERS = 'taskUsers';

        self.RETAINER_DAILY_HOURS = 8;

        return self;
    };
    //shortcut
    /** @type Schedule */
    window.Schedule = new Schedule();

    //endregion

    //region Validation

    /**
     * This class is the primary validation logic for the schedule. It primarily consists of field change functions for various fields that were changed.
     * @param parms
     * @param item The data row that contains the data to be validated
     * @param response The response object that is returned with any errors that may occur
     * @param validateRow Boolean. Pass true if the entire row is being validated to keep the scheduling logic from running
     * @param dataSetName Enum on window.Schedule. Either Tasks or TaskUsers. Tells the validation class which set of functions to use for validation.
     */
    var ScheduleValidation = function (parms, item, response, validateRow, dataSetName) {

        var ci = item; //using "ci" because that's the variable that was used in the validation code being copied from Flex
        var scheduleMgr = parms.scheduleMgr;
        var dsPreds = parms.dsPreds;
        var dsTaskUsers = parms.dsTaskUsers;
        var project = parms.project;
        var options = parms.options;
        var scheduleDirection = u.obn(project.ScheduleDirection);

        this.dataSetName = dataSetName = dataSetName || window.Schedule.TASKS;

        switch (dataSetName) {
            /**
             * TASK field change functions
             */
            case window.Schedule.TASKS:
                this.TaskID = function () {

                    if (ci.TaskID) {
                        //check for uniqueness
                        var lookupTask = parms.dsTasks.findByKey(ci.TaskID, 'TaskID');
                        if (lookupTask) {
                            if (Array.isArray(lookupTask)) {
                                if (lookupTask.some(
                                        function (item) {
                                            return item.TaskKey !== ci.TaskKey && (!item.rowstate || item.rowstate.indexOf('D') < 0);
                                        })
                                ) {
                                    addError("|26626|This Task ID is already in use for this project.  Task IDs must be unique within each project.|");
                                }
                            } else {
                                if (lookupTask.TaskKey !== ci.TaskKey) {
                                    addError("|26626|This Task ID is already in use for this project.  Task IDs must be unique within each project.|");
                                }
                            }
                        }
                    } else {
                        //if task id is null and auto iding, dont worry, it will get one right after.
                        if (ci.TrackBudget === 1 && project.AutoIDTask === 0) {
                            addError("|26627|A Task ID is required when tracking budgets and expenses|");
                        }
                    }

                    if (!response.hasEntityErrors()) {
                        if (project.AutoIDTask === 1 && u.obs(ci.TaskID).length === 0 && (ci.rowstate.indexOf('I') > -1 || ci.rowstate.indexOf('U') > -1)) {
                            invalidateSchedule();
                        }
                    }
                };

                this.TaskName = function () {
                    if (!validateRequired(ci.TaskName)) {
                        addError('|26628|A Task Name is required|');
                    }
                };

                this.ScheduleTask = function () {

                    if (ci.ScheduleTask === 0) {
                        //for both task types (summary, tracking), verify predecessor relatonships do not exist
                        //is this task a predecessor to other tasks?
                        if (scheduleMgr.isPredecessor(ci)) {//PredecessorKey
                            ci.ScheduleTask = 1;
                            addNotification("|26629|This task is a predecessor to one or more tasks, it must be shown on the schedule|");
                        }

                        //is this task a preceded by other tasks?
                        if (scheduleMgr.hasPredecessors(ci)) {//"TaskKey"
                            ci.ScheduleTask = 1;
                            addNotification("|26630|This task is preceded by one or more tasks, it must be shown on the schedule|")
                        }

                        //check tracking tasks to make sure they do not have any assigned users
                        if (ci.TaskType === Enum.ProjectScheduleTaskType.TRACKING_TASK) {
                            if (ci.taskUsers && ci.taskUsers.length > 0) {
                                ci.ScheduleTask = 1;
                                addNotification("|26631|This task must remain a scheduled task because it has staff assigned|");
                            }
                        }
                        if (ci.ScheduleTask === 0) {
                            //if it's still off, turn off PercCompSeparate
                            ci.PercCompSeparate = 0;
                        }
                    } else {
                        setScheduleTaskFlagUp(ci);
                    }
                    //for both turning on or off, if no errors rebuild the predecessor list and reschedule
                    if (!hasErrors()) {
                        invalidateSchedule(true);
                    }
                      
                };

                this.TrackBudget = function () {

                    if (ci.TrackBudget === 1) {
                        //verify that no parents or children are tracking budget already
                        if (checkTrackBudgetState(ci, 1)) {
                            //this task can track budget, cascade MoneyTask = 1 to all parent tasks above it and BTK down
                            setTrackBudgetFlags(ci, 1);
                        } else {
                            ci.TrackBudget = 0;
                            addNotification("|26632|This task can not track budget and expenses because one of it's parents or children is already tracking budget and expenses|");
                        }
                    } else {
                        ci.AllowAnyone = 0;
                        Remote.remoteDataCall('Modules.Projects.Schedule', 'checkMoneyTrans', {taskKey: ci.TaskKey}).done(function (results) {
                            if (results.returnValue > 0) {
                                if (ci.TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                                    addNotification("|26633|The Scheduling Constraint Selected Must Track Budget And Expenses.|");
                                    ci.TrackBudget = 1;
                                } else {
                                    //the flag can be turned off, cascade MoneyTask = 0 to all parent tasks above it
                                    setTrackBudgetFlags(ci, 0);
                                    ci.AllowAnyone = 0;
                                }
                            } else {
                                ci.TrackBudget = 1;

                                switch (results.returnValue) {
                                    case -2:
                                        addNotification("|26634|You must track budgets on this task because time entries have already been tied to this task|");
                                        break;
                                    case -3:
                                        addNotification("|26635|You must track budgets on this task because expense reports have already been tied to this task|");
                                        break;
                                    case -4:
                                        addNotification("|26636|You must track budgets on this task because misc cost entries have already been tied to this task|");
                                        break;
                                    case -5:
                                        addNotification("|26637|You must track budgets on this task because vendor invoices have already been tied to this task|");
                                        break;
                                    case -6:
                                        addNotification("|26638|You must track budgets on this task because purchase orders have already been tied to this task|");
                                        break;
                                    case -7:
                                        addNotification("|26639|You must track budgets on this task because quotes have already been tied to this task|");
                                        break;
                                    case -8:
                                        addNotification("|26640|You must track budgets on this task because change orders have already been tied to this task|");
                                        break;
                                    case -9:
                                        addNotification("|26641|You must track budgets on this task because estimates/COs have already been tied to this task|");
                                        break;
                                }
                            }

                            if (hasNotifications()) {
                                System.notify.fire('deferredScheduleValidationNotification', getNotifications());
                            }

                        });
                    }
                };

                this.ConstraintName =
                    this.TaskConstraint = function () {

                        if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                        // If not a schedule task, exit
                        if (ci.ScheduleTask === 0) return;

                        //if we have a constraint type that requires a constraint date, validate
                        if (ci.TaskConstraint !== Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT) {
                            if (!ci.ConstraintDate) {
                                if (ci.TaskConstraint === Enum.ProjectDirection.FROM_PROJECT_START) {
                                    ci.ConstraintDate = ci.PlanStart;
                                }
                                if (ci.TaskConstraint === Enum.ProjectDirection.FROM_PROJECT_END) {
                                    ci.ConstraintDate = ci.PlanComplete;
                                }
                            }
                            if (u.obn(ci.ConstraintDayOfTheWeek) !== 0) {
                                ci.ConstraintDayOfTheWeek = 0;
                            }

                            if (ci.TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                                if (scheduleMgr.hasPredecessors(ci)) {
                                    addError("|26642|The Scheduling Constraint Selected Does Not Allow Any Predecessors|", "TaskConstraint");
                                }
                                if (scheduleMgr.isPredecessor(ci)) {
                                    addError("|26643|The Scheduling Constraint Selected Can Not Be a Predecessor On Another Task|", "TaskConstraint");
                                }
                                ci.ConstraintDate = null;
                            }
                            this.ConstraintDate();
                        } else {
                            invalidateSchedule(true);
                        }
                    };

                this.ConstraintDate = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    // Convert to what the sm wants
                    var findNextDateDirection = 0;
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        findNextDateDirection = Enum.ProjectScheduleDirection.DIRECTION_FWD;
                    } else {
                        findNextDateDirection = Enum.ProjectScheduleDirection.DIRECTION_BWD;
                    }

                    //if Must Start On or Must Finish On, it has to be a working day
                    if (ci.TaskConstraint === Enum.ProjectScheduleConstraintTypes.MUST_START_ON || ci.TaskConstraint === Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON) {
                        // if does not work anyday, sm.isWorkingDay checks flag also but faster
                        if (ci.WorkAnyDay === 0) {
                            if (!isWorkingDay(ci, ci.ConstraintDate)) {
                                ci.ConstraintDate = findNextWorkDate(ci.ConstraintDate, findNextDateDirection);
                                addNotification("|26644|The Scheduling Constraint date you entered is not a working date. The next open date has been selected for you|");
                            }
                        }
                    }

                    if (validateRow) return;
                    invalidateSchedule(true);
                };

                this.PlanDuration = function () {

                    var workAnyDay = (ci.WorkAnyDay === 1);

                    // If a summary task, exit
                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    if (isNaN(ci.PlanDuration)) ci.PlanDuration = 0;
                    //if duration > 0 then it cannot be show on the calendar
                    if (ci.PlanDuration > 1 && ci.ShowOnCalendar) {
                        ci.ShowOnCalendar = 0;
                        ci.EventStart = null;
                        ci.EventEnd = null;
                        addNotification("|26645|The Plan Duration of the task must be 0 or 1 to show on Calendar|");
                    }
                    if (ci.PlanDuration < 0) {
                        ci.PlanDuration = 0
                        addNotification("|26646|The Plan Duration can not be less than zero|");
                    }

                    //make sure that the PlanStart is not on a blockout day
                    if (!workAnyDay) {
                        while (!isWorkDay(ci.PlanStart, true, true)) {
                            ci.PlanStart = Date.add(ci.PlanStart, 1, Date.DAY);
                        }
                    }

                    //for row we will reschedule after all validations
                    if (validateRow) return;
                    //set the constraint and set complete data
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        //ci.TaskConstraint = ProjectConstants.START_NO_EARLIER_THAN;
                        if (ci.PlanDuration > 1) ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                    } else {
                        //ci.TaskConstraint = ProjectConstants.FINISH_NO_LATER_THAN;
                        if (ci.PlanDuration > 1) ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                    }
                    if (!hasErrors()) {
                        scheduleMgr.calcColumns(ci);
                        invalidateSchedule(true);
                    }
                };

                this.ExcludeFromStatus = function () {
                    invalidateSchedule(true);
                };

                this.WorkAnyDay = function () {
                    invalidateSchedule(true);
                };

                this.PlanStart = function () {

                    var showWorkingDayError = true;
                    var workAnyDay = (ci.WorkAnyDay === 1);

                    // If a summary task, exit
                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                        return;
                    }
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) {
                        return;
                    }

                    if (!ci.PlanStart) {
                        ci.PlanStart = ci.PlanComplete; // for lack of a better value
                        ci.PlanDuration = 0;
                        invalidateSchedule(true);
                        return;
                    }

                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {

                        if (!isWorkingDay(ci, ci.PlanStart)) {
                            // if the user entered an invalid ActStart we will not display an error
                            if (ci.ActStart !== null) {
                                if ((!isWorkingDay(ci.TaskKey, ci.ActStart)) && (parms.prefs.ActualDateUpdatesPlan === 1)) {
                                    showWorkingDayError = false;
                                }
                            }
                            if (showWorkingDayError) {
                                ci.PlanStart = findNextWorkDate(ci.PlanStart, 1);
                                addNotification("|26647|The date you entered is not a working date. The next open date has been selected for you|");
                            }
                        }
                        if (validateRow){
                            if(ci.PlanComplete && ci.PlanStart > ci.PlanComplete){
                                ci.PlanStart = ci.PlanComplete;
                            }
                            return;
                        }

                        switch (ci.TaskConstraint) {
                            case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                                //Change to Start no earlier than and set the constraint
                                setConstraint(ci, Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN);
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_START_ON:
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.ConstraintDate);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.ConstraintDate);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.ConstraintDate);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN:
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN:
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                        }
                    } else {
                        if (!isWorkingDay(ci, ci.PlanStart)) {
                            // if the user entered an invalid ActStart we will not display an error
                            if (ci.ActStart !== null) {
                                if ((!isWorkingDay(ci, ci.ActStart)) && (pm.ActualDateUpdatesPlan === 1)) {
                                    showWorkingDayError = false;
                                }
                            }
                            if (showWorkingDayError) {
                                ci.PlanStart = findNextWorkDate(ci.PlanStart, 1);
                                addNotification("|26647|The date you entered is not a working date. The next open date has been selected for you|");
                            }

                        }
                        if (validateRow) return;
                        switch (ci.TaskConstraint) {
                            case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_START_ON:
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.ConstraintDate);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.ConstraintDate);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.ConstraintDate);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN:
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN:
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                        }
                    }
                    if (ci.PlanDuration > 1 && ci.ShowOnCalendar) {
                        ci.ShowOnCalendar = 0;
                        ci.EventStart = null;
                        ci.EventEnd = null;
                        addNotification("|26645|The Plan Duration of the task must be 0 or 1 to show on Calendar|");
                    }
                    if(ci.PlanComplete && ci.PlanStart > ci.PlanComplete){
                        ci.PlanStart = ci.PlanComplete;
                    }

                    scheduleMgr.calcColumns(ci);
                    invalidateSchedule(true);
                };

                this.PlanComplete = function () {

                    var workAnyDay;

                    // If a summary task, exit
                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    if (!ci.PlanComplete) {
                        ci.PlanComplete = ci.PlanStart; // for lack of a better value
                        ci.PlanDuration = 0;
                        invalidateSchedule(true);
                        return;
                    }
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        if (!isWorkingDay(ci, ci.PlanComplete)) {
                            var showWorkingDayError = true;
                            if (ci.ActComplete !== null) {
                                if ((!isWorkingDay(ci, ci.ActComplete)) && (parms.prefs.ActualDateUpdatesPlan === 1)) {
                                    showWorkingDayError = false;
                                }
                            }
                            if (showWorkingDayError) {
                                ci.PlanComplete = findNextWorkDate(ci.PlanComplete, Enum.ProjectScheduleDirection.DIRECTION_FWD);
                                addNotification("|26647|The date you entered is not a working date. The next open date has been selected for you|");
                            }
                        }
                        if (validateRow) return;
                        workAnyDay = (ci.WorkAnyDay === 1);
                        switch (ci.TaskConstraint) {
                            case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                                // Modify the duration
                                //CRG: See DEV1197, Task 3.6 for info on this modification:
                                setConstraint(ci, Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON);
                                ci.ConstraintDate = ci.PlanComplete;
                                ci.PlanStart = calcNextDate(ci.PlanDuration, ci.PlanComplete, Enum.ProjectScheduleDirection.DIRECTION_BWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_START_ON:
                                // Modify the duration
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON:
                                ci.ConstraintDate = ci.PlanComplete;
                                // Duration says the same
                                // Plan start must move
                                ci.PlanStart = calcNextDate(ci.PlanDuration, ci.PlanComplete, Enum.ProjectScheduleDirection.DIRECTION_BWD, workAnyDay)
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN:
                                ci.ConstraintDate = ci.PlanComplete;
                                // Duration says the same
                                // Plan start must move
                                ci.PlanStart = calcNextDate(ci.PlanDuration, ci.PlanComplete, Enum.ProjectScheduleDirection.DIRECTION_BWD, workAnyDay)
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN:
                                ci.ConstraintDate = ci.PlanComplete;
                                // Duration says the same
                                // Plan start must move
                                ci.PlanStart = calcNextDate(ci.PlanDuration, ci.PlanComplete, Enum.ProjectScheduleDirection.DIRECTION_BWD, workAnyDay)
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                        }
                    } else {
                        if (!isWorkingDay(ci, ci.PlanComplete)) {
                            ci.PlanComplete = findNextWorkDate(ci.PlanComplete, Enum.ProjectScheduleDirection.DIRECTION_FWD);
                            addNotification("|26647|The date you entered is not a working date. The next open date has been selected for you|");
                        }
                        if (validateRow) return;
                        workAnyDay = (ci.WorkAnyDay === 1);
                        switch (ci.TaskConstraint) {
                            case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                                setConstraint(ci, Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN);
                                ci.ConstraintDate = ci.PlanStart;
                                ci.PlanComplete = calcNextDate(ci.PlanDuration, ci.PlanStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, workAnyDay);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_START_ON:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON:
                                ci.ConstraintDate = ci.PlanComplete;
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN:
                                ci.ConstraintDate = ci.PlanComplete;
                                break;
                            case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN:
                                ci.ConstraintDate = ci.PlanComplete;
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                            case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN:
                                ci.PlanDuration = calcTaskDuration(ci.TaskKey, ci.PlanStart, ci.PlanComplete);
                                break;
                        }
                    }
                    if (ci.PlanDuration > 1 && ci.ShowOnCalendar) {
                        ci.ShowOnCalendar = 0;
                        ci.EventStart = null;
                        ci.EventEnd = null;
                        addNotification("|26645|The Plan Duration of the task must be 0 or 1 to show on Calendar|");
                    }
                    scheduleMgr.calcColumns(ci);
                    invalidateSchedule(true);
                };

                this.GanttPlanComplete = function () {
                    this.PlanComplete();
                };

                this.BaseStart = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    if (ci.BaseStart) {
                        if (ci.BaseComplete) {
                            if (ci.BaseStart > ci.BaseComplete) {
                                addError("|26648|The Original Start date cannot be greater than the Original Complete date|");
                            }
                        }
                    }
                };

                this.BaseComplete = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                        return;
                    }
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) {
                        return;
                    }

                    if (ci.BaseComplete) {
                        if (ci.BaseStart) {
                            if (ci.BaseComplete < ci.BaseStart) {
                                addError("|26649|The Original Complete date cannot be less than the Original Start date|");
                            }
                        }
                    }
                };

                this.ActStart = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    if (!ci.ActStart) {
                        if (ci.PercComp > 0) {
                            ci.PercComp = 0;
                        }
                        if (ci.ActComplete) {
                            ci.ActComplete = null;
                        }
                    } else {
                        if (ci.ActComplete) {
                            if (ci.ActStart > ci.ActComplete) {
                                ci.ActComplete = ci.ActStart;
                            }
                        }
                    }
                    pushDownActuals(ci, "ActStart");
                    ci.UpdateActuals = 'ActStart';

                    //for row we will reschedule after all validations
                    if (validateRow) return;
                    //do we really need to do this now?
                    if (!hasErrors()) {
                        invalidateSchedule(true);
                    }
                        
                };

                this.ActComplete = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    if (!ci.ActComplete) {
                        if (ci.PercComp === 100) {
                            ci.PercComp = 0;
                        }
                    } else {
                        if (!ci.ActStart) {
                            ci.ActStart = Date.ISOToday;
                        }
                        if (ci.ActStart) {
                            if (ci.ActComplete < ci.ActStart) {
                                ci.ActStart = ci.ActComplete;
                            }
                        }
                        if (ci.PercComp < 100) {
                            ci.PercComp = 100;
                            if(ci.TCProjectStatusKey > 0){
                                //set this task as the "last" task to have it's percent set to 100
                                project.SetProjectStatusTaskKey = ci.TaskKey;
                            }
                        }
                    }
                    pushDownActuals(ci, "ActComplete");
                    ci.UpdateActuals = 'ActComplete';

                    //for row we will reschedule after all validations
                    if (validateRow) return;

                    if (!hasErrors()) {
                        calcBucketHours(ci.tuhTaskParms, ci);
                        invalidateSchedule(true);
                    }
                };

                this.PercComp = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;

                    if (ci.PercComp > 100) {
                        ci.PercComp = 100;
                    }
                    if (ci.PercComp > 0 && !ci.ActStart) {
                        ci.ActStart = Date.ISOToday;
                    }
                    if (ci.PercComp === 100 && !ci.ActComplete) {
                        ci.ActComplete = Date.ISOToday;
                    }
                    if (ci.PercComp < 100 && ci.ActComplete) {
                        ci.ActComplete = null;
                    }
                    if (ci.ActStart > ci.ActComplete) {
                        ci.ActComplete = ci.ActStart;
                    }

                    if(ci.PercComp === 100 && ci.TCProjectStatusKey > 0){
                        //set this task as the "last" task to have it's percent set to 100
                        project.SetProjectStatusTaskKey = ci.TaskKey;
                    }

                    //Now pushing down the PercComp in all cases
                    pushDownActuals(ci, 'PercComp');

                    ci.UpdateActuals = 'PercComp';

                    //for row we will reschedule after all validations
                    if (validateRow) return;

                    if (!hasErrors()) {
                        invalidateSchedule(true);
                    }
                       
                };

                this.ConstraintDayOfTheWeek = function () {

                    // If a summary task, exit
                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;
                    // If not a schedule task, exit
                    if (ci.ScheduleTask === 0) return;

                    if (ci.TaskConstraint === Enum.ProjectScheduleConstraintTypes.AS_SOON_AS_POSSIBLE) {
                        if (ci.WorkAnyDay === 0) {
                            if (!isProjectWorkDay(u.obn(ci.ConstraintDayOfTheWeek))) {
                                addError("|26650|This date is not a working day for the project. Select a valid work date|");
                                return;
                            }
                        }
                    }

                    if (validateRow) return;
                    invalidateSchedule(true);
                };

                this.Actuals = function () {

                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;

                    if (ci.PercComp > 100) {
                        ci.PercComp = 100;
                    }

                    if (ci.PercComp > 0 && !ci.ActStart) {
                        ci.ActStart = Date.ISOToday;
                    }
                    if (ci.PercComp === 100 && !ci.ActComplete) {
                        ci.ActComplete = Date.ISOToday;
                    }
                    if (ci.ActStart > ci.ActComplete) {
                        ci.ActComplete = ci.ActStart;
                    }
                    if (ci.PercComp < 100 && ci.ActComplete) {
                        ci.PercComp = 100;
                    }
                };

                this.CalendarEvent = function () {

                    // If a summary task, exit
                    if (ci.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) return;

                    // ShowOnCalendar = 1 ==> EventStart and EventEnd are required
                    // ShowOnCalendar = 0 ==> No EventStart on UI but EventEnd is optional

                    if (ci.ShowOnCalendar === 1) {
                        if (!validateRequired(ci.EventStart)) {
                            addError('|26651|A Start Time is required to show a task on the calendar|');
                        }
                        if (!validateRequired(ci.EventEnd)) {
                            addError('|26652|An End Time is required to show a task on the calendar|');
                        }

                        if (ci.EventStart && ci.EventEnd) {
                            if (ci.EventEnd < ci.EventStart) {
                                addError("|26653|The End Time cannot be less than the Start Time|");
                            }
                        }
                    } else {
                        ci.EventStart = null;
                        ci.EventEnd = null;
                    }
                };

                this.AssignedPreds = function () {

                    var predList = scheduleMgr.getPreds(ci, true, true);

                    //Build associative array of the current DS lag values and cross-project preds
                    var lagList = {};
                    var crossProject = {};
                    var rowstateItems = {}; //if the user edits the preds in the grid on the right, they'll have a rowstate (if they used the select list in the grid, it will not yet)
                    for (var p = 0; p < dsPreds.length; p++) {
                        lagList[dsPreds[p].PredecessorKey] = u.obn(dsPreds[p].Lag);
                        if (dsPreds[p].CrossProject === 1) {
                            crossProject[dsPreds[p].PredecessorKey] = dsPreds[p];
                        }
                        if (dsPreds[p].rowstate) {
                            rowstateItems[dsPreds[p].PredecessorKey] = dsPreds[p].rowstate;
                        }
                    }

                    //Build associative array of the last lag values prior to this change
                    var currentPreds = {};
                    var i;
                    for (i = 0; i < predList.length; i++) {
                        currentPreds[predList[i].PredKey] = predList[i].Lag || 0;
                    }

                    var selectedPredKeys = ci.AssignedPredsKeys ? ci.AssignedPredsKeys.split(',') : [];

                    var key;
                    var insertPreds = [];
                    var deletePreds = [];
                    var predLagUpdated = false;

                    for (i = 0; i < selectedPredKeys.length; i++) {
                        key = u.obn(selectedPredKeys[i]);
                        //if the key is in currentPreds, that means the pred was already set (or was set using the lookup in the grid at the right...which in that case we have to use the rowstate)
                        if (currentPreds[key] !== undefined) {
                            if (lagList[key] !== undefined) {
                                if (lagList[key] !== currentPreds[key]) {
                                    predLagUpdated = true; //set this flag so that the scheduleMgr pred arrays can be reloaded at the end
                                    currentPreds[key] = 's'; //still selected
                                } else {
                                    currentPreds[key] = 's'; //still selected
                                }
                            } else {
                                currentPreds[key] = 'd'; //it's not in the DS anymore, mark it for delete
                            }
                            if (rowstateItems[key]) {
                                if (rowstateItems[key].indexOf('I') >= 0 || rowstateItems[key].indexOf('U') >= 0) {
                                    insertPreds.push({key: key, lag: lagList[key] || 0});
                                }
                            }
                        } else {
                            if (key !== ci.TaskKey) {
                                insertPreds.push({key: key, lag: lagList[key] || 0});
                            }
                        }
                    }

                    for (var k in currentPreds) {
                        if (currentPreds[k] !== 's') {
                            deletePreds.push(k);
                        }
                    }

                    var linkTask;

                    for (i = 0; i < insertPreds.length; i++) {
                        linkTask = crossProject[insertPreds[i].key] ? crossProject[insertPreds[i].key] : scheduleMgr.findTaskByKey(insertPreds[i].key);
                        scheduleMgr.linkToTask(ci, linkTask, insertPreds[i].lag, true);
                        scheduleMgr.markTaskDirty(ci);
                        if (hasGeneralErrors()) {
                            var findErrorPred = dsPreds.findByKey(insertPreds[i].key, 'PredecessorKey');
                            if (findErrorPred) {
                                scheduleMgr.removePredecessorFromSource(findErrorPred);
                                scheduleMgr.formatPredValues();
                                return;
                            }
                        }
                    }
                    for (i = 0; i < deletePreds.length; i++) {
                        linkTask = scheduleMgr.findTaskByKey(deletePreds[i]);
                        if(!linkTask){
                            //see if it's a cross project
                            linkTask = crossProject[u.obn(deletePreds[i])];
                        }
                        scheduleMgr.unlinkFromTask(ci, linkTask, true);
                        scheduleMgr.markTaskDirty(ci);
                        if (hasGeneralErrors()) {
                            var findErrorPred = dsPreds.findByKey(deletePreds[i].key, 'PredecessorKey');
                            if (findErrorPred) {
                                scheduleMgr.removePredecessorFromSource(findErrorPred);
                                scheduleMgr.formatPredValues();
                                return;
                            }
                        }
                    }

                    if (!hasGeneralErrors()) {
                        //if lag has been updated, tell scheduleMgr to reload the internal arrays from the DS
                        if (predLagUpdated) scheduleMgr.lagUpdated(ci);

                        if (validateRow) return;
                        invalidateSchedule(true);
                    }
                };

                this.validateRowCleanup = function () {

                    if (ci.ScheduleTask === 0) {
                        // do some kind of cleanup
                        // all these fields do not mean anything if not a schedule task
                        setConstraint(ci, Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT);
                        ci.TaskStatus = 1;
                        ci.ScheduleNote = "";
                        ci.PlanDuration = 0;
                        ci.PercComp = 0;

                        if (project.ScheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                            ci.PlanStart = project.StartDate;
                            ci.PlanComplete = project.StartDate;
                            ci.ActStart = null;
                            ci.ActComplete = null;
                        } else {
                            ci.PlanStart = project.CompleteDate;
                            ci.PlanComplete = project.CompleteDate;
                            ci.ActStart = null;
                            ci.ActComplete = null;
                        }
                    }
                };

                break;

            /**
             * TASK USER field change functions
             */
            case window.Schedule.TASK_USERS:

                var currentTask = scheduleMgr.findTaskByKey(ci.TaskKey);
                var tuhTaskParms = currentTask ? currentTask.tuhTaskParms : null;
                if (currentTask) {
                    scheduleMgr.markTaskDirty(currentTask);
                    response.task = currentTask;
                }

                this.Hours = function (fromPercAssigned) {

                    if (!options.retainer && ci.ManualAdjustment === 1) {
                        ci.LogManualFlagOffComment = 'Task ID: ' + currentTask.TaskID + '. Assignment: ' + u.obs(ci.UserName) + '/' + u.obs(ci.ServiceDescription) + ' manual adjustment turned off due to modification of Allocated Hours';
                    }
                    ci.ManualAdjustment = 0;
                    var duration = calcBucketHours(tuhTaskParms, ci);
                    this.rollupActuals();

                    if(options.retainer){
                        if (!fromPercAssigned) {
                            if(duration === 0){
                                ci.PercAssigned = 0;
                            }else {
                                ci.PercAssigned = Math.roundTo((ci.Hours / window.Schedule.RETAINER_DAILY_HOURS / duration) * 100, 0);
                            }
                        }

                        //after it calcs the hours, set it to Manual if it's a retainer
                        ci.ManualAdjustment = 1;
                    }

                    invalidateSchedule(true);
                };

                //this only applies to retainers
                this.PercAssigned = function () {
                    ci.Hours = Math.roundTo((ci.PercAssigned / 100) * window.Schedule.RETAINER_DAILY_HOURS * calcBucketHours(tuhTaskParms, ci, true, null, true), 2);
                    this.Hours(true);
                };

                this.APlanStart = this.APlanComplete = function () {

                    if (!options.retainer && ci.ManualAdjustment === 1) {
                        ci.LogManualFlagOffComment = 'Task ID: ' + currentTask.TaskID + '. Assignment: ' + u.obs(ci.UserName) + '/' + u.obs(ci.ServiceDescription) + ' manual adjustment turned off due to modification of Assignment Dates';
                    }
                    ci.ManualAdjustment = 0;
                    calcBucketHours(tuhTaskParms, ci);
                    this.rollupActuals();

                    if(options.retainer){
                        //after it calcs the hours, set it to Manual if it's a retainer
                        ci.ManualAdjustment = 1;
                    }

                    invalidateSchedule(true);
                };

                this.Offset = function (field) {
                    ci.ManualAdjustment = 1;
                    ci.LogManualFlagOffComment = null;

                    if (field.length > 7) {
                        var offset = u.obn(field.substring(7));
                        TaskUserHour.calcBucketTotal(tuhTaskParms, ci, offset);

                        if(options.retainer){
                            var duration = calcBucketHours(tuhTaskParms, ci, true, null, true);
                            if(duration === 0){
                                ci.PercAssigned = 0;
                            }else{
                                ci.PercAssigned = Math.roundTo((ci.Hours / window.Schedule.RETAINER_DAILY_HOURS / duration) * 100, 0);
                            }
                        }
                    }
                };

                this.ActStart = function () {

                    if (currentTask && currentTask.PercCompSeparate === 0) return;

                    if (!ci.ActStart) {
                        if (ci.PercComp > 0) {
                            ci.PercComp = 0;
                        }
                        if (ci.ActComplete) {
                            ci.ActComplete = null;
                        }
                    } else {
                        if (ci.ActComplete) {
                            if (ci.ActStart > ci.ActComplete) {
                                ci.ActComplete = ci.ActStart;
                            }
                        }
                    }

                    calcBucketHours(tuhTaskParms, ci);
                    this.rollupActuals();
                    invalidateSchedule(true);
                };

                this.ActComplete = function () {

                    if (currentTask && currentTask.PercCompSeparate === 0) return;

                    if (!ci.ActComplete) {
                        if (ci.PercComp === 100) {
                            ci.PercComp = 0;
                        }
                    } else {
                        if (!ci.ActStart) {
                            ci.ActStart = Date.ISOToday;
                        }
                        if (ci.ActStart) {
                            if (ci.ActComplete < ci.ActStart) {
                                ci.ActStart = ci.ActComplete;
                            }
                        }
                        if (ci.PercComp < 100) {
                            ci.PercComp = 100;
                        }
                    }

                    calcBucketHours(tuhTaskParms, ci);
                    this.rollupActuals();
                    invalidateSchedule(true);
                };

                this.PercComp = function () {

                    if (currentTask && currentTask.PercCompSeparate === 0) return;

                    if (ci.PercComp > 100) {
                        ci.PercComp = 100;
                    }
                    if (ci.PercComp > 0 && !ci.ActStart) {
                        ci.ActStart = Date.ISOToday;
                    }
                    if (ci.PercComp === 100 && !ci.ActComplete) {
                        ci.ActComplete = Date.ISOToday;
                    }
                    if (ci.PercComp < 100 && ci.ActComplete) {
                        ci.ActComplete = null;
                    }
                    if (ci.ActStart > ci.ActComplete) {
                        ci.ActComplete = ci.ActStart;
                    }

                    if(ci.PercComp === 100 && currentTask.TCProjectStatusKey > 0){
                        //set this task as the "last" task to have it's percent set to 100
                        project.SetProjectStatusTaskKey = currentTask.TaskKey;
                    }

                    calcBucketHours(tuhTaskParms, ci);
                    this.rollupActuals();
                    invalidateSchedule(true);
                };

                this.rollupActuals = function () {

                    if (currentTask) {
                        rollupAssignedUsersToTask(currentTask);
                    }
                };

                this.validateRowCleanup = function () {

                    if (ci.rowstate !== 'D') {
                        if (options.retainer) {
                            if (!ci.ServiceKey) {
                                addError('|27584|A service is required.|');
                            }
                            for (var i = 0; i < dsTaskUsers.source.length; i++) {
                                if (dsTaskUsers.source[i] !== ci) {
                                    //if the row has not been changed at all, OR it has been changed and it's not being deleted...
                                    if (!dsTaskUsers.source[i].rowstate || (dsTaskUsers.source[i].rowstate && dsTaskUsers.source[i].rowstate.indexOf('D') < 0)) {
                                        if(ci.UserKey > 0) {
                                            if (dsTaskUsers.source[i].ServiceKey === ci.ServiceKey && dsTaskUsers.source[i].UserKey === ci.UserKey) {
                                                addError('|27585|The user and service must be unique|');
                                            }
                                        }
                                    }
                                }
                            }
                        }else {
                            if (!ci.UserKey && !ci.ServiceKey) {
                                addError('|26654|Either a service or a user is required.|');
                            }
                        }
                    }

                    this.rollupActuals();
                };
                break;
        }


        //************************************************************************
        // Helper Functions
        //************************************************************************
        function addError(msg) {
            //field is added to the error automatically by the response object
            response.addError(msg);
        }

        function addNotification(msg) {
            response.addNotification(msg);
        }

        function hasErrors() {
            return response.hasErrors();
        }

        function hasNotifications() {
            return response.hasNotifications();
        }

        function getErrors() {
            return response.getErrors();
        }

        function getNotifications() {
            return response.getNotifications();
        }

        function getParentTask(task) {
            return scheduleMgr.getParentTask(task);
        }

        function findTaskByKey(keyValue, keyField) {
            return scheduleMgr.findTaskByKey(keyValue, keyField);
        }

        function isWorkDay(dte, includeCompany, includeProject) {
            //this makes it easier to bring over code copied from Flex that called isWorkDay directly
            return scheduleMgr.isWorkDay(dte, includeCompany, includeProject);
        }

        function isWorkingDay(task, dte) {

            if (task && task.WorkAnyDay) {
                return true;
            }

            return isWorkDay(dte, true, true);
        }

        function isProjectWorkDay(day) {
            return scheduleMgr.isProjectWorkDay(day)
        }

        function findNextWorkDate(dte, direction) {
            return scheduleMgr.findNextWorkDate(dte, direction);
        }

        function invalidateSchedule(skipAutoTaskID) {
            scheduleMgr.invalidateSchedule(skipAutoTaskID);
        }

        function validateSchedule() {
            return scheduleMgr.validateSchedule();
        }

        function hasGeneralErrors() {
            return scheduleMgr.hasGeneralErrors();
        }

        function getGeneralErrors() {
            return scheduleMgr.getGeneralErrors();
        }

        function hasGeneralNotifications() {
            return scheduleMgr.hasGeneralNotifications();
        }

        function getGeneralNotifications() {
            return scheduleMgr.getGeneralNotifications();
        }

        function setScheduleTaskFlagUp(task) {

            if (!task) return;

            if (task.ScheduleTask !== 1) {
                task.ScheduleTask = 1;
                if (task.SummaryTaskKey) {
                    setScheduleTaskFlagUp(getParentTask(task));
                }
            }
        }

        function checkTrackBudgetState(task, trackBudget) {
            return scheduleMgr.checkTrackBudgetState(task, trackBudget);
        }

        function setTrackBudgetFlags(task, trackBudget) {
            scheduleMgr.setTrackBudgetFlags(task, trackBudget);
        }


        function pushDownActuals(task, field) {
            scheduleMgr.pushDownActuals(task, field);
        }

        function rollupAssignedUsersToTask(task) {
            scheduleMgr.rollupAssignedUsersToTask(task);
            scheduleMgr.markTaskDirty(currentTask);
        }

        function calcBucketHours(tuhTaskParms, taskUser, skipDSFire, bucketDelta, durationOnly) {
            return scheduleMgr.calcBucketHours(tuhTaskParms, taskUser, skipDSFire, bucketDelta, durationOnly)
        }

        function calcBucketTotal(tuhTaskParms, taskUser, offset) {
            scheduleMgr.calcBucketTotal(tuhTaskParms, taskUser, offset)
        }

        function calcTaskDuration(task, startDate, endDate) {

            if (!startDate) {
                return 0;
            }
            if (!endDate) {
                return 0;
            }

            return calcDuration(startDate, endDate, task.PlanDuration, task.WorkAnyDay);
        }

        function calcDuration(dteStart, dteEnd, origDur, workAnyDay) {
            return scheduleMgr.calcDuration(dteStart, dteEnd, origDur, workAnyDay);
        }

        function calcNextDate(dur, dteDate, direction, workAnyDay) {
            return scheduleMgr.calcNextDate(dur, dteDate, direction, workAnyDay);
        }

        function setConstraint(task, constraint) {
            scheduleMgr.setConstraint(task, constraint);
        }

        /**
         * Returns true if a value is present
         * @param value
         * @returns {boolean}
         */
        function validateRequired(value) {
            return u.obs(value).trim().length > 0;
        }

        this.validateSchedule = validateSchedule;
        this.getGeneralErrors = getGeneralErrors;
        this.hasGeneralErrors = hasGeneralErrors;
        this.hasGeneralNotifications = hasGeneralNotifications;
        this.getGeneralNotifications = getGeneralNotifications;
    };
    window.ScheduleValidation = ScheduleValidation;

    //endregion

    /**
     * This class contains the core scheduling logic
     * @param parms
     * @param response
     * @constructor
     */
    var ScheduleMgr = function (parms, response) {

        var errors;
        var notifications;

        var options = parms.options || {};
        var usingDM = options.usingDM;

        var dsTasks = parms.dsTasks;
        var setOriginalValues = parms.setOriginalValues; //set a pointer to this function so that we can execute it later

        var dsPreds = parms.dsPreds;
        var predList = dsPreds.source;
        var crossProjectPreds = {};
        var deletedPreds = [];
        var deletedTaskUsers = [];
        var removeProjectUsers = [];
        var eventLog = [];

        var dsTaskUsers = parms.dsTaskUsers;
        var taskUsers = dsTaskUsers.source;

        var project = parms.project;

        var blockoutDays = parms.blockoutDays;
        var actualDateUpdatesPlan = u.obb(parms.prefs.ActualDateUpdatesPlan);
        var autoScheduleProjects = u.obb(parms.prefs.AutoScheduleProjects);
        var autoScheduleTaskEndDate = u.obb(parms.prefs.AutoScheduleTaskEndDate);
        var scheduleDirection = u.obn(project.ScheduleDirection);

        var invalidSchedule = false;
        var setNewTaskID = undefined;

        var scheduleOrder;
        var predPairs = "";
        var parentCheckedInCircDetection; // this is to speed up the search when detecting circ pred
        var taskProcessedInScheduling; // To recover from circular relationships in scheduling

        /**
         * Call this to set the invalidSchedule flag to true. This will cause the schedule logic to recalculate when validateSchedule is called
         */
        function invalidateSchedule(skipAutoTaskID) {

            if (skipAutoTaskID || options.retainer) {
                if(setNewTaskID === undefined){
                    setNewTaskID = false;
                }
            }else{
                setNewTaskID = true;
            }

            invalidSchedule = true;
        }

        /**
         * Returns true if the invalidSchedule flag has been set to true
         * @returns {boolean}
         */
        function scheduleIsInvalid() {
            return invalidSchedule;
        }

        /**
         * Causes the schedule to be recalculated if the invalidSchedule flag is true
         * @returns {boolean}
         */
        function validateSchedule() {
            if (scheduleIsInvalid()) {
                scheduleProject();
            }
            return !hasGeneralErrors(); //return false if there are errors, i.e. schedule did not validate
        }

        /**
         * Runs through from top to bottom and ensures the DisplayOrder, ProjectOrder, and TaskLevel are set properly
         */
        function validateStructure() {

            autoNumberTasks();
            setStructureValues();
            validatePredecessorKeys();
            schedulingSetPredecessorsComplete();
            logChanges();
        }

        function logChanges() {

            var item;
            var origValue;
            var newValue;

            for (var i = 0; i < dsTasks.source.length; i++) {
                item = dsTasks.source[i];
                //if an item is updated and then deleted, its rowstate = 'UD'. If inserted, then updated, its rowstate = 'IU'. So we need to make sure that it hasn't been inserted or deleted before logging changes
                if (item.rowstate.indexOf('U') >= 0 && item.rowstate.indexOf('D') < 0 && item.rowstate.indexOf('I') < 0) {
                    origValue = u.obn(originalValue(item, 'PlanDuration'));
                    newValue = u.obn(item.PlanDuration);
                    if (newValue !== origValue) {
                        logEvent(System.localize('|26655|Project Schedule Change|'), System.localize('|26656|Task|') + ' ' + item.TaskID + '-' + item.TaskName + ' ' +
                            System.localize('|26657|plan duration was updated from|') + ' ' + origValue.toString() + ' ' + System.localize('|26658|to|') + ' ' + newValue.toString());
                    }
                    origValue = u.obn(originalValue(item, 'PercComp'));
                    newValue = u.obn(item.PercComp);
                    if (newValue !== origValue) {
                        if (newValue === 100) {
                            logEvent(System.localize('|26659|Task Completed|'), System.localize('|26656|Task|') + ' ' + item.TaskID + '-' + item.TaskName + ' ' +
                                System.localize('|26660|was marked completed|'));
                        } else {
                            logEvent(System.localize('|26661|Task Percent Updated|'), System.localize('|26656|Task|') + ' ' + item.TaskID + '-' + item.TaskName + ' ' +
                                System.localize('|26662|task percent was updated from|') + ' ' + origValue.toString() + ' ' + System.localize('|26658|to|') + ' ' + newValue.toString());
                        }
                    }
                } else {
                    //check for marked completed
                    origValue = u.obn(originalValue(item, 'PercComp'));
                    newValue = u.obn(item.PercComp);
                    if (newValue === 100 && origValue !== 100) {
                        logEvent(System.localize('|26659|Task Completed|'), System.localize('|26656|Task|') + ' ' + item.TaskID + '-' + item.TaskName + ' ' +
                            System.localize('|26660|was marked completed|'));
                    }
                }
            }

            var taskItem;
            for (var i = 0; i < dsTaskUsers.source.length; i++) {
                item = dsTaskUsers.source[i];
                taskItem = findTaskByKey(item.TaskKey);
                if(taskItem) {
                    if (item.rowstate && item.rowstate.indexOf('U') >= 0) {
                        origValue = u.obn(originalValue(item, 'UserKey'));
                        newValue = u.obn(item.UserKey);
                        if (newValue !== origValue) {
                            if (newValue > 0) {
                                logEvent(System.localize('|26663|Task Assigned|'), System.localize('|26656|Task|') + ' ' + taskItem.TaskID + '-' + taskItem.TaskName + ' ' +
                                    System.localize('|26664|was reassigned to|') + ' ' + u.obs(item.UserName) + ' ' + System.localize('|26665|from|') + ' ' + u.obs(originalValue(item, 'UserName')) + ' ' + System.localize('|26666|for|') + ' (' + u.obn(item.Hours).toString() + ' ' + System.localize('|26667|hours|') + ')');
                            } else {
                                logEvent(System.localize('|26663|Task Assigned|'), System.localize('|26656|Task|') + ' ' + taskItem.TaskID + '-' + taskItem.TaskName + ' ' +
                                    System.localize('|26668|was unassigned from|') + ' ' + u.obs(originalValue(item, 'UserName')));
                            }
                        }
                        origValue = u.obn(originalValue(item, 'Hours'));
                        newValue = u.obn(item.Hours);
                        if (newValue !== origValue) {
                            logEvent(System.localize('|26669|Assignment Allocated Hours|'), System.localize('|26670|The allocated hours for|') + ' ' + u.obs(item.UserName, item.ServiceDescription) + ' ' + System.localize('|26671|on Task|') + ' ' + taskItem.TaskID + '-' + taskItem.TaskName + ' ' +
                                System.localize('|26672|was changed to|') + ' ' + u.obs(item.Hours) + ' ' + System.localize('|26665|from|') + ' ' + u.obs(originalValue(item, 'Hours')));
                        }
                    }
                    //if item.rowstate is missing, then it's an insert (copied from a template)
                    if (item.rowstate === undefined || item.rowstate.indexOf('I') >= 0) {
                        newValue = u.obn(item.UserKey);
                        if (newValue > 0) {
                            logEvent(System.localize('|26663|Task Assigned|'), System.localize('|26656|Task|') + ' ' + taskItem.TaskID + '-' + taskItem.TaskName + ' ' +
                                System.localize('|26673|was assigned to|') + ' ' + u.obs(item.UserName) + ' ' + System.localize('|26666|for|') + ' (' + u.obn(item.Hours).toString() + ' ' + System.localize('|26667|hours|') + ')');
                        }
                    }
                    if (item.rowstate && item.rowstate.indexOf('D') >= 0) {
                        origValue = u.obn(originalValue(item, 'UserKey'));
                        if (origValue > 0) {
                            logEvent(System.localize('|26663|Task Assigned|'), System.localize('|26656|Task|') + ' ' + taskItem.TaskID + '-' + taskItem.TaskName + ' ' +
                                System.localize('|26668|was unassigned from|') + ' ' + u.obs(originalValue(item, 'UserName')));
                        }
                    }
                }
            }
        }

        function originalValue(item, field) {
            if (!item) return;
            if (!item.original) return;
            return item.original[field];
        }

        /**
         * Recursive routine that sets the DisplayOrder, ProjectOrder, and TaskLevel
         * @param parentNode
         * @param projectOrder
         * @param taskLevel
         * @returns {*|number}
         */
        function setStructureValues(parentNode, projectOrder, taskLevel) {

            projectOrder = projectOrder || 1;
            taskLevel = taskLevel || 0;

            var taskList = parentNode ? getChildren(parentNode) : dsTasks.source.filter(function (item) { return !item.SummaryTaskKey || item.SummaryTaskKey === 0});
            var summaryTaskKey = parentNode ? parentNode.TaskKey : 0;

            var task;
            for (var i = 0; i < taskList.length; i++) {
                task = taskList[i];

                task.ProjectOrder = projectOrder;
                projectOrder++;

                task.DisplayOrder = i + 1;
                task.TaskLevel = taskLevel;
                task.SummaryTaskKey = summaryTaskKey;
                task.PercCompSeparate = u.obn(task.PercCompSeparate);


                checkMoneyTaskFlag(task);

                if(u.obn(task.TrackBudget) === 1){
                    setTrackBudgetFlags(task, 1);
                }

                if (hasChildren(task)) {
                    task.TaskType = Enum.ProjectScheduleTaskType.SUMMARY_TASK;
                    projectOrder = setStructureValues(task, projectOrder, taskLevel + 1);

                    //Check to see if the task has assignments. If so, remove them from the task because Summary tasks can't have assignments
                    if (task.taskUsers) {
                        var tempTU = [];
                        for (var j = 0; j < task.taskUsers.length; j++) {
                            tempTU.push(task.taskUsers[j]);
                        }
                        for (var j = 0; j < tempTU.length; j++) {
                            if(!usingDM) dsTaskUsers.removeItem(tempTU[j]);
                            removeTaskUser(tempTU[j]);
                        }
                    }
                } else {
                    task.TaskType = Enum.ProjectScheduleTaskType.TRACKING_TASK;
                    if (u.obn(task.TrackBudget) === 0) task.MoneyTask = 0;

                    //ensure that ScheduleTask is set, if needed
                    if (task.ScheduleTask === 0) {
                        if (task.taskUsers && task.taskUsers.length > 0) {
                            var tu = task.taskUsers.filter(function (item) {
                                return !item.rowstate || item.rowstate.indexOf('D') <= 0;
                            });

                            if (tu.length > 0) task.ScheduleTask = 1;
                        }
                    }
                    if (task.ScheduleTask === 0) {
                        if (isPredecessor(task)) task.ScheduleTask = 1;
                    }
                    if (task.ScheduleTask === 0) {
                        if (hasPredecessors(task)) task.ScheduleTask = 1;
                    }
                }
            }
            return projectOrder;
        }

        function validatePredecessorKeys(){

            var item;
            var invalid = false;
            for (var i = 0; i < dsPreds.source.length; i++) {
                item = dsPreds.source[i];

                if(item.CrossProject !== 1) {
                    if (!findTaskByKey(item.TaskKey) || !findTaskByKey(item.PredecessorKey)) {
                        removePredecessorFromSource(item);
                    }
                }
            }
        }

        /**
         * Adds to the array of general errors (not row specific errors)
         * @param msg
         */
        function addGeneralError(msg) {
            if (!errors) errors = [];
            errors.push(System.localize(msg));
        }

        /**
         * Returns true if the general errors array contains messages
         * @returns {*|boolean}
         */
        function hasGeneralErrors() {
            return errors && errors.length > 0;
        }

        /**
         * Returns the array of general errors, and clears the array for the next schedule manager operation
         * @returns {Array}
         */
        function getGeneralErrors() {
            var rtn = errors || [];
            errors = []; //clear out the errors array for next time
            return rtn;
        }

        /**
         * Adds to the array of general notifications
         * @param msg
         */
        function addGeneralNotification(msg) {
            if (!notifications) notifications = [];
            notifications.push({msg: msg});
        }

        /**
         * Returns true if the general notifications array contains messages
         * @returns {*|boolean}
         */
        function hasGeneralNotifications() {
            return notifications && notifications.length > 0;
        }

        /**
         * Returns the array of general notifications, and clears the array for the next schedule manager operation
         * @returns {Array}
         */
        function getGeneralNotifications() {
            var rtn = notifications || [];
            notifications = []; //clear out the notifications array for next time
            return rtn;
        }

        /**
         * Searches the task datasource for the record by key. If no keyField is specified, TaskKey is searched
         * @param keyValue The value to search for
         * @param keyField Optional. The field name to search, default is TaskKey
         * @returns {*|{}|Array}
         */
        function findTaskByKey(keyValue, keyField) {
            keyField = keyField || 'TaskKey';
            return dsTasks.findByKey(u.obn(keyValue), keyField);
        }

        /**
         * Returns the parent task of the task that is passed in
         * @param task This is a task object (NOT THE TaskKey!! Use getParentByKey below if you only have a TaskKey)
         * @returns {Object}
         */
        function getParentTask(task) {
            return DataManager.getParent(dsTasks, task);
        }

        /**
         * Returns an array of children for the given task
         * @param task
         * @returns {*}
         */
        function getChildren(task){
            return DataManager.getChildren(dsTasks, task, true);
        }

        /**
         * Returns true if the task has children
         * @param task
         * @returns {boolean}
         */
        function hasChildren(task){
            return getChildren(task).length > 0;
        }

        //Be careful if copying from Flex!!!!
        //In Flex, getParentByKey just returns the node of the key passed in, not the parent of it!!!!
        //They make it work by passing in the SummaryTaskKey of the child task
        /**
         * Returns the parent task of the taskKey that is passed in
         * @param taskKey
         * @returns {Object}
         */
        function getParentByKey(taskKey) {
            var task = findTaskByKey(taskKey);
            if (task) return getParentTask(task);
        }

        /**
         * Travels up the tree looking for the budget task
         * @param task
         * @returns {*}
         */
        function findTrackBudgetInParents(task) {

            if (task) {
                if (task.BudgetTaskKey) {
                    return true;
                } else {
                    return findTrackBudgetInParents(getParentTask(task));
                }
            }
            return false;
        }

        /**
         * Travels down the tree looking for the budget task
         * @param children
         * @returns {boolean}
         */
        function findTrackBudgetInChildren(children) {
            if (children) {
                var childTask;
                for (var i = 0; i < children.length; i++) {
                    if (children[i].MoneyTask === 1) return true;
                }
            }
            return false;
        }

        /**
         * Sets the TaskConstraint and its label
         * @param task
         * @param constraint
         */
        function setConstraint(task, constraint) {
            task.TaskConstraint = constraint

            for (var i = 0; i < parms.constraintListWithDuration.length; i++) {
                if (parms.constraintListWithDuration[i].k === constraint) {
                    task.ConstraintName = parms.constraintListWithDuration[i].l;
                    break;
                }
            }
        }

        var taskEffPreds = {};
        var taskPreds = {};
        var predKeyList = {}; //associative array of PredecessorKeys, just for quick lookup

        /**
         * Rolls through the predecessors and sets up the internal predecessor pointer objects
         */
        function loadPredecessors() {

            taskEffPreds = {};
            taskPreds = {};
            predKeyList = {};

            var task;
            var preds;
            var predecessorKey;
            for (var i = 0; i < dsTasks.source.length; i++) {
                task = dsTasks.source[i];
                if (task.ScheduleTask === 1) {
                    preds = lookupPredecessors(task.TaskKey);

                    for (var j = 0; j < preds.length; j++) {
                        predecessorKey = preds[j].PredecessorKey;
                        if (preds[j].isInvalid) {
                            //ignore invalid preds....these get cleaned up or should have been cleaned up
                        } else {
                            // Load the effective or real predecessors
                            addTaskEffectivePredecessor(task, predecessorKey, preds[j].Lag, preds[j].InWeeks);
                            if (project.ScheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                                // Now load the logical predecessors
                                addTaskPredecessor(task, predecessorKey, preds[j].Lag, preds[j].InWeeks, preds[j].CrossProject);
                            }else{
                                // Successors should be scheduled first
                                addTaskPredecessor(findTaskByKey(predecessorKey), task.TaskKey, preds[j].Lag, preds[j].InWeeks, preds[j].CrossProject);
                            }
                        }
                    }
                }
            }
        }

        /**
         * Returns an array of predecessors of the given task
         * @param taskKey
         */
        function lookupPredecessors(taskKey) {
            return predList.filter(function (item) {
                return item.TaskKey === taskKey;
            });
        }

        /**
         * Returns an array of successors of the given task
         * @param taskKey
         */
        function lookupSuccessors(taskKey) {
            return predList.filter(function (item) {
                return item.PredecessorKey === taskKey;
            });
        }

        function getEffPreds(task, createIfNotFound) {
            var preds = taskEffPreds[task.TaskKey];
            if (!preds) {
                return createIfNotFound ? (taskEffPreds[task.TaskKey] = []) : null;
            }
            return preds;
        }

        /**
         * Returns the predecessors of a task
         * @param task
         * @param createIfNotFound If true, it will return an empty array if no preds are found
         * @param reloadPreds If true, it will reload the predecessor objects first (this is for a new row that hasn't been edited yet, it's not in the source until it's been edited)
         * @param ignoreEffPreds This function is used for two purposes. If true, it always returns preds from the taskPreds object. If false, and the direction is not FROM_PROJECT_START, it will return taskEffPreds
         * @returns {*}
         */
        function getPreds(task, createIfNotFound, reloadPreds, ignoreEffPreds) {

            if(reloadPreds) loadPredecessors();

            var preds = ignoreEffPreds ? taskPreds[task.TaskKey] : (project.ScheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) ? taskPreds[task.TaskKey] : taskEffPreds[task.TaskKey];
            if (!preds) {
                return createIfNotFound ? (taskPreds[task.TaskKey] = []) : null;
            }
            return preds;
        }

        function addTaskEffectivePredecessor(task, newPredKey, newLag, inWeeks) {

            predKeyList[newPredKey] = true;

            var preds = getEffPreds(task, true);
            for (var i = 0; i < preds.length; i++) {
                if (preds[i].PredKey === newPredKey) {
                    if (newLag > preds[i].Lag) preds[i].Lag = newLag;
                    return;
                }
            }
            preds.push({PredKey: newPredKey, Lag: newLag, InWeeks: inWeeks});
        }

        function addTaskPredecessor(task, newPredKey, newLag, inWeeks, crossProject) {
            if (!task) return;

            var preds = getPreds(task, true, false, true);
            for (var i = 0; i < preds.length; i++) {
                if (preds[i].PredKey === newPredKey) {
                    if (newLag > preds[i].Lag) preds[i].Lag = newLag;
                    return;
                }
            }
            preds.push({PredKey: newPredKey, Lag: newLag, InWeeks: inWeeks, CrossProject: crossProject});
        }

        /**
         * Removes predecessors and successors of the given task
         * @param task
         */
        function deleteTaskPredecessors(task) {

            var removedItems = dsPreds.removeByKey(task.TaskKey, 'TaskKey');
            if (removedItems) {
                if (Array.isArray(removedItems)) {
                    deletedPreds = deletedPreds.concat(removedItems);
                } else {
                    deletedPreds.push(removedItems);
                }
            }

            removedItems = dsPreds.removeByKey(task.TaskKey, 'PredecessorKey');
            if (removedItems) {
                if (Array.isArray(removedItems)) {
                    deletedPreds = deletedPreds.concat(removedItems);
                } else {
                    deletedPreds.push(removedItems);
                }
            }

            //predList should still be pointing to dsPreds.source, so it should reflect the removals from above...hopefully
            loadPredecessors();
        }

        /**
         * Removes all taskUser assignments for the given task
         * @param task
         */
        function deleteTaskUsers(task) {

            var removedItems = dsTaskUsers.removeByKey(task.TaskKey, 'TaskKey');
            if (removedItems) {
                if (Array.isArray(removedItems)) {
                    for (var i = 0; i < removedItems.length; i++) {
                        removeTaskUser(removedItems[i]);
                    }
                } else {
                    removeTaskUser(removedItems);
                }
            }
        }

        /**
         * Returns true if the task has successors. This is an extra function that is here to be consistent with the terminology
         * @param task
         * @returns {*}
         */
        function hasSuccessors(task) {
            return isPredecessor(task);
        }

        /**
         * Returns true if the task is a predecessor of another task
         * @param task
         * @returns {boolean}
         */
        function isPredecessor(task) {
            if (predKeyList[task.TaskKey]) {
                return true;
            } else {
                return false;
            }
        }

        /**
         * Returns true if the task has predecessors
         * @param task
         * @returns {boolean}
         */
        function hasPredecessors(task) {
            if (getPreds(task)) {
                return true;
            } else {
                return false;
            }
        }

        var idleDay = {};

        function initBlockoutDays() {

            var item;
            for (var i = 0; i < blockoutDays.length; i++) {
                item = blockoutDays[i];
                if (item.UserKey === -1) {
                    setIdleDay(item.CalendarDate);
                } else if (project.OfficeKey > 0 && item.UserKey === -2) {
                    //Office specific calendar
                    if(item.OfficeKey === project.OfficeKey){
                        setIdleDay(item.CalendarDate);
                    }
                } else {
                    setIdleDay(item.CalendarDate, u.obn(item.UserKey));
                }
            }
        }

        function setIdleDay(calendarDate, userKey) {

            userKey = userKey || '-1';
            userKey = userKey.toString();
            if (!idleDay[calendarDate]) idleDay[calendarDate] = {};
            idleDay[calendarDate][userKey.toString()] = true;
        }

        function isWorkDay(dte, includeCompany, includeProject) {

            //copied from existing CMP schedule edit screen...

            var isCompanyWorkDay = true;
            var isProjectWorkDay = true;

            if (!dte) {
                return false;
            }
            // convert to the way it is indexed or stored in the hash table
            var d = Date.fromISOString(dte);
            var year = d.getFullYear().toString();
            var month = (d.getMonth() + 1).toString();
            var dt = d.getDate().toString();
            if (dt < 10) {
                dt = '0' + dt;
            }
            if (month < 10) {
                month = '0' + month;
            }
            var storedDate = year + month + dt;

            if (includeCompany && idleDay && idleDay[storedDate] && idleDay[storedDate]['-1']) {
                isCompanyWorkDay = false;
            }

            if (includeProject) {
                isProjectWorkDay = projectWorkDay(d.getDay() + 1);
            }

            // return a logical AND of all flags
            return isCompanyWorkDay && isProjectWorkDay;
        }

        function findNextWorkDate(dte, direction) {
            //add does a copy as well
            var dteCurDate = Date.add(dte, direction, Date.DAY);

            while (!(isWorkDay(dteCurDate, true, true))) {
                dteCurDate = Date.add(dteCurDate, direction, Date.DAY);
            }
            return dteCurDate;
        }

        function getNextWorkDate() {
            var newDate;
            if (project.ScheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                //from the project start
                newDate = Date.add(Date.ISOToday, -1, Date.DAY);
                newDate = findNextWorkDate(newDate, Enum.ProjectScheduleDirection.DIRECTION_FWD);
                if (newDate < project.StartDate) {
                    //new date cannot be before the start date
                    newDate = project.StartDate;
                }
            } else {
                //from the project complete
                newDate = project.CompleteDate;
            }
            return newDate;
        }

        function projectWorkDay(day) {
            switch (day) {
                case 1:
                    return project.WorkSun === 1;
                case 2:
                    return project.WorkMon === 1;
                case 3:
                    return project.WorkTue === 1;
                case 4:
                    return project.WorkWed === 1;
                case 5:
                    return project.WorkThur === 1;
                case 6:
                    return project.WorkFri === 1;
                case 7:
                    return project.WorkSat === 1;
                default:
                    return true;
            }
        }

        function calcNextDate(dur, dteDate, direction, workAnyDay) {
            //iDirection = 1 Forward, -1 Backward
            if (!dteDate) {
                return Date.ISOToday;
            }
            if (dur === 0) {
                // This is a milestone, stay on same day
                return dteDate;
            }

            if (dur === 1) {
                // This is a 1 day task, stay on same day
                return dteDate;
            }
            //change the duration into number of days we want to move by
            return moveDate(dteDate, dur - 1, direction, workAnyDay);
        }

        function moveDate(dteDate, moveBy, direction, workAnyDay, moveByWeeks) {
            if (moveBy === 0) {
                return dteDate;
            } else {
                if (workAnyDay) {
                    if (moveByWeeks) {
                        return Date.add(dteDate, (moveBy * direction) * 7, Date.DAY);
                    } else {
                        return Date.add(dteDate, moveBy * direction, Date.DAY);
                    }
                } else {
                    var dteNewDate = dteDate;
                    var currMoved = 0;
                    while (currMoved < moveBy) {
                        dteNewDate = Date.add(dteNewDate, direction, Date.DAY);
                        if (isWorkDay(dteNewDate, true, true)) {
                            currMoved++;
                        }
                    }
                    return dteNewDate;
                }
            }
        }

        function calcDuration(dteStart, dteEnd, origDur, workAnyDay) {
            // When task is a milestone, original duration = 0, return 0
            if ((origDur === 0) && (dteStart === dteEnd)) {
                return 0;
            }
            if (workAnyDay) {
                return Date.diff(dteStart, dteEnd) + 1;
            } else {
                var dur = Date.diff(dteStart, dteEnd) + 1;
                // Beware of the calendar control, it returns dates which contain hours
                // This creates problems with DateUtils.diff so we clear timmmme
                var dteStart2 = dteStart;
                var dayCount = 0;
                for (var i = 0; i < dur; i++) {
                    if (isWorkDay(dteStart2, true, true, false)) {
                        dayCount++;
                    }
                    dteStart2 = Date.add(dteStart2, 1, Date.DAY);
                }
                return dayCount;
            }
        }

        function calcMaxProjectDate(findMax) {

            findMax = (typeof findMax !== 'undefined') ? findMax : true;

            var dateTemp;
            var allDurationOfProjectTasks = true;
            if (findMax) {
                dateTemp = new Date(1901, 0, 1).toISOString();
            } else {
                dateTemp = new Date(2079, 0, 1).toISOString();
            }

            var task;
            for (var i = 0; i < dsTasks.source.length; i++) {
                task = dsTasks.source[i];
                if (task.TaskConstraint !== Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                    if (allDurationOfProjectTasks) {
                        allDurationOfProjectTasks = false;
                    }
                    if (findMax) {
                        if (task.PlanComplete > dateTemp) {
                            dateTemp = task.PlanComplete;
                        }
                    } else {
                        if (task.PlanStart < dateTemp) {
                            dateTemp = task.PlanStart;
                        }
                    }
                }
            }
            if (allDurationOfProjectTasks === true) {
                if (findMax) {
                    dateTemp = project.CompleteDate || project.StartDate;
                } else {
                    dateTemp = project.StartDate || project.CompleteDate;
                }
            }
            return dateTemp

        }

        function checkTrackBudgetState(task, iTrackBudget, parentTask) {

            parentTask = parentTask || getParentByKey(task.TaskKey);
            if (iTrackBudget === 1) {
                if (findTrackBudgetInParents(parentTask)) {
                    return false;
                }
                if (hasChildren(task)) {
                    if (findTrackBudgetInChildren(getChildren(task))) {
                        return false;
                    }
                }
            } else {
                if (findTrackBudgetInParents(parentTask)) {
                    if (hasChildren(task)) {
                        if (findTrackBudgetInChildren(getChildren(task))) {
                            return false;
                        }
                    }
                }
            }
            return true;
        }

        function setTrackBudgetFlags(task, trackBudget, parentTask) {

            parentTask = parentTask || getParentTask(task);

            if (trackBudget === 1) {
                //GG says..
                task.TrackBudget = 1;
                task.MoneyTask = 1;
                task.BudgetTaskKey = task.TaskKey;
                //set BTK down MT up
                setBudgetTaskKeyDown(task, task.TaskKey);
                setMoneyTaskFlagUp(parentTask);
            } else {
                //GG says..
                task.TrackBudget = 0;
                task.MoneyTask = 0;
                task.BudgetTaskKey = 0;
                //clear BTK down and check MT up in case it can be cleared
                setBudgetTaskKeyDown(task, 0);
                checkMoneyTaskFlag(parentTask);
            }
        }

        function setBudgetTaskKeyDown(task, budgetTaskKey, childrenOnly) {

            //default optional parm to true
            childrenOnly = (typeof childrenOnly !== 'undefined') ? childrenOnly : true;

            //set BTK on the Task
            if (!childrenOnly) task.BudgetTaskKey = budgetTaskKey;
            //check children
            if (task.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                if (hasChildren(task)) {
                    var children = getChildren(task);
                    var childTask;
                    for (var i = 0; i < children.length; i++) {
                        childTask = children[i];
                        childTask.BudgetTaskKey = budgetTaskKey;
                        setBudgetTaskKeyDown(childTask, budgetTaskKey);
                    }
                }
            }
        }

        function setMoneyTaskFlagUp(task) {
            if (task) {
                if (task.MoneyTask !== 1) {
                    task.MoneyTask = 1;
                    if (task.SummaryTaskKey) {
                        setMoneyTaskFlagUp(getParentTask(task));
                    }
                }
            }
        }

        function checkMoneyTaskFlag(task, flag) {
            //task can be null if we hit root because a root object doesnt exist
            //at this point we have already checked for illegal children
            if (task) {
                if (task.MoneyTask === 1 && task.TrackBudget === 1) {
                    //skip any check, this task is already valid
                    return;
                }
                var children = getChildren(task);
                for (var i = 0; i < children.length; i++) {
                    if (children[i].MoneyTask === 1) {
                        flag = 1;
                        break;
                    }
                }
                if (flag === 1) {
                    if (task.MoneyTask === 0) {
                        task.MoneyTask = 1;
                        //set parent task
                        setMoneyTaskFlagUp(getParentTask(task));
                    }
                } else {
                    //if the MoneyTask is true we switch it to false but only if TrackBudget
                    if (task.MoneyTask === 1) {
                        task.MoneyTask = 0;
                        //check parent task
                        checkMoneyTaskFlag(getParentTask(task));
                    }
                }
            }
        }

        function pushDownActuals(ci, source) {

            var users = lookupAssignedUsers(ci.TaskKey);
            for (var i = 0; i < users.length; i++) {
                if (ci.PercComp === 100) {
                    users[i].PercComp = ci.PercComp;
                    users[i].ActStart = users[i].ActStart ? users[i].ActStart : ci.ActStart;
                    users[i].ActComplete = users[i].ActComplete ? users[i].ActComplete : ci.ActComplete;
                } else {
                    if (source === "PercComp") {
                        users[i].PercComp = ci.PercComp;
                        if (users[i].ActComplete !== null) {
                            users[i].ActComplete = null;
                        }
                    }
                    if (source === "ActStart") {
                        users[i].ActStart = ci.ActStart;
                        if (!ci.ActStart) {
                            if (users[i].ActComplete) {
                                users[i].ActComplete = ci.ActComplete;
                                users[i].PercComp = ci.PercComp;
                            }
                        }
                    }
                    if (source === "ActComplete") {
                        users[i].ActComplete = ci.ActComplete;
                        users[i].PercComp = ci.PercComp;
                    }
                }
            }
        }

        function lookupAssignedUsers(taskKey) {
            return taskUsers.filter(function (item) {
                return (item.TaskKey === taskKey);
            });
        }

        /**
         * Primary function to cause the scheduling logic to run
         */
        function scheduleProject() {

            invalidSchedule = false;

            //create a new empty errors array if it's undefined or if its length > 0
            //if we already have an empty errors array, no need to create a new one
            if (!errors || errors.length > 0) errors = [];

            if (parms.isTaskOnly()) {
                return;
            }

            loadPredecessors();
            schedulingPushPredecessorsDown();
            schedulingResetScheduleFlags();

            var tasks = dsTasks.source;
            var projectConstraintTaskArray = [];
            for (var i = 0; i < tasks.length; i++) {
                if (tasks[i].ScheduleTask === 1) {
                    taskProcessedInScheduling = {};
                    if (tasks[i].TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                        projectConstraintTaskArray.push(tasks[i]);
                    } else {
                        //update non project constraint task rows
                        schedulingSetTaskDates(tasks[i]);
                    }
                }
            }
            if (projectConstraintTaskArray.length > 0) {
                for (var j = 0; j < projectConstraintTaskArray.length; j++) {
                    //update project constraint task rows to correctly calculate plan start and end
                    schedulingSetTaskDates(projectConstraintTaskArray[j]);
                }
            }

            validateAllBuckets();

            //update task rows predComplete flag
            schedulingSetPredecessorsComplete();

            if (setNewTaskID) {
                autoNumberTasks();
            }
           
            formatPredValues();

            //update the projectData
            schedulingUpdateProject();
        }

        /**
         * Returns an array of the top level tasks
         * @returns {Array}
         */
        function getScheduleRoot() {

            var rtn = [];
            for (var i = 0; i < dsTasks.length; i++) {
                rtn.push(dsTasks[i]);
            }
            return rtn;
        }

        function schedulingPushPredecessorsDown(effectivePredecessors) {

            var nodes = getScheduleRoot();
            if (effectivePredecessors) {
                for (var i = 0; i < nodes.length; i++) if (nodes[i].ScheduleTask === 1) schedulingPushEffectivePredecessorDown(nodes[i]);
            } else {
                for (var i = 0; i < nodes.length; i++) if (nodes[i].ScheduleTask === 1) schedulingPushPredecessorDown(nodes[i]);
            }
        }

        function schedulingPushPredecessorDown(parentTask, effectivePredecessors) {

            if (parentTask.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                var children = getChildren(parentTask);
                if (children) {
                    if (effectivePredecessors) {
                        var effPreds = getEffPreds(parentTask, true);
                        if (effPreds) {
                            for (var i = 0; i < children.length; i++) {
                                childTask = children[i];
                                for (var j = 0; j < effPreds.length; j++) {
                                    // Push this predecessor to the child
                                    // AddPredecessor will take the greatest of the lags if same predecessor
                                    addTaskPredecessor(childTask, effPreds[j].PredKey, effPreds[j].Lag, effPreds[j].InWeeks, effPreds[j].CrossProject);
                                }
                                schedulingPushPredecessorDown(childTask, true);
                            }
                        }
                    } else {
                        var childTask;
                        var preds = getPreds(parentTask, true, false, true);
                        if (preds) {
                            for (var i = 0; i < children.length; i++) {
                                childTask = children[i];
                                for (var j = 0; j < preds.length; j++) {
                                    // Push this predecessor to the child
                                    // AddPredecessor will take the greatest of the lags if same predecessor
                                    addTaskPredecessor(childTask, preds[j].PredKey, preds[j].Lag, preds[j].InWeeks, preds[j].CrossProject);
                                }
                                schedulingPushPredecessorDown(childTask);
                            }
                        }
                    }

                }
            }
        }

        function schedulingPushEffectivePredecessorDown(parentTask) {

            if (parentTask.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                var childTask;
                var children = getChildren(parentTask);
                if (!children) return;
                var effPreds = getEffPreds(parentTask, true);
                if (!effPreds) return;
                for (var i = 0; i < children.length; i++) {
                    childTask = children[i];
                    for (var j = 0; j < effPreds.length; j++) {
                        // Push this predecessor to the child
                        // AddPredecessor will take the greatest of the lags if same predecessor
                        addTaskEffectivePredecessor(childTask, effPreds[j].PredKey, effPreds[j].Lag, effPreds[j].InWeeks);
                    }
                    schedulingPushEffectivePredecessorDown(childTask);
                }
            }
        }

        function schedulingResetScheduleFlags() {
            var task;
            for (var i = 0; i < dsTasks.source.length; i++) {
                task = dsTasks.source[i];
                task.scheduled = false;
                task.scheduleOrder = 0;
            }
            scheduleOrder = 0;
        }

        function schedulingSetTaskDates(task, successorTask) {

            //if already scheduled don't bother doing it again
            if (task.scheduled) return;

            TaskUserHour.setStatus(getTaskParms(task), TaskUserHour.STATUS_RECALC_REQUIRED);

            var sKey;
            var sTaskName;
            var maxDate;
            var maxEndDate;
            var sPlanStart;// = new Date();
            var sPlanComplete;// = new Date();
            var sPredPlanDate;// = new Date();
            var bWorkAnyDay;
            var lDuration;
            var PredKey;
            var Lag;
            var lPredDir;
            var childKey;
            var childTask;
            var predTask;
            var bPushToToday;
            var today = Date.ISOToday;
            if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                maxDate = project.StartDate;
                maxEndDate = task.PlanComplete;

                bPushToToday = false;

                // General case we will push if they requested it
                if (autoScheduleProjects && maxDate < today) {
                    bPushToToday = true;
                }

                // Special case when actual dates do not update plan and there are actuals
                // Nerland requested to stop pushing otherwise it would continue bumping for ever
                // Nerland configuration:
                // autoScheduleProjects = true
                // actualDateUpdatesPlan = false
                // This concerns 4% of companies only

                if (!actualDateUpdatesPlan && autoScheduleProjects) {
                    if (task.ActStart) {
                        bPushToToday = false;
                        // Keep it as the same plan start
                        maxDate = task.PlanStart;
                    }
                }
                // new option to push end dates to todays date if there is an actual start and end date is before today
                if (autoScheduleTaskEndDate && maxEndDate < today) {
                    if (task.ActStart && !task.ActComplete) {
                        maxEndDate = Date.add(today, -1, Date.DAY);
                        maxEndDate = moveDate(maxEndDate, 1, 1, task.WorkAnyDay);
                        task.PlanComplete = maxEndDate;
                        task.PlanDuration = calcDuration(task.ActStart, task.PlanComplete, task.PlanDuration, task.WorkAnyDay);

                    }
                }

                if (bPushToToday) {
                    maxDate = Date.add(today, -1, Date.DAY);
                    maxDate = moveDate(maxDate, 1, 1, task.WorkAnyDay);
                }
            } else {
                maxDate = project.CompleteDate;
            }

            sKey = task.TaskKey;
            sTaskName = task.TaskName;

            // To fight circular relationships
            if (taskProcessedInScheduling[sKey]) {
                //remove the task, and warn the user
                if(successorTask) {
                    if (unlinkFromTask(successorTask, task, true)) {
                        addGeneralNotification('|26674|A Circular Relationship found with task ' + sTaskName + ' has been removed.|');
                    }
                    return;
                }
            }
            taskProcessedInScheduling[sKey] = 1;
            var alPreds = getPreds(task, false, false, true);
            if (alPreds) {
                for (var i = 0; i <= alPreds.length - 1; i++) {
                    PredKey = alPreds[i].PredKey;
                    predTask = findTaskByKey(PredKey);
                    if (predTask) {
                        if (!(predTask.scheduled)) {
                            schedulingSetTaskDates(predTask, task);
                        }
                    }
                }
            }
            if (task.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                var children = getChildren(task);
                if (children.length > 0) {
                    var i = 0;
                    while (children.length > 0 && i < children.length) {
                        childTask = children[i];
                        childKey = childTask.TaskKey;
                        if ((!childTask.scheduled) && (childTask.ScheduleTask === 1)) {
                            schedulingSetTaskDates(childTask);
                        }

                        i++;

                        //after incrementing i, if i has become greater than the length of the array (due to items being deleted by other processes),
                        // set i to the first child that has not been scheduled
                        if (i >= children.length) {
                            for (var j = 0; j < children.length; j++) {
                                if (!children[j].scheduled) {
                                    i = j;
                                    break;
                                }
                            }
                            //if all children have been scheduled, set i to the length of the array which will end the while loop
                            i = children.length;
                        }
                        //if there are no task children, or i has exceeded the length of the array, then the while loop will end
                    }

                    sPlanStart = new Date(2050, 1, 1);
                    sPlanComplete = new Date(1975, 1, 1);
                    bWorkAnyDay = true;
                    lDuration = 0;
                    for (var i = 0; i <= children.length - 1; i++) {
                        childTask = children[i];
                        childKey = childTask.TaskKey;
                        if ((childTask.ScheduleTask === 1)) {
                            sPlanStart = Date.min(childTask.PlanStart, sPlanStart);
                            sPlanComplete = Date.max(childTask.PlanComplete, sPlanComplete);
                            bWorkAnyDay = bWorkAnyDay && childTask.WorkAnyDay;
                            if (childTask.PlanDuration > 0) {
                                lDuration = childTask.PlanDuration; //I dont get this????why set it to anyhting since dur is calc'd or if not wy not eh max this is just the "last"
                            }
                        }
                    }
                    task.WorkAnyDay = bWorkAnyDay ? 1 : 0;

                    task.PlanDuration = lDuration;
                    schedulingScheduleTask2(maxDate, sPlanStart, sPlanComplete, task);
                    schedulingSetPercComp(task);
                } else {
                    schedulingScheduleTask3(maxDate, maxDate, maxDate, 0, task);
                }
            } else {
                var alPreds = getPreds(task, false, false, true);
                var setWeek = false;
                if (alPreds) {
                    for (var i = 0; i <= alPreds.length - 1; i++) {
                        var useWeeks = false;
                        PredKey = alPreds[i].PredKey;
                        Lag = alPreds[i].Lag;
                        if (u.obn(alPreds[i].InWeeks) === 1) {
                            useWeeks = true;
                            if (!setWeek) {
                                setWeek = true;
                            }
                        }
                        if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START || alPreds[i].CrossProject === 1) {
                            if (Lag < 0) {
                                lPredDir = -1;
                            } else {
                                lPredDir = 1;
                            }
                            Lag = Math.abs(Lag);

                            if(alPreds[i].CrossProject === 1){
                                predTask = crossProjectPreds[PredKey];
                            }else{
                                predTask = findTaskByKey(PredKey);
                            }

                            if (predTask) {
                                sPredPlanDate = predTask.PlanComplete;
                                if (useWeeks) {
                                    sPredPlanDate = moveDate(sPredPlanDate, Lag, lPredDir, true, true);
                                } else {
                                    sPredPlanDate = moveDate(sPredPlanDate, Lag, lPredDir, task.WorkAnyDay);
                                }

                                if (predTask.PlanDuration > 0) {
                                    if (useWeeks) {
                                        sPredPlanDate = moveDate(sPredPlanDate, 1, 1, true);
                                    } else {
                                        sPredPlanDate = moveDate(sPredPlanDate, 1, 1, task.WorkAnyDay);
                                    }

                                }
                                if (sPredPlanDate > maxDate) {
                                    maxDate = sPredPlanDate;
                                }
                            }
                        } else {
                            if (Lag >= 0) {
                                lPredDir = -1;
                            } else {
                                lPredDir = 1;
                            }
                            Lag = Math.abs(Lag);
                            predTask = findTaskByKey(PredKey)
                            sPredPlanDate = predTask.PlanStart;
                            if (useWeeks) {
                                sPredPlanDate = moveDate(sPredPlanDate, Lag, lPredDir, true, true);
                            } else {
                                sPredPlanDate = moveDate(sPredPlanDate, Lag, lPredDir, task.WorkAnyDay);
                            }

                            if (predTask.PlanDuration > 0) {
                                if (useWeeks) {
                                    sPredPlanDate = moveDate(sPredPlanDate, 1, -1, true);
                                } else {
                                    sPredPlanDate = moveDate(sPredPlanDate, 1, -1, task.WorkAnyDay);
                                }

                            }
                            if (sPredPlanDate < maxDate) {
                                maxDate = sPredPlanDate;
                            }
                        }

                    }
                }
                schedulingScheduleTask(maxDate, task, setWeek);
            }
        }

        /**
         * Primary function for calculating the schedule for a task
         * @param maxDate
         * @param task
         * @param setInWeek True if this task's predecessor has lag defined in weeks
         */
        function schedulingScheduleTask(maxDate, task, setInWeek) {

            var bActStart = true;
            var bActComplete = true;

            if (!task.ActStart) {
                bActStart = false;
            }
            if (!(actualDateUpdatesPlan)) {
                bActStart = false;
            }
            if (!task.ActComplete) {
                bActComplete = false;
            }
            if (!(actualDateUpdatesPlan)) {
                bActComplete = false;
            }
            var sNewStart;// = new Date();
            var sNewComplete;// = new Date();
            //var sConstraint:String = UtilNamedConstraint(task.TaskConstraint);
            var taskConstraint = task.TaskConstraint;
            var sConstraintDate = task.ConstraintDate;
            var lDuration = task.PlanDuration;
            var bWorkAnyDay = task.WorkAnyDay;
            var bSetDayOfWeek = false;
            var constraintDayOfTheWeek = u.obn(task.ConstraintDayOfTheWeek);
            var nextDayOfTheWeek;
            var maxDateWKSd;
            if (constraintDayOfTheWeek > 0) {
                bSetDayOfWeek = true;
                nextDayOfTheWeek = maxDate;

                if (setInWeek) {
                    // this should be the week we want now just need to find the right day to set next day of the wk to
                    nextDayOfTheWeek = Date.getStartOfWeek(nextDayOfTheWeek);
                    // getUTCDay is sunday through sat Constraint day is monday through sunday start with 1
                    var maxWkSDDay = Date.getDayOfWeek(nextDayOfTheWeek);
                    var moveDayTo = constraintDayOfTheWeek - 1;

                    while (maxWkSDDay !== moveDayTo) {
                        nextDayOfTheWeek = Date.add(nextDayOfTheWeek, 1, Date.DAY);
                        maxWkSDDay = Date.getDayOfWeek(nextDayOfTheWeek);
                    }

                    if (!bWorkAnyDay) {
                        // changed to only checking for company holiday if so go back a wk
                        while (!isWorkDay(nextDayOfTheWeek, true, false, false)) {
                            nextDayOfTheWeek = Date.add(nextDayOfTheWeek, -7, Date.DAY);
                        }
                    }
                } else {
                    while (Date.getDayOfWeek(nextDayOfTheWeek) !== (constraintDayOfTheWeek - 1)) {
                        nextDayOfTheWeek = Date.add(nextDayOfTheWeek, 1, Date.DAY);
                    }

                    // now check if day selected is a workday
                    if (!bWorkAnyDay) {
                        while (!isWorkDay(nextDayOfTheWeek, true, true, false)) {
                            nextDayOfTheWeek = Date.add(nextDayOfTheWeek, 7, Date.DAY);
                        }
                    }
                }
            }

            if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                switch (taskConstraint) {
                    case Enum.ProjectScheduleConstraintTypes.AS_SOON_AS_POSSIBLE://"As Soon As Possible":
                        if (bSetDayOfWeek) {
                            maxDate = nextDayOfTheWeek;
                        }
                        sConstraintDate = maxDate;
                        break;
                    case Enum.ProjectScheduleConstraintTypes.MUST_START_ON://"Must Start On":
                    case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN://"Start No Earlier Than":
                    case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN://"Start No Later Than":
                        if (!sConstraintDate) {
                            sConstraintDate = maxDate;
                        } else if (sConstraintDate < project.StartDate) {
                            sConstraintDate = maxDate;
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON://"Must Finish On":
                    case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN://"Finish No Earlier Than":
                    case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN://"Finish No Later Than":
                        if (!sConstraintDate) {
                            sConstraintDate = maxDate;
                        } else if (sConstraintDate < calcNextDate(lDuration, project.StartDate, Enum.ProjectScheduleDirection.DIRECTION_FWD, bWorkAnyDay)) {
                            // This was causing problems at Franklin Covey, so I removed it
                            // ProjectStartDate = 2/13/2007, PlanComplete = 3/1/2007, PlanStart = 2/13/2007
                            // maybe a border stituation
                            //sConstraintDate = calcNextDate(lDuration, this.projectStartDate, DIRECTION_FWD, bWorkAnyDay);
                        }
                        break;
                }
            } else {
                switch (taskConstraint) {
                    case Enum.ProjectScheduleConstraintTypes.AS_LATE_AS_POSSIBLE://"As Late As Possible":
                        sConstraintDate = maxDate;
                        break;
                    case Enum.ProjectScheduleConstraintTypes.MUST_START_ON://"Must Start On":
                    case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN://"Start No Earlier Than":
                    case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN://"Start No Later Than":
                        if (!sConstraintDate) {
                            sConstraintDate = maxDate;
                        } else if (sConstraintDate > calcNextDate(lDuration, project.CompleteDate, Enum.ProjectScheduleDirection.DIRECTION_BWD, bWorkAnyDay)) {
                            sConstraintDate = maxDate;
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON://"Must Finish On":
                    case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN://"Finish No Earlier Than":
                    case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN://"Finish No Later Than":
                        if (!sConstraintDate) {
                            sConstraintDate = maxDate;
                        } else if (sConstraintDate > project.CompleteDate) {
                            sConstraintDate = maxDate;
                        }
                        break;
                }
            }
            switch (taskConstraint) {
                //case ProjectConstants.AS_SOON_AS_POSSIBLE://"As Soon As Possible":
                //case ProjectConstants.AS_LATE_AS_POSSIBLE://"As Late As Possible":
                //sNewStart = DateUtils.copyDate(maxDate);
                //sNewComplete = DateUtils.copyDate(maxDate);
                //break;
                case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        sNewStart = maxDate;
                    } else {
                        sNewComplete = maxDate;
                    }
                    break;
                case Enum.ProjectScheduleConstraintTypes.MUST_START_ON://"Must Start On":
                    sNewStart = sConstraintDate;
                    break;
                case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN://"Start No Earlier Than":
                    sNewStart = Date.max(sConstraintDate, maxDate);
                    break;
                case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN://"Start No Later Than":
                    sNewStart = Date.min(sConstraintDate, maxDate);
                    break;
                case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON://"Must Finish On":
                    sNewComplete = sConstraintDate;
                    break;
                case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN://"Finish No Earlier Than":
                    if (bActStart) {
                        var minDuration;
                        if (task.ActStart > sConstraintDate) {
                            minDuration = calcDuration(task.ActStart, maxDate, lDuration, bWorkAnyDay);
                        } else {
                            if (sConstraintDate > maxDate) {
                                minDuration = calcDuration(task.ActStart, sConstraintDate, lDuration, bWorkAnyDay);
                            } else {
                                minDuration = calcDuration(task.ActStart, maxDate, lDuration, bWorkAnyDay);
                            }
                        }

                        if (lDuration > minDuration) {
                            sNewComplete = calcNextDate(lDuration, task.ActStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, bWorkAnyDay);
                        } else {
                            sNewComplete = calcNextDate(minDuration, task.ActStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, bWorkAnyDay);
                        }
                        sNewComplete = Date.max(sConstraintDate, sNewComplete);
                    } else {
                        sNewComplete = calcNextDate(lDuration, maxDate, Enum.ProjectScheduleDirection.DIRECTION_FWD, bWorkAnyDay);
                        sNewComplete = Date.max(sConstraintDate, sNewComplete);
                    }
                    break;
                case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN://"Finish No Later Than":
                    sNewComplete = calcNextDate(lDuration, maxDate, Enum.ProjectScheduleDirection.DIRECTION_FWD, bWorkAnyDay);
                    sNewComplete = Date.min(sConstraintDate, sNewComplete);
                    break;
                case Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT://"Duration Of Project":
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        sNewStart = project.StartDate;
                    } else {
                        sNewComplete = project.CompleteDate;
                    }
                    break;
            }
            var mode = "START";
            switch (taskConstraint) {
                //case ProjectConstants.AS_SOON_AS_POSSIBLE://"As Soon As Possible":
                //case ProjectConstants.AS_LATE_AS_POSSIBLE://"As Late As Possible":
                case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        mode = "START";
                    } else {
                        mode = "COMPLETE";
                    }
                    break;
                case Enum.ProjectScheduleConstraintTypes.MUST_START_ON://"Must Start On":
                case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN://"Start No Earlier Than":
                case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN://"Start No Later Than":
                    mode = "START";
                    break;
                case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON://"Must Finish On":
                case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN://"Finish No Earlier Than":
                case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN://"Finish No Later Than":
                    mode = "COMPLETE";
                    break;
                case Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT:
                    if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                        mode = "START";
                    } else {
                        mode = "COMPLETE";
                    }
                    break;
            }
            if (mode === "START") {
                if (taskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                    if (project.CompleteDate) {
                        sNewComplete = project.CompleteDate;
                    } else {
                        sNewComplete = calcMaxProjectDate(true);
                    }
                    lDuration = calcDuration(sNewStart, sNewComplete, lDuration, bWorkAnyDay);
                } else {
                    if (bActStart) {
                        sNewStart = task.ActStart;
                    } else {
                        //ensure that the new start date is a work day
                        if (bWorkAnyDay === 0) {
                            while (!isWorkDay(sNewStart, true, true)) {
                                sNewStart = Date.add(sNewStart, 1, Date.DAY);
                            }
                        }
                    }
                    if (bActComplete) {
                        sNewComplete = task.ActComplete;
                        lDuration = calcDuration(sNewStart, sNewComplete, lDuration, bWorkAnyDay);
                    } else {
                        sNewComplete = calcNextDate(lDuration, sNewStart, Enum.ProjectScheduleDirection.DIRECTION_FWD, bWorkAnyDay);
                    }
                }

            } else {
                if (taskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                    if (project.StartDate) {
                        sNewStart = project.StartDate;
                    } else {
                        sNewStart = calcMaxProjectDate();
                    }
                    lDuration = calcDuration(sNewStart, sNewComplete, lDuration, bWorkAnyDay);
                } else {
                    if (bActComplete) {
                        sNewComplete = task.ActComplete;
                    } else {
                        //ensure that the new complete date is a work day
                        if (bWorkAnyDay === 0) {
                            while (!isWorkDay(sNewComplete, true, true)) {
                                sNewComplete = Date.add(sNewComplete, -1, Date.DAY);
                            }
                        }
                    }
                    if (bActStart) {
                        sNewStart = task.ActStart;
                        lDuration = calcDuration(sNewStart, sNewComplete, lDuration, bWorkAnyDay);
                    } else {
                        sNewStart = calcNextDate(lDuration, sNewComplete, Enum.ProjectScheduleDirection.DIRECTION_BWD, bWorkAnyDay);
                    }
                }
            }
            //This happened when 'Must Finish On' on 4/18/2006 and an actual start 12/11/2006 was entered
            if (sNewStart > sNewComplete) {
                sNewComplete = sNewStart;
                lDuration = 0;
            }

            task.maxDate = maxDate;
            task.ConstraintDate = sConstraintDate;
            task.PlanStart = sNewStart;
            task.PlanComplete = sNewComplete;
            task.PlanDuration = lDuration;
            task.scheduled = true;

            task.scheduleOrder = ++scheduleOrder;
            schedulingSetTaskStatus(task);
            schedulingSetEventDates(task);
        }

        function schedulingSetEventDates(task) {
            var dtePlanStart;
            var dteEventStart;
            var dteEventEnd;
            var dMinutes;
            if (task.ShowOnCalendar === 1) {
                dtePlanStart = task.PlanStart;
                dteEventStart = task.EventStart;
                dteEventEnd = task.EventEnd;
                if (dteEventStart && dteEventEnd && dtePlanStart) {
                    dMinutes = Date.diff(dteEventStart, dteEventEnd, Date.MINUTE);

                    dteEventStart = dtePlanStart;
                    dteEventEnd = Date.add(dteEventStart, dMinutes, Date.MINUTE);
                    task.EventStart = dteEventStart;
                    task.EventEnd = dteEventEnd;
                }
            }
        }

        /**
         * Sets the schedule dates, and calculates the duration
         * @param maxDate
         * @param planStart
         * @param planComplete
         * @param task
         */
        function schedulingScheduleTask2(maxDate, planStart, planComplete, task) {

            var duration;
            var workAnyDay;

            if (planStart > planComplete) {
                //if PlanStart is later than PlanComplete then alter them both to this.projectStartDate
                planStart = project.StartDate;
                planComplete = project.StartDate;
            }
            workAnyDay = task.WorkAnyDay;
            duration = task.PlanDuration;
            duration = calcDuration(planStart, planComplete, duration, workAnyDay);
            task.maxDate = maxDate;
            task.PlanStart = planStart;
            task.PlanComplete = planComplete;
            task.PlanDuration = duration;
            task.scheduled = true;

            task.scheduleOrder = ++scheduleOrder;
            if (task.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                setConstraint(task, Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT);
                task.ConstraintDate = maxDate;
            }
            schedulingSetTaskStatus(task);
        }

        /**
         * Sets the schedule dates but takes the duration as a parameter
         * @param maxDate
         * @param planStart
         * @param planComplete
         * @param duration
         * @param task
         */
        function schedulingScheduleTask3(maxDate, planStart, planComplete, duration, task) {

            if (planStart > planComplete) {
                //if PlanStart is later than PlanComplete then alter them both to this.projectStartDate
                planStart = project.StartDate;
                planComplete = project.StartDate;
            }
            task.maxDate = maxDate;
            task.ConstraintDate = maxDate;
            task.PlanStart = planStart;
            task.PlanComplete = planComplete;
            task.PlanDuration = duration;
            task.scheduled = true;

            task.scheduleOrder = ++scheduleOrder;
            if (task.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                setConstraint(task, Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT);
                task.ConstraintDate = maxDate;
            }
            schedulingSetTaskStatus(task);
        }

        /**
         * Calculates the green/yellow/red status of a task
         * @param task
         */
        function schedulingSetTaskStatus(task) {
            // Note: If this function is changed, please change also the VB code and the Task Manager
            var today = Date.ISOToday;

            task.TaskStatus = Enum.ProjectTaskStatus.STATUS_GREEN;
            task.ScheduleNote = "";
            //var sConstraint:String = UtilNamedConstraint(task.TaskConstraint);
            var taskConstraint = task.TaskConstraint;

            //dont do anyhting if the task hasnt been scheduled yet
            if (!(task.scheduled)) {
                return;
            }
            // if task is set not to track status
            if (u.obn(task.ExcludeFromStatus) === 1) {
                return;
            }
            if (task.TaskType !== Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                switch (taskConstraint) {
                    //case ProjectConstants.AS_SOON_AS_POSSIBLE://"As Soon As Possible":
                    //case ProjectConstants.AS_LATE_AS_POSSIBLE://"As Late As Possible":
                    case Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT:
                        if (!task.ActStart) {
                            // No Act Start
                            if (task.PlanStart < today) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                task.ScheduleNote = "The task should have been started by today, but there is no actual start date.";
                            }
                            if (task.PlanComplete < today) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                task.ScheduleNote = "The task should have been completed by today, but there is no actual complete date.";
                            }
                        } else {
                            if (!task.ActComplete) {
                                // There is an Act Start
                                if (task.ActStart > task.PlanStart) {
                                    task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                    task.ScheduleNote = "The task should have been started earlier.";
                                }
                                // No Act Complete
                                if (task.PlanComplete < today) {
                                    task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                    task.ScheduleNote = "The task should have been completed before today, but there is no actual completion date.";
                                }
                            } else {
                                // There is an Act Complete
                                if (task.ActComplete > task.PlanComplete) {
                                    task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                    task.ScheduleNote = "The task should have been completed earlier.";
                                }
                            }
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.MUST_START_ON://"Must Start On":
                        if (!task.ActStart) {
                            if (task.ConstraintDate < today) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                task.ScheduleNote = "The task should have been started by today, but there is no actual start date.";
                            }
                        } else {
                            if (task.ConstraintDate !== task.ActStart) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                task.ScheduleNote = "The task start date has not been met.";
                            }
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN://"Start No Earlier Than":
                        if (task.ActStart) {
                            if (task.ActStart < task.ConstraintDate) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                task.ScheduleNote = "The task was started too early.";
                            }
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN://"Start No Later Than":
                        if (!task.ActStart) {
                            if (task.ConstraintDate < today) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                task.ScheduleNote = "The task should have been started by today, but there is no actual start date.";
                            }
                        } else {
                            if (task.ConstraintDate < task.ActStart) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_YELLOW;
                                task.ScheduleNote = "The task start date has not been met.";
                            }
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON://"Must Finish On":
                        if (!task.ActComplete) {
                            if (task.ConstraintDate < today) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                task.ScheduleNote = "The task should have been completed by today, but there is no actual complete date.";
                            }
                        } else {
                            if (task.ConstraintDate < task.ActComplete) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                task.ScheduleNote = "The task was not completed in time.";
                            }
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN://"Finish No Earlier Than":
                        if (task.ActComplete) {
                            if (task.ActComplete < task.ConstraintDate) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                task.ScheduleNote = "The task was completed too early.";
                            }
                        }
                        break;
                    case Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN://"Finish No Later Than":
                        if (!task.ActComplete) {
                            if (task.ConstraintDate < today) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                task.ScheduleNote = "The task should have been completed by today, but there is no actual complete date.";
                            }
                        } else {
                            if (task.ConstraintDate < task.ActComplete) {
                                task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                                task.ScheduleNote = "The task complete date has not been met.";
                            }
                        }
                        break;
                }

                if (task.ActComplete) {
                    if (task.ActComplete > task.PlanComplete) {
                        task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                        task.ScheduleNote = "The task should have been completed earlier.";
                    }
                } else {
                    if (today > task.PlanComplete) {
                        task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                        task.ScheduleNote = "The task should have been completed earlier.";
                    }
                }

            }
            if (project.CompleteDate) {
                if (task.PlanComplete > project.CompleteDate) {
                    task.TaskStatus = Enum.ProjectTaskStatus.STATUS_RED;
                    task.ScheduleNote = "The task will not be completed by the project completion date.";
                }
            }
        }

        function schedulingSetPercComp(task) {

            if (task.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                var children = getChildren(task);
                if (children.length > 0) {
                    var childTask;
                    var numerator = 0;
                    var denominator = 0;
                    var duration = 1;
                    for (var i = 0; i < children.length; i++) {
                        childTask = children[i];
                        if(childTask.rowstate && childTask.rowstate.includes('D')) continue; //skip this child if it's marked for delete
                        duration = childTask.PlanDuration;
                        // If duration = 0, then make it a 1
                        // Bug 6151, users could not understand weighted or arithmetic average concepts
                        if (duration == 0) duration = 1;

                        // Weighted average
                        numerator += (duration * childTask.PercComp);
                        denominator += duration;
                    }
                    //check for div by zero, constraint result to 0-100
                    task.PercComp = Math.round(Math.min(100, Math.max(0, numerator / Math.max(denominator, 1))));
                }
            }
        }

        function schedulingSetPredecessorsComplete() {

            if (!dsTasks) return;
            if (dsTasks.length == 0) return;

            var task;
            var predTask;
            var bPredecessorsComplete;

            // Only logical predecessors have been pushed down

            // Patch for the effective predecessors
            // They have not been pushed down yet
            schedulingPushPredecessorsDown(true);

            for (var i = 0; i < dsTasks.source.length; i++) {
                task = dsTasks.source[i];
                if (task.ScheduleTask === 1) {
                    bPredecessorsComplete = true;
                    var effPreds = getEffPreds(task);
                    if (effPreds) {
                        for (var j = 0; j <= effPreds.length - 1; j++) {
                            predTask = findTaskByKey(effPreds[j].PredKey);
                            if (predTask) {
                                if (predTask.PercComp !== 100) {
                                    bPredecessorsComplete = false;
                                    break;
                                }
                            }
                        }
                    }
                    if (bPredecessorsComplete) {
                        task.PredecessorsComplete = 1;
                    } else {
                        task.PredecessorsComplete = 0;
                    }
                }
            }
        }

        /**
         * Rolls the task data up to the project level
         */
        function schedulingUpdateProject() {

            var rootTasks = getScheduleRoot();

            var planStart = new Date(2050, 1, 1).toISOString();
            var planComplete = new Date(1975, 1, 1).toISOString();

            var numerator = 0;
            var denominator = 0;
            var percComp = 0;
            var duration = 0;
            var found = false;
            var task;

            for (var i = 0; i < rootTasks.length; i++) {
                task = rootTasks[i];
                //Only process root tasks that are scheduled tasks
                if (task.ScheduleTask === 1) {
                    found = true;
                    duration = task.PlanDuration;
                    // If duration = 0, then make it a 1
                    // Bug 6151, users could not understand weighted or arithmetic average concepts
                    if (duration === 0) {
                        duration = 1;
                    }
                    // Weighted average
                    numerator += (duration * task.PercComp);
                    denominator += duration;

                    planStart = Date.min(task.PlanStart, planStart);
                    planComplete = Date.max(task.PlanComplete, planComplete);
                }
            }

            if (found) {
                //Calculate PercComp for the project
                percComp = Math.min(100, Math.max(0, numerator / Math.max(1, denominator)));
                if (percComp >= 100) {
                    percComp = 100;
                } else {
                    if (percComp >= 99) {
                        percComp = 99;
                    } else {
                        if (percComp <= 0) {
                            percComp = 0;
                        } else {
                            percComp = Math.round(percComp);
                        }
                    }
                }

                // And duration for the project
                if (planStart > planComplete) {
                    planStart = planComplete;
                }
                duration = calcDuration(planStart, planComplete, 0, false);
            } else {
                // Could not find at least one task with ScheduleTask = 1
                percComp = 0;
                duration = 0;
            }
            // Now update the project list
            project.PercComp = percComp;
            project.Duration = duration;
        }

        /**
         * Takes the item (which is the newly added task), and sets defaults. Note: this does not add a task to the datasource
         * @param prev Optional. If supplied, it will set this task as a predecessor of the new task
         * @param item The newly inserted task from the datasource
         */
        function addTask(prev, item) {

            //The item object is the new row just before it's added
            item.ProjectKey = project.ProjectKey;
            item.SummaryTaskKey = prev ? prev.SummaryTaskKey : 0; //so that they are on the same level
            item.TaskName = "New Task";
            item.TaskType = Enum.ProjectScheduleTaskType.TRACKING_TASK;
            item.TaskAssignmentTypeKey = null;

            if (!findTrackBudgetInParents(item)) {
                item.TrackBudget = 1;
                item.MoneyTask = 1;
                item.MoneyTask = 1;
            } else {
                item.TrackBudget = 0;
                item.MoneyTask = 0;
            }

            item.HourlyRate = 0;
            item.IOCommission = 0;
            item.ShowDescOnEst = 0;
            item.ServiceKey = null;
            item.Taxable = 0;
            item.Taxable2 = 0;
            item.HideFromClient = 0;
            item.EstHours = 0;
            item.BudgetLabor = 0;
            item.EstLabor = 0;
            item.BudgetExpense = 0;
            item.EstExpense = 0;
            item.ApprovedCOHours = 0;
            item.ApprovedCHLabor = 0;
            item.ApprovedCOBudgetLabor = 0;
            item.ApprovedCOExpense = 0;
            item.ApprovedCOBudgetExp = 0;
            item.Contingency = 0;
            item.MasterTaskKey = 0;
            item.AllowAnyone = 0;
            item.EstimateHours = 0;
            item.FormCount = 0;
            item.ActualHours = 0;
            item.AllocatedHours = 0;
            item.AttachmentCount = 0;
            setConstraint(item, Enum.ProjectScheduleConstraintTypes.DEFAULT_CONSTRAINT);
            if (project.ScheduleDirection === 1) {
                item.ConstraintName = System.localize("|26675|As Soon As Possible|");
            } else {
                item.ConstraintName = System.localize("|26676|As Late As Possible|");
            }

            item.WorkAnyDay = 0;
            item.PercComp = 0;
            item.PercCompSeparate = 1;
            item.PredecessorsComplete = 0;
            item.Priority = 2;
            item.PriorityName = System.localize("|26677|Medium|");
            item.ScheduleTask = 1;

            var newDate;
            var findNextDateDirection = 0;
            if (scheduleDirection === Enum.ProjectDirection.FROM_PROJECT_START) {
                findNextDateDirection = Enum.ProjectScheduleDirection.DIRECTION_FWD;
            } else {
                findNextDateDirection = Enum.ProjectScheduleDirection.DIRECTION_BWD;
            }
            if (prev) {
                linkToTask(item, prev);
                item.AssignedPredsKeys = u.obs(prev.TaskKey); //must be a string because of the split function
                item.AssignedPreds = u.obs(prev.ProjectOrder);

                newDate = findNextWorkDate(prev.PlanComplete, findNextDateDirection);
                item.PlanStart = newDate;
                item.PlanComplete = newDate;
                item.PlanDuration = 1;
            } else {
                newDate = findNextWorkDate(project.StartDate, findNextDateDirection);
                item.PlanStart = newDate;
                item.PlanComplete = newDate;
                item.PlanDuration = 1;
            }
        }

        /**
         * Flag that is set when a batch deleting process is happening
         * @type {boolean}
         */
        var deletingTasks = false;

        /**
         * Takes an array of tasks and marks them as deleted in the datasource
         * @param tasks Array of tasks to delete
         */
        function deleteTasks(tasks) {

            if (deletingTasks) return;

            //see if the tasks have any transactions that could cause problems
            if (tasks && tasks.length > 0) {
                var deleteTaskItems = tasks.slice();
                var deleteTasksListLookup = {};
                var deleteTasksList = [];
                var i;
                var j;
                var aPredecessors;
                var aUsers;
                var task;
                var hasErrors = false;

                //sort by project order
                deleteTaskItems.sort(function (a, b) {
                    return a.ProjectOrder - b.ProjectOrder;
                });

                // now get the complete list of tasks to delete (including children)
                for (i = 0; i < deleteTaskItems.length; i++) {
                    deleteTasksAddTaskToList(deleteTaskItems[i], deleteTasksListLookup, deleteTasksList);
                }

                for (i = 0; i < deleteTasksList.length; i++) {
                    task = deleteTasksList[i].task;
                    if (task.HasTransactions === 1) {
                        addGeneralError('Task ' + task.TaskName + ' could not be deleted because it has transactions or approved estimates linked to it');
                        hasErrors = true;
                        break;
                    }
                }

                if (!hasErrors) {
                    //sort the list into reverse ProjectOrder.. so we delete children before parents
                    deleteTasksList.sort(function (a, b) {
                        return b.task.ProjectOrder - a.task.ProjectOrder;
                    });

                    //now process the removal
                    deletingTasks = true;
                    for (i = 0; i < deleteTasksList.length; i++) {
                        task = deleteTasksList[i].task;
                        deleteTaskPredecessors(task);
                        deleteTaskUsers(task);
                        if (task.rowstate) task.rowstate = task.rowstate.replace('U', ''); //deleting taskUsers adds a "U" to the rowstate. Then when "D" is added, it becomes "DU", which will cause it to still display in the UI
                        dsTasks.itemUpdated('D', task, true);

                        var parent = getParentTask(task);
                        if(parent){
                            schedulingSetPercComp(parent);
                        }
                    }
                    deletingTasks = false;

                    autoNumberTasks();

                    //reschedule the project
                    scheduleProject();
                } else {
                    dsTasks.rollbackModifiedItems({deletedItems: tasks});
                }
            }
        }

        /**
         * Rolls through the project tasks and autonumbers the task IDs
         */
        function autoNumberTasks() {
            setNewTaskID = undefined;
            if (!project.AutoIDTask) return;
            DataManager.reindex(dsTasks, dsTasks.hdd);
        }

        /**
         * Sets the Task ID of an individual task
         * @param task
         * @param parentNumber
         * @param currentNumber
         */
        function setTaskNumbers(task, parentNumber, currentNumber) {

            if (!project.AutoIDTask) return;
            if (task.rowstate === 'D') return;

            if (parentNumber.length === 0) {
                task.TaskID = currentNumber.toString();
            } else {
                task.TaskID = parentNumber + "." + currentNumber.toString();
            }
        }

        /**
         * Called recursively to add all of the child tasks to the list of tasks to delete
         * @param task
         * @param deleteTasksListLookup Must be an instance of an object that is an associative array of tasks to delete
         * @param deleteTasksList Must be an instance of an array
         */
        function deleteTasksAddTaskToList(task, deleteTasksListLookup, deleteTasksList) {
            if (task) {
                // Add task to the list
                if (deleteTasksListLookup[task.TaskKey]) {
                    //already been added to the list so no further processing required for this node
                } else {
                    deleteTasksListLookup[task.TaskKey] = true;
                    deleteTasksList.push({task: task, response: new ScheduleResponse(task)});
                    //if it has any children we need to process them as well
                    var children = getChildren(task);
                    for (var i = 0; i < children.length; i++) {
                        // now process recursively
                        deleteTasksAddTaskToList(children[i], deleteTasksListLookup, deleteTasksList);
                    }
                }
            }
        }

        /**
         * Returns false if the task cannot be moved to the new location
         * @param taskBeingMoved The task being dragged or moved in or out
         * @param referenceTask The task that the user is either dropping onto, or in front of.
         * @param dropAsChild If true, the taskBeingMoved is about to be a child of referenceTask
         */
        function isMoveTaskAllowed(taskBeingMoved, referenceTask, dropAsChild) {

            if (!taskBeingMoved) return false;
            if (!referenceTask) return false;

            //verify that taskBeingMoved is not a parent of referenceTask already
            var parent = getParentTask(referenceTask);
            if (parent === taskBeingMoved) {
                addGeneralError('|26679|You cannot move a task to one of its subtasks|');
                return false;
            }

            if(dropAsChild && taskBeingMoved.TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT){
                addGeneralError("|26681|Constraint conflict: A task with a constraint Duration Of Project which can not have a summary task.|");
                return false;

            }

            var oldParentTask = getParentTask(taskBeingMoved);
            if (parent && oldParentTask !== parent) {
                if (taskBeingMoved.MoneyTask === 1) {
                    //if MoneyTask is true it indicates that TrackBudget is true below us so we need to check the parent
                    if (findTrackBudgetInParents(parent)) {
                        addGeneralError("|26680|Track Expenses conflict: There is a parent task which tracks expenses|");
                        return false;
                    }
                }
                if (taskBeingMoved.TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                    addGeneralError("|26681|Constraint conflict: A task with a constraint Duration Of Project which can not have a summary task.|");
                    return false;
                }
            }

            while (parent && parent !== referenceTask) {
                if (parent === taskBeingMoved) {
                    addGeneralError('|26679|You cannot move a task to one of its subtasks|');
                    return false;
                }
                parent = getParentTask(parent);
                if (!parent) break; //we're at the root
            }
            return true;
        }

        /**
         * Returns false if the proposed parent cannot be set as a parent
         * @param referenceTask The task that is about to be the parent.
         */
        function isDropAsChildAllowed(referenceTask) {

            if(!referenceTask) return true;

            if (referenceTask.TaskType === Enum.TaskRestriction.BUDGETONLY) {
                if (referenceTask.taskUsers) {
                    return false;
                }
            }
            return true;
        }

        /**
         * Removes all assignments from a task
         * @param task
         */
        function removeTaskAssignedUsers(task) {

            if (!dsTaskUsers) {
                return;
            } else {
                var item;
                for (var i = 0; i < dsTaskUsers.source.length; i++) {
                    item = dsTaskUsers.source[i];
                    if (item.TaskKey === task.TaskKey) {
                        if(!usingDM) dsTaskUsers.itemUpdated('D', item, false);
                        removeTaskUser(item);
                    }
                }
            }

            rollupAssignedUsersToTask(task);
        }

        /**
         * This will rollup PercComp, ActStart, ActComplete from tTaskUser to tTask
         * It will also rollup tTaskUser.Hours to tTask.AllocatedHours
         * @param task
         */
        function rollupAssignedUsersToTask(task) {
            //
            // It will also rollup tTaskUser.Hours to tTask.AllocatedHours
            if (task) {
                var taskKey = task.TaskKey;
                var percCompSeparate = task.PercCompSeparate;
                var lstTaskList = dsTaskUsers.source;
                var dteActStart = Date.getISODate(1, 1, 2050);
                var actStart = false;
                var dteActComplete = Date.getISODate(1, 1, 1975);
                var actComplete = false;
                var userHours = 0;
                var totalHours = 0;
                var dAllocatedHours = 0;
                var totalHoursComp = 0;
                var percComp = 0;
                var numReviews = 0;
                var actCompleteCount = 0;
                var taskUserItem;

                for (var i = 0; i < lstTaskList.length; i++) {
                    if (taskKey === lstTaskList[i].TaskKey) {
                        taskUserItem = lstTaskList[i];
                        if(!taskUserItem.rowstate || taskUserItem.rowstate.indexOf('D') < 0) {
                            if ((taskUserItem.UserKey > 0) || (taskUserItem.ServiceKey > 0)) {
                                userHours = u.obn(taskUserItem.Hours);

                                // sum up the allocated hours, this is an exact number
                                dAllocatedHours += userHours;

                                // sum up the total hours, this is a number used for other calcs, i.e not exact
                                if (userHours < 1) {
                                    userHours = 1;
                                }
                                totalHours += userHours;

                                if (percCompSeparate === 1) {
                                    if (taskUserItem.ActStart) {
                                        actStart = true;
                                        dteActStart = Date.min(taskUserItem.ActStart, dteActStart);
                                    }
                                    if (taskUserItem.ActComplete) {
                                        actCompleteCount += 1;
                                        actComplete = true;
                                        dteActComplete = Date.max(taskUserItem.ActComplete, dteActComplete);
                                    }
                                    if (taskUserItem.PercComp) {
                                        totalHoursComp += userHours * (taskUserItem.PercComp / 100);
                                    }
                                    if (taskUserItem.ReviewedByTraffic === 1) {
                                        numReviews += 1;
                                    }
                                }
                            }
                        }
                    }
                }

                task.AllocatedHours = dAllocatedHours;

                if (percCompSeparate === 1) {
                    percComp = Math.roundTo(totalHoursComp / Math.max(totalHours, 1) * 100, 0, 4);
                    percComp = Math.min(100, Math.max(0, percComp));//constrain 100-0
                    if (actComplete) {
                        // we must have 100 % or all have finished
                        if (actCompleteCount === lstTaskList.length) {
                            percComp = 100;
                        } else if (percComp === 100) {

                        } else {
                            actComplete = false;
                        }
                    }
                    task.PercComp = percComp;

                    if (actStart) {
                        task.ActStart = dteActStart;
                    } else {
                        //don't blank out the Actual Start if it was entered in at the header level
                        // task.ActStart = null;
                    }

                    if (actComplete) {
                        task.ActComplete = dteActComplete;
                    } else {
                        task.ActComplete = null;
                    }

                    // if (numReviews === lstTaskList.length) {
                    //     dp.taskReviewedByTraffic = 1;
                    // } else {
                    //     dp.taskReviewedByTraffic = 0;
                    // }
                }
            }
        }

        /**
         * Must be called after a task has been moved to process any final cleanup logic
         * @param taskBeingMoved
         * @param oldParentTask
         * @returns {boolean}
         */
        function afterMoveTask(taskBeingMoved, oldParentTask) {
            if (!taskBeingMoved) return false;
            processAfterMoveTask(taskBeingMoved, getParentTask(taskBeingMoved), oldParentTask);
        }

        function processAfterMoveTask(taskBeingMoved, newParentTask, oldParentTask) {

            //Set the new parent to Summary, if it wasn't already
            if (newParentTask) newParentTask.TaskType = Enum.ProjectScheduleTaskType.SUMMARY_TASK;

            if (newParentTask !== oldParentTask) {
                if (taskBeingMoved.ScheduleTask === 1) {
                    //make sure all the parent tasks have ScheduleTask true
                    setScheduleTaskFlagUp(taskBeingMoved);
                }
                if (taskBeingMoved.MoneyTask === 1) {
                    //we need to recheck the old parent to see if it can now release its MoneyTask flag.
                    checkMoneyTaskFlag(oldParentTask);

                    //if the flag is up then there must be a child that is TrackBudget
                    //we know there cant be a TrackBudget above us or we would have failed the previous validation
                    //so we therefore only need to set the new parents MoneyTask flag to match
                    if (newParentTask && newParentTask.MoneyTask === 0) {
                        setMoneyTaskFlagUp(newParentTask);
                    }

                    if (newParentTask && newParentTask.TrackBudget === 1 && taskBeingMoved.TrackBudget === 1) {
                        newParentTask.TrackBudget = 0;
                        newParentTask.BudgetTaskKey = 0;
                        setTrackBudgetFlags(taskBeingMoved, 1);

                    }
                } else {
                    if (newParentTask) {
                        if (newParentTask.TrackBudget === 1) {
                            //if TrackBudget is set ...
                            //so we need to set the BudgetTaskKey down into child
                            setBudgetTaskKeyDown(taskBeingMoved, newParentTask.TaskKey, false);
                        } else if (newParentTask.BudgetTaskKey > 1) {
                            //if BudgetTaskKey is set then somewhere above someone is TrackBudget
                            //so we need to set the BudgetTaskKey down into child
                            setBudgetTaskKeyDown(taskBeingMoved, newParentTask.BudgetTaskKey, false);
                        }
                    }
                }
            }

            //check if the old parent still has children. If not, set it to a tracking task
            if (oldParentTask) {
                var oldParentChildren = findTaskByKey(oldParentTask.TaskKey, "SummaryTaskKey");
                if (!oldParentChildren) oldParentTask.TaskType = Enum.ProjectScheduleTaskType.TRACKING_TASK;
            }

            markTaskDirty(taskBeingMoved);

            circDetectCircularPredecessorSafe(taskBeingMoved.TaskKey);

            scheduleProject();
        }

        /**
         * Makes a copy of a task and adds it to the datasource
         * @param taskBeingCopied
         * @param dropTask Optional. The task that is the reference to where the new task is being copied. If specified, and dropAsChild is true, it will become the parent of the newly copied task.
         *      If specified, and dropAsChild is false, the newly copied task will be placed before this task. If it's not specified, the newly copied task will be placed after the taskBeingCopied
         * @param dropAsChild If true, the newly copied task will be a child of dropTask.
         */
        function copyTask(taskBeingCopied, dropTask, dropAsChild) {

            if (!taskBeingCopied) return;

            //use the same logic as when a template task is dropped
            dropTemplateTask(taskBeingCopied, dropTask ? dropTask : taskBeingCopied, dropAsChild, true);
        }

        /**
         * Checks for circular references within a node and it children.
         * There cannot be circular references caused by removing nodes but adding nodes can cuase circ references
         * This test for this and removes any invalid refs
         * @param moveTaskKey
         */

        function circDetectCircularPredecessorSafe(moveTaskKey) {
            // This function can be used when loading the predecessor relationships
            // AFTER the parent-child relationships have been changed (during Moves)

            // returns: Success true, Error false

            var taskKey;
            var predecessorKey;
            var taskPredecessorKey;
            var preds;
            var tasksToMove = {}; // Associative Array ex: parentMap[taskKey]
            var invalidPreds = [];

            //if there are no preds nothing to do
            if (dsPreds.length === 0) return;

            // Tasks to check will be saved in tasksToMove
            circGetTasksToMove(moveTaskKey, tasksToMove);

            var tasks = dsTasks.source;
            if (tasks.length) {
                for (var i = 0; i < tasks.length; i++) {
                    taskKey = tasks[i].TaskKey;
                    preds = lookupPredecessors(taskKey);
                    for (var j = 0; j < preds.length; j++) {
                        taskPredecessorKey = preds[j].TaskPredecessorKey;
                        predecessorKey = preds[j].PredecessorKey;
                        if (preds[j].isInvalid) {
                            invalidPreds.push(taskPredecessorKey);
                        } else {
                            //if the relation is used by one of the tasks to move
                            if (tasksToMove[taskKey] || tasksToMove[predecessorKey]) {
                                // keep it only if the 2 are part of the branch
                                if (!(tasksToMove[taskKey] && tasksToMove[predecessorKey])) {
                                    invalidPreds.push(preds[j]);
                                }
                            }
                        }
                    }
                }
                if (invalidPreds.length !== 0) {
                    for (var i = 0; i < invalidPreds.length; i++) {
                        if (checkCircularPredecessor(invalidPreds[i].TaskKey, invalidPreds[i].PredecessorKey, false)) {
                            removePredecessorFromSource(invalidPreds[i]);
                        }
                    }
                }
            }
        }

        function circGetTasksToMove(moveTaskKey, tasksToMove) {
            var parentTask = findTaskByKey(moveTaskKey);
            if (parentTask) {
                tasksToMove[moveTaskKey] = 1;
                var children = getChildren(parentTask);
                for (var i = 0; i < children.length; i++) {
                    circGetTasksToMove(children[i], tasksToMove);
                }
            }
        }

        function checkCircularPredecessor(taskKey, predecessorKey, addProposed) {

            //default addProposed to true
            addProposed = (typeof addProposed === 'undefined') ? true : addProposed;

            // returns: There is circular relationship = true
            //          There is NO circular relationship or Success = false
            if (taskKey === predecessorKey) {
                return true;
            }
            // work off the unfiltered list since the predecessor list is often filtered
            // 1 Load the predecessor records from the database to predPairs
            // These should have been already validated
            //clear the predPairs String
            predPairs = '';
            //Build the predPair string
            var preds;
            var tasks = dsTasks.source;
            for (var i = 0; i < tasks.length; i++) {
                // Fight memory problems with Flash, get a new object
                preds = lookupPredecessors(tasks[i].TaskKey);
                if (preds && preds.length > 0) {
                    for (var j = 0; j < preds.length; j++) {
                        if (!preds[j].isInvalid) {
                            circAddPredPair(preds[j].PredecessorKey, preds[j].TaskKey);
                        }
                    }
                }
            }
            if (addProposed) {
                // 2 Load the new proposed one
                // We can only do it if the reverse is not there yet
                var predPair = circBuildPredPair(taskKey, predecessorKey);
                if (predPairs.indexOf(predPair) >= 0) {
                    return true;
                }//add the proposed predPair....
                circAddPredPair(predecessorKey, taskKey);
            }
            if (!circExpandPredecessors()) {
                return true;
            }

            return circDetectCircularPredecessors(taskKey, predecessorKey);
        }

        function circDetectCircularPredecessors(taskKey, predecessorKey) {

            var sPredPair;
            var predPair;
            var allPreds = predPairs.split("#");
            for (var i = 0; i <= allPreds.length - 1; i++) {
                sPredPair = allPreds[i];
                predPair = sPredPair.split("*").join("").split("|");
                allPreds[i] = {TaskKey: predPair[1], PredecessorKey: predPair[0]};
            }

            var findKeys = {taskKey: true};

            // speed up search
            parentCheckedInCircDetection = new Object();

            return circFindTaskKey(findKeys, predecessorKey, allPreds);
        }

        function circFindTaskKey(oFindKeys, parentKey, allPreds) {
            var taskPreds;
            var predecessorKey;

            // if the parent has been checked already exit
            if (parentCheckedInCircDetection[parentKey]) {
                return false;
            }
            parentCheckedInCircDetection[parentKey] = true;

            taskPreds = allPreds.filter(circGetFilteredArray, parentKey);

            if (taskPreds.length === 0) {
                return false;
            } else {
                for (var i = 0; i < taskPreds.length; i++) {
                    predecessorKey = taskPreds[i].PredecessorKey;
                    // if one of the predecessors points back, then there is a circ ref
                    if (oFindKeys[predecessorKey]) {
                        return true;
                    } else {
                        var findKeys = $.cloneObjectDeep(oFindKeys);
                        // Add this predecessor into the chain
                        findKeys[parentKey] = true;
                        if (circFindTaskKey(findKeys, predecessorKey, allPreds)) return true;
                    }
                }
            }
            return false;
        }

        function circGetFilteredArray(item) {
            return item.TaskKey === this;
        }

        function circAddPredPair(predKey, succKey) {
            // Build something like *152|123*#*456|5657*

            var predPair = circBuildPredPair(predKey, succKey);
            if (predPairs) {
                if (predPairs.indexOf(predPair) === -1) {
                    predPairs += "#" + predPair;
                }
            } else {
                predPairs = predPair;
            }
        }

        function circBuildPredPair(predKey, succKey) {
            return "*" + predKey + "|" + succKey + "*";
        }

        function circExpandPredecessors() {
            // true Success
            // false Error
            var predecessorKey;
            var successorKey;
            var asPredPair;
            var sPredPair = "";
            var sCopy = "";
            var iIdx;
            var asTask;
            // Now push the predecessors down
            sCopy = predPairs;
            asPredPair = sCopy.split("#");
            for (iIdx = 0; iIdx <= asPredPair.length - 1; iIdx++) {
                sPredPair = asPredPair[iIdx];
                asTask = sPredPair.split("*").join("").split("|");
                predecessorKey = asTask[0];
                successorKey = asTask[1];
                if (!(circPushPredecessorsDown(predecessorKey, successorKey))) {
                    return false;
                }
            }

            // Now push the Successors down
            sCopy = predPairs;
            asPredPair = sCopy.split("#");
            for (iIdx = 0; iIdx <= asPredPair.length - 1; iIdx++) {
                sPredPair = asPredPair[iIdx];
                asTask = sPredPair.split("*").join("").split("|");
                predecessorKey = asTask[0];
                successorKey = asTask[1];
                if (!(circPushSuccessorDown(findTaskByKey(predecessorKey), successorKey))) {
                    return false;
                }
            }
            return true;
        }

        /**
         * Removes predecessors from the datasource
         * @param pred
         * @returns {boolean}
         */
        function removePredecessorFromSource(pred) {
            dsPreds.removeItem(pred);
            deletedPreds.push(pred);
            return true;
        }

        function circPushPredecessorsDown(predecessorKey, successorKey) {
            // true Success
            // false Error
            return circPushPredecessorDown(findTaskByKey(successorKey), predecessorKey);
        }

        function circPushPredecessorDown(parentTask, predecessorKey) {
            if (parentTask) {
                // true Success
                // false Error
                if (parentTask.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK) {
                    var children = getChildren(parentTask);
                    if (children && children.length) {
                        var childKey;
                        var predPair;
                        for (var i = 0; i <= children.length - 1; i++) {
                            childKey = children[i].TaskKey;
                            if (predecessorKey === childKey) {
                                //								UtilErrorAdd("Circular reference between "+parentTask.TaskName+" and "+children[i].TaskName);
                                return false;
                            }
                            // We can only do it if the reverse is not there yet
                            predPair = circBuildPredPair(childKey, predecessorKey);
                            if (predPairs.indexOf(predPair) >= 0) {
                                //								UtilErrorAdd("Circular reference between "+parentTask.TaskName+" and "+children[i].TaskName);
                                return false;
                            }
                            circAddPredPair(predecessorKey, childKey);
                            if (!(circPushPredecessorDown(children[i], predecessorKey))) {
                                return false;
                            }
                        }
                    }
                }
            }
            return true;
        }

        function setScheduleTaskFlagUp(task) {
            if (task.ScheduleTask !== 1) {
                task.ScheduleTask = 1;
                if (task.SummaryTaskKey) {
                    setScheduleTaskFlagUp(getParentTask(task));
                }
            }
        }

        function circPushSuccessorDown(parentTask, successorKey) {
            // true Success
            // false Error
            if (parentTask) {
                if ((parentTask.TaskType === Enum.ProjectScheduleTaskType.SUMMARY_TASK)) {
                    var children = getChildren(parentTask);
                    if (children && children.length) {
                        var childKey;
                        var predPair;
                        for (var i = 0; i <= children.length - 1; i++) {
                            childKey = children[i].TaskKey;
                            if (successorKey === childKey) {
                                //								UtilErrorAdd("Circular reference between "+parentTask.TaskName+" and "+children[i].TaskName);
                                return false;
                            }
                            // We can only do it if the reverse is not there yet
                            predPair = circBuildPredPair(successorKey, childKey);
                            if (predPairs.indexOf(predPair) >= 0) {
                                //								UtilErrorAdd("Circular reference between "+parentTask.TaskName+" and "+children[i].TaskName);
                                return false;
                            }
                            circAddPredPair(childKey, successorKey);
                            if (!(circPushSuccessorDown(children[i], successorKey))) {
                                return false;
                            }
                        }
                    }
                }
            }
            return true;
        }

        /**
         * Returns the taskParms object that is attached to the task which is used by the TaskUserHour class. This will create a taskParms object if it hasn't been created yet for this task
         * @param task
         * @returns {*}
         */
        function getTaskParms(task) {
            if (!task.tuhTaskParms) {
                createTaskParms(task);
            }
            return task.tuhTaskParms;
        }

        /**
         * Creates an instance of the taskParms object that's used by the TaskUserHour class, and attaches it to the task. It then calls calcColumns for the task to calculate the bucket columns.
         * @param task
         */
        function createTaskParms(task) {
            task.tuhTaskParms = TaskUserHour.initTaskParms(task.PlanStart, task.PlanComplete, idleDay);
            calcColumns(task, true);
        }

        /**
         * This will roll through all tasks that don't have taskParms yet and create them
         */
        function initBuckets() {
            loadTaskBucket();
        }

        /**
         * This is called to load the task buckets for the task assignments. It will continue to be called until all tasks have a taskParms object
         */
        function loadTaskBucket() {

            var arr = dsTasks.source.filter(filterNoTUHParms);
            if (arr.length > 0) {
                var task;
                var assignments = []; //array of taskUser records that need buckets from the server
                var filteredList;
                var taskList = [];
                for (var i = 0; i < arr.length; i++) {
                    task = arr[i];
                    if (task.TaskKey > 0) {
                        filteredList = dsTaskUsers.source.filter(filterTaskUserForTask, task);
                        if(filteredList.length > 0){
                            assignments = assignments.concat(filteredList);
                            taskList.push(task);
                        }
                    }
                    createTaskParms(task);
                }

                if (assignments.length > 0) {
                    //if there are existing tasks, get their buckets
                    Remote.remoteDataCall('Modules.Projects.Schedule', 'getAssignmentsForMultipleTasks', {
                        projectKey: project.ProjectKey,
                        tasks: taskList,
                        assignments: assignments,
                        application: 'cm.schedule.js_loadBuckets'
                    }).done(loadBucketReturn);
                } else {
                    loadTaskBucket(); //recursively call this in case there are now more tasks that need to have the tuhTaskParms added since this was called
                }
            }
        }

        function filterNoTUHParms(item) {
            return !item.tuhTaskParms;
        }

        function filterTaskUserForTask(item) {
            return (item.TaskKey === this.TaskKey) && (item.RecalcRequired === 1 || item.ManualAdjustment === 1 || TaskUserHour.isCalculating(this.tuhTaskParms) || TaskUserHour.recalcRequired(this.tuhTaskParms));
        }

        function loadBucketReturn(results) {

            var tuBucketLoad = results.data.assignments;
            if (tuBucketLoad) {
                var taskUser;
                var item;
                for (var i = 0; i < tuBucketLoad.length; i++) {
                    item = tuBucketLoad[i];
                    taskUser = dsTaskUsers.findByKey(item.TaskUserKey);
                    if (taskUser) {
                        for (var f in item) {
                            if (f.indexOf('Offset') >= 0) {
                                taskUser[f] = item[f];
                            }
                            if (f.indexOf('Actual_') >= 0) {
                                taskUser[f] = item[f];
                            }
                        }
                    }
                }
            }
            loadTaskBucket();
        }

        /**
         * This will call the TaskUserHour logic to calculate the bucket columns for the task assignments
         * @param task
         * @param fromLoad Pass in true if this is from the initial load. This will cause it to skip unnecessary calculations
         */
        function calcColumns(task, fromLoad) {

            var tuhTaskParms = getTaskParms(task);

            if (TaskUserHour.isCalculating(tuhTaskParms)) return;
            TaskUserHour.setStatus(tuhTaskParms, TaskUserHour.STATUS_CALCULATING);

            if (!fromLoad) TaskUserHour.setDates(tuhTaskParms, task.PlanStart, task.PlanComplete);
            TaskUserHour.calcRowColumnsFromDates(tuhTaskParms);

            if (!fromLoad) {
                //the columns have been recalculated, so we need to re-bucket all assignments for the current task
                var arr = dsTaskUsers.source.filter(filterTaskUserForTask, task);
                for (var i = 0; i < arr.length; i++) {
                    calcBucketHours(tuhTaskParms, arr[i], true, tuhTaskParms.bucketDelta);
                }
            }
            dsTaskUsers.fire();
            TaskUserHour.setStatus(tuhTaskParms, TaskUserHour.STATUS_NONE);
        }

        /**
         * Calls the TaskUserHour calcBucketHours method
         * @param tuhTaskParms
         * @param taskUser
         * @param skipDSFire
         * @param bucketDelta
         */
        function calcBucketHours(tuhTaskParms, taskUser, skipDSFire, bucketDelta, durationOnly) {

            bucketDelta = bucketDelta || 0;

            var duration = TaskUserHour.calcBucketHours(tuhTaskParms, taskUser, skipDSFire, bucketDelta, durationOnly);

            if (!skipDSFire) {
                dsTaskUsers.fire();
            }
            return duration;
        }

        /**
         * Calls the TaskUserHour calcBucketTotal method
         * @param tuhTaskParms
         * @param taskUser
         * @param offset
         */
        function calcBucketTotal(tuhTaskParms, taskUser, offset) {
            TaskUserHour.calcBucketTotal(tuhTaskParms, taskUser, offset);
            dsTaskUsers.fire();
        }

        /**
         * Causes the buckets to be recalculated if necessary
         * @param task
         */
        function validateTaskBuckets(task) {

            var tuhTaskParms = getTaskParms(task);
            if (tuhTaskParms) {
                if (TaskUserHour.recalcRequired(tuhTaskParms)) {
                    calcColumns(task);
                }

            }
        }

        /**
         * Runs the bucket validation on all tasks
         */
        function validateAllBuckets() {

            for (var i = 0; i < dsTasks.source.length; i++) {
                validateTaskBuckets(dsTasks.source[i]);
            }
        }

        /**
         * Called after blockout days have been refreshed from the server to set the blockout days object on each of the taskParms objects
         */
        function refreshBlockoutDays() {

            var tuhTaskParms;
            for (var i = 0; i < dsTasks.source.length; i++) {
                tuhTaskParms = getTaskParms(dsTasks.source[i]);
                if (tuhTaskParms) TaskUserHour.setBlockoutDays(tuhTaskParms, blockoutDays);
            }
        }

        /**
         * Adds cross-project predecessor information to the list for use in calculations
         * NOTE: Does not add a cross-project predecessor to a task!
         * @param predKey
         * @param predPlanComplete
         * @param predPlanDuration
         */
        function addCrossProjectPredInformation(predKey, predPlanComplete, predPlanDuration){
            if(u.obn(predKey) > 0) {
                crossProjectPreds[predKey] = {PlanComplete: predPlanComplete, PlanDuration: predPlanDuration};
            }
        }

        /**
         * Call this if a lag value has been updated so that the internal predecessor arrays can be updated with the new lag
         * @param task Task that had a predecessor lag updated
         */
        function lagUpdated(task) {
            markTaskDirty(task);
            loadPredecessors();
            formatPredValues();
        }

        /**
         * Links a task to another task as a predecessor
         * @param task Task that will be the successor
         * @param linkToThisTask Task that will be the predecessor
         * @param lag
         * @param delaySchedule Pass in true if you are linking multiple tasks at one time so that the scheduling logic doesn't run after each link
         * @returns Boolean: false if the new link will cause a circular reference
         * @param inWeeks
         */
        function linkToTask(task, linkToThisTask, lag, delaySchedule, inWeeks) {

            if (linkWillCauseCircularRef(task, linkToThisTask)) {
                addGeneralError('This predecessor would have caused a circular reference and has been removed');
                return false;
            }

            linkTasks([linkToThisTask, task], delaySchedule, lag, inWeeks);
            return true;
        }

        function linkWillCauseCircularRef(task, linkToThisTask) {

            var testLog = {};
            return testPreds(task, linkToThisTask.TaskKey, testLog);
        }

        function testPreds(task, testTaskKey, testLog) {

            if (testTaskKey === task.TaskKey) {
                return true;
            }

            var preds = getPreds({TaskKey: testTaskKey});
            if (preds) {
                for (var i = 0; i < preds.length; i++) {
                    if (testPreds(task, preds[i].PredKey, testLog)) {
                        return true;
                    }
                }
            }
            return false;
        }

        /**
         * Links multiple tasks as predecessors
         * @param tasks Array of tasks to link
         * @param delaySchedule Pass in true if you are linking multiple tasks at one time so that the scheduling logic doesn't run after each link
         * @param lag lag value that will be used for every row in the array (usually only used when linking only one task)
         * @param inWeeks 0 or 1 to indicate if the lag is in weeks (usually only used when linking only one task)
         */
        function linkTasks(tasks, delaySchedule, lag, inWeeks) {

            var task1;
            var task2;
            var preds;
            var newPred;
            var validLink;
            var found = false;

            lag = lag || 0;
            inWeeks = inWeeks || 0;

            if (tasks.length > 1) {
                var showErrors = tasks.length === 2; //if there are more than two tasks being linked, don't bother showing errors

                // go one less than the end cause we don't need the last one
                for (var i = 0; i < tasks.length - 1; i++) {
                    found = false;
                    validLink = true;

                    task1 = tasks[i];

                    if (!task1) continue;

                    if (task1.CrossProject !== 1) {
                        if (task1.ScheduleTask === 0) {
                            validLink = false;
                            if (showErrors) addGeneralError('|26682|Cannot link a non-schedule Task: ' + task1.TaskName + '|');
                        }
                        if (task1.TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                            validLink = false;
                            if (showErrors) addGeneralError('|26683|Task: ' + task1.TaskName + ' can not be linked because of task constraint|');
                        }
                    }

                    task2 = tasks[i + 1];

                    if (!task2) continue;

                    if (task2.ScheduleTask === 0) {
                        validLink = false;
                        if (showErrors) addGeneralError('|26684|Cannot link a non-schedule Task: ' + task2.TaskName + '|');
                    }
                    if (task2.TaskConstraint === Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT) {
                        validLink = false;
                        if (showErrors) addGeneralError('|26685|Task: ' + task2.TaskName + ' can not be linked because of task constraint|');
                    }
                    if (task2.SummaryTaskKey === task1.TaskKey) {
                        validLink = false;
                        if (showErrors) addGeneralError('|28770|Task: ' + task2.TaskName + ' can not be linked to its parent|');
                    }
                    if (!validLink || hasGeneralErrors()) {
                        if (showErrors) {
                            return;
                        } else {
                            continue;
                        }
                    }

                    preds = lookupPredecessors(task2.TaskKey);
                    // otherwise, try and add it
                    for (var j = 0; j < preds.length; j++) {
                        if (preds[j].PredecessorKey === task1.TaskKey) {
                            if (preds[j].rowstate && preds[j].rowstate.contains('I')) {
                                if (!task2.rowstate || task2.rowstate.indexOf('N') < 0) markTaskDirty(task2);
                                if (task1.CrossProject !== 1 && (!task1.rowstate || task1.rowstate.indexOf('N') < 0)) markTaskDirty(task1);
                            }
                            found = true;
                            break;
                        }
                        //if it's cross-project, then their PredecessorKeys will be the same
                        if(preds[j].CrossProject === 1) {
                            if (preds[j].PredecessorKey === task1.PredecessorKey) {
                                found = true;
                                break;
                            }
                        }
                    }
                    if (found) {
                        ///throw new ErrorWarning("Some predecessors already exist for this task "+task1.taskName);
                    } else {
                        //if it's cross-project OR there is not a circular predecessor
                        if (task1.CrossProject === 1 || !linkWillCauseCircularRef(task2, task1)) {
                            var newPred;
                            if(task1.CrossProject === 1){
                                task1.TaskKey = task2.TaskKey;
                                task1.Type = 'FS';
                                newPred = task1;
                            }else {
                                newPred = dsPreds.addItem({TaskPredecessorKey: dsPreds.getNextKey(), Lag: lag, InWeeks: inWeeks, PredecessorKey: task1.TaskKey, TaskKey: task2.TaskKey, Type: 'FS', ProjectOrder: task1.ProjectOrder, TaskID: task1.TaskID, TaskName: task1.TaskName, ProjectKey: project.ProjectKey, ProjectNumber: project.ProjectNumber, ProjectName: project.ProjectName});
                            }
                            dsPreds.itemUpdated('I', newPred, false);
                            if (!task2.rowstate || task2.rowstate.indexOf('N') < 0) markTaskDirty(task2);
                            if (task1.CrossProject !== 1 && (!task1.rowstate || task1.rowstate.indexOf('N') < 0)) markTaskDirty(task1);
                            loadPredecessors();
                            if (delaySchedule){
                                invalidateSchedule(true);
                            }else{
                                scheduleProject();
                            }
                        }
                    }
                }
                formatPredValues();
            } else {
                addGeneralError('|26686|You must select more than one task to link|');
            }
        }

        /**
         * Removes the predecessor link of a task from another task
         * @param task
         * @param unlinkFromThisTask
         * @param delaySchedule Pass in true if you are un-linking multiple tasks at one time so that the scheduling logic doesn't run after each un-link
         * @returns {boolean} True if a link was found to remove
         */
        function unlinkFromTask(task, unlinkFromThisTask, delaySchedule) {
            return unlinkTasks([unlinkFromThisTask, task], delaySchedule);
        }

        /**
         * Un-links multiple tasks as predecessors
         * @param tasks Array of tasks to unlink
         * @param delaySchedule Pass in true if you are un-linking multiple tasks at one time so that the scheduling logic doesn't run after each un-link
         * @returns {boolean} True if a link was found to remove
         */
        function unlinkTasks(tasks, delaySchedule) {

            var task1;
            var task2;
            var preds;
            var newPred;
            var compareKey;
            var found = false;
            if (tasks.length > 1) {
                ///process in reverse because otherwise the index becomes invalid
                // go one less than the end because we don't need the last one
                for (var i = tasks.length - 1; i > 0; i--) {
                    task2 = tasks[i];
                    if(task2) {
                        preds = lookupPredecessors(task2.TaskKey);

                        if (preds && preds.length > 0) {
                            //loop through every other task selected in the list and try to unlink it from task2
                            for (var k = tasks.length - 1; k >= 0; k--) {
                                if (k !== i) {
                                    task1 = tasks[k];

                                    if (task1 && task2) {
                                        // otherwise, try and add it
                                        for (var j = 0; j < preds.length; j++) {
                                            compareKey = (task1.CrossProject === 1) ? task1.PredecessorKey : task1.TaskKey; //if it's cross project, then the PredecessorKey is the key we're looking for from the list
                                            if (preds[j].PredecessorKey === compareKey) {
                                                removePredecessorFromSource(preds[j]);
                                                markTaskDirty(task2);
                                                markTaskDirty(task1);
                                                found = true;
                                                break;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                if (found) {
                    if (!delaySchedule) scheduleProject();
                } else {
                    //addGeneralError('|26687|These tasks are not linked|');
                }
                formatPredValues();
            } else if (tasks.length === 1) {
                task2 = tasks[0];
                preds = lookupPredecessors(task2.TaskKey);
                for (var j = 0; j < preds.length; j++) {
                    removePredecessorFromSource(preds[j]);
                    markTaskDirty(task2);
                    found = true;
                }
                preds = lookupSuccessors(task2.TaskKey);
                for (var j = 0; j < preds.length; j++) {
                    removePredecessorFromSource(preds[j]);
                    markTaskDirty(task2);
                    found = true;
                }
                if (found) {
                    if (!delaySchedule) scheduleProject();
                } else {
                    addGeneralError('|26688|No linked tasks were found|');
                }
                formatPredValues();
            } else {
                addGeneralError('|26689|You must select at least one task to unlink|');
            }
            return found;
        }

        /**
         * Used to default the values for a new taskUser assignment
         * @param item The new taskUser item
         */
        function taskUserDefaults(item) {

            item.ServiceKey = 0;
            item.ServiceDescription = '';
            item.ActualHours = 0;
            item.UserKey = 0;
            item.UserName = '';
            item.PercComp = 0;
            item.ProjectKey = project.ProjectKey;

            if(options.retainer){
                item.ManualAdjustment = 1;
                item.RetainerKey = project.RetainerKey;

                if (dsTasks.si.tuhTaskParms) {
                    var buckets = dsTasks.si.tuhTaskParms.buckets;
                    for (var b = 0; b < buckets.length; b++) {
                        if(buckets[b].prevWeek){
                            item['Offset_' + b] = 0;
                        }
                    }
                }
            }
        }

        /**
         * Called after a taskUser has been deleted from the datasource. Note- does not remove it from the datasource
         * @param item
         */
        function removeTaskUser(item) {

            if(usingDM) dsTaskUsers.itemUpdated('D', item, false);

            deletedTaskUsers.push(item);
            var task = findTaskByKey(item.TaskKey);
            if (task) {
                rollupAssignedUsersToTask(task);
                markTaskDirty(task);
            }
        }

        /**
         * Adds user to the list of users that should be removed from the project team upon save
         * @param userKey
         */
        function removeUsersFromProject(userKey){
            if(!removeProjectUsers) removeProjectUsers = [];
            removeProjectUsers.push(userKey);
        }

        /**
         * Checks to see if any assignments that are about to be reassigned are 100%
         * @param opt Option object. The reassign settings are passed in through this object.
         *      ReasignOption:  1-Assign a person to a service
         *                      2-Replace a person with a different person
         *                      3-Remove this person from their assignments
         *                      4-Replace one service with another
         *                      5-Remove assignments
         *                      6-All unassigned
         * @param selectedTasks Array. Optional. If passed in, only use these tasks, otherwise, get the tagged rows from the main schedule
         * @returns Boolean: false if any assignment is 100%
         */
        function reassignCheck(opt, selectedTasks) {

            var users = dsTaskUsers.source;

            selectedTasks = selectedTasks || dsTasks.getTaggedRows('TaskKey');
            var sKeys = "," + selectedTasks.join(',') + ',';

            for (var i = 0; i < users.length; i++) {
                if (sKeys.indexOf(',' + users[i].TaskKey + ',') > -1) {
                    switch (opt.ReasignOption) {
                        case 1:  //Assign a Person to a Service
                            if (users[i].ServiceKey === opt.ServiceKey && u.obn(users[i].PercComp) >= 100) {
                                return false;
                            }
                            break;
                        case 2:  //Replace a person w a different person
                            if (users[i].UserKey === opt.UserKey && u.obn(users[i].PercComp) >= 100) {
                                return false;
                            }
                            break;
                        case 3:  //Remove this person from their assignments
                            if (users[i].UserKey === opt.UserKey && u.obn(users[i].PercComp) >= 100) {
                                return false;
                            }
                            break;
                        case 4:  //Replace one service with another
                            if (users[i].ServiceKey === opt.ServiceKey && u.obn(users[i].PercComp) >= 100) {
                                return false;
                            }
                            break;
                        case 5:  //Remove assignments
                            if ((users[i].UserKey === opt.UserKey || users[i].ServiceKey === opt.ServiceKey) && u.obn(users[i].PercComp) >= 100) {
                                return false;
                            }
                            break;
                    }
                }
            }
            return true;
        }


        /**
         * Reassigns multiple users
         * @param opt Option object. The reassign settings are passed in through this object.
         *      ReasignOption:  1-Assign a person to a service
         *                      2-Replace a person with a different person
         *                      3-Remove this person from their assignments
         *                      4-Replace one service with another
         *                      5-Remove assignments
         *                      6-All Unassigned
         * @param selectedTasks Array. Optional. If passed in, only use these TaskKeys, otherwise, get the tagged rows from the main schedule
         * @param sendNotifications. If passed in, sent the SendNotifications field on the new assignment to the value
         *
         */
        function reassign(opt, selectedTasks, sendNotifications) {

            //associative array of taskKeys that will need to have the assignment rolled up
            var updatedTasks = {};

            selectedTasks = selectedTasks || dsTasks.getTaggedRows('TaskKey');

            var newTaskArray = [];
            for (var i = 0; i < selectedTasks.length; i++) {
                var taskItem = findTaskByKey(selectedTasks[i]);
                if (taskItem && taskItem.TaskType === 2) {
                    newTaskArray.push(selectedTasks[i]);
                }
            }

            selectedTasks = newTaskArray;

            if (opt.ReasignOption === 6) {
                if (!opt.UserKey && !opt.ServiceKey) {
                    addGeneralError('|28771|You must select either a user or service to assign|');
                    return;
                }
                //Assign new users to unassigned tasks (rather than replacing an existing assignment, like below)
                var newTaskUser;
                //selectedTasks is just an array of TaskKeys (not the whole task object)
                for (var i = 0; i < selectedTasks.length; i++) {
                    newTaskUser = {
                        TaskUserKey: dsTaskUsers.getNextKey(),
                        TaskKey: selectedTasks[i],
                        ProjectKey: project.ProjectKey,
                        UserKey: opt.UserKey,
                        UserName: opt.UserName,
                        Initials: opt.Initials,
                        ServiceKey: opt.ServiceKey,
                        ServiceCode: opt.ServiceCode,
                        ServiceDescription: opt.ServiceDescription,
                        SendNotifications: sendNotifications
                    };

                    dsTaskUsers.addItem(newTaskUser);
                    dsTaskUsers.itemUpdated('I');

                    updatedTasks[selectedTasks[i]] = true;
                }
            } else {
                var users = dsTaskUsers.source;
                var assignmentsToRemove = [];
                var sKeys = "," + selectedTasks.join(',') + ',';

                for (var i = 0; i < users.length; i++) {
                    if (sKeys.indexOf(',' + users[i].TaskKey + ',') > -1) {
                        switch (opt.ReasignOption) {
                            case 1:  //Assign a Person to a Service
                                if (users[i].ServiceKey === opt.ServiceKey && u.obn(users[i].PercComp) < 100) {
                                    users[i].UserKey = opt.UserKey;
                                    users[i].UserName = opt.UserName;
                                    users[i].Initials = opt.Initials;
                                    users[i].SendNotifications = sendNotifications;
                                    dsTaskUsers.itemUpdated('U', users[i]);
                                    updatedTasks[users[i].TaskKey] = true;
                                }
                                break;
                            case 2:  //Replace a person w a different person
                                if (users[i].UserKey === opt.UserKey && u.obn(users[i].PercComp) < 100) {
                                    users[i].UserKey = opt.UserKey2;
                                    users[i].UserName = opt.UserName2;
                                    users[i].Initials = opt.Initials2;
                                    users[i].SendNotifications = sendNotifications;
                                    dsTaskUsers.itemUpdated('U', users[i]);
                                    updatedTasks[users[i].TaskKey] = true;

                                    if(opt.RemoveFromProject === 1 && opt.UserKey > 0){
                                        removeUsersFromProject(opt.UserKey);
                                    }
                                }
                                break;
                            case 3:  //Remove this person from their assignments
                                if (users[i].UserKey === opt.UserKey && u.obn(users[i].PercComp) < 100) {
                                    users[i].UserKey = undefined;
                                    users[i].UserName = undefined;
                                    users[i].Initials = undefined;
                                    if (u.obn(users[i].ServiceKey) === 0) {
                                        users[i].ServiceKey = opt.ServiceKey;
                                        users[i].ServiceCode = opt.ServiceCode;
                                        users[i].ServiceDescription = opt.ServiceDescription;

                                    }
                                    dsTaskUsers.itemUpdated('U', users[i]);
                                    updatedTasks[users[i].TaskKey] = true;

                                    if(opt.RemoveFromProject === 1 && opt.UserKey > 0){
                                        removeUsersFromProject(opt.UserKey);
                                    }
                                }
                                break;
                            case 4:  //Replace one service with another
                                if (users[i].ServiceKey === opt.ServiceKey && u.obn(users[i].PercComp) < 100) {
                                    users[i].ServiceKey = opt.ServiceKey2;
                                    users[i].ServiceCode = opt.ServiceCode2;
                                    users[i].ServiceDescription = opt.ServiceDescription2;
                                    dsTaskUsers.itemUpdated('U', users[i]);
                                    updatedTasks[users[i].TaskKey] = true;
                                }
                                break;
                            case 5:  //Remove assignments
                                if (!opt.UserKey && !opt.ServiceKey) {
                                    addGeneralError('|26690|You must select either a user or service to remove|');
                                    return;
                                }
                                if ((users[i].UserKey === opt.UserKey || users[i].ServiceKey === opt.ServiceKey) && u.obn(users[i].PercComp) < 100) {
                                    assignmentsToRemove.push(users[i]);
                                    updatedTasks[users[i].TaskKey] = true;

                                    if(opt.RemoveFromProject === 1 && opt.UserKey > 0){
                                        removeUsersFromProject(opt.UserKey);
                                    }
                                }
                                break;
                        }
                    }
                }

                for (var i = 0; i < assignmentsToRemove.length; i++) {
                    if(!usingDM) dsTaskUsers.removeItem(assignmentsToRemove[i]);
                    removeTaskUser(assignmentsToRemove[i]);
                }
            }

            //rollup the newly assigned users to the tasks
            var task;
            for (var taskKey in updatedTasks) {
                task = findTaskByKey(taskKey);
                rollupAssignedUsersToTask(task);
                markTaskDirty(task);
            }
        }

        /**
         * Used by getModifiedItems to get the array of addedUsers so that they can be added to the assignment team. (Note: deletedUsers are not used at this time)
         * @returns {{addedUsers: Array.<*>, deletedUsers: Array}}
         */
        function getAssignmentChanges() {

            //this is the list of taskUsers that have been inserted
            var insertedTaskUsers = dsTaskUsers.source.filter(function (item) {
                return item.UserKey > 0 && item.rowstate && item.rowstate.indexOf('I') >= 0;
            });
            var insertedKeys = [];
            for (var i = 0; i < insertedTaskUsers.length; i++) {
                insertedKeys.push(insertedTaskUsers[i].UserKey);
            }
            insertedKeys = u.uniquify(insertedKeys);

            //this is the list of current taskUsers that have not been inserted nor deleted
            var currentTaskUsers = dsTaskUsers.source.filter(function (item) {
                return item.UserKey > 0 && item.rowstate && item.rowstate.indexOf('I') < 0 && item.rowstate.indexOf('D') < 0;
            });
            var currentKeys = [];
            for (var i = 0; i < currentTaskUsers.length; i++) {
                currentKeys.push(currentTaskUsers[i].UserKey);
            }
            currentKeys = u.uniquify(currentKeys);

            //this is the list of inserted users that were not in the list before, which means they should be added to the project team
            var addedUsers = insertedKeys.filter(function (item) {
                return isNaN(currentKeys.find(function (item2) {
                    return item === item2;
                }));
            });

            //at this point we are NEVER deleting users from the project team
            return {addedUsers: addedUsers, deletedUsers: []};
        }

        /**
         * Called when a template task is dragged over and dropped on a schedule
         * @param templateTask
         * @param scheduleTask
         * @param dropAsChild
         * @param dropAfterScheduleTask
         * @param keyMap
         * @returns {*}
         */
        function dropTemplateTask(templateTask, scheduleTask, dropAsChild, dropAfterScheduleTask, keyMap) {

            keyMap = keyMap || [];

            if(keyMap.length === 0){
                addTemplatePreds(); //this will clear out the previously set "NewKeys"
            }

            var taskBeingCopied = $.cloneObjectDeep(templateTask); //just converts it to JSON and converts it back to an object
            var children = dsTasks.source.includes(templateTask) ? DataManager.getChildren(dsTasks, templateTask, true) : null;
            var newTask = addTemplateTask(taskBeingCopied, scheduleTask, dropAsChild, dropAfterScheduleTask, keyMap, children);

            var position;
            if (scheduleTask) {
                if (dropAsChild) {
                    //new task goes directly below the dropTask and will be a child
                    position = 'child';
                } else {
                    position = dropAfterScheduleTask ? 'after' : 'before';
                }
            }

            position = position || 'after';
            dsTasks.addItem(newTask, position, scheduleTask);

            if (dsTasks.source.includes(templateTask)) {
                //the "templateTask" is actually an existing schedule task being copied
                var preds = getPreds(scheduleTask);
                if(preds) {
                    var linkToThisTask;
                    for (var i = 0; i < preds.length; i++) {
                        linkToThisTask = findTaskByKey(preds[i].PredKey);
                        linkToTask(newTask, linkToThisTask, u.obn(preds[i].Lag), true, u.obn(preds[i].InWeeks));
                    }
                }
            } else {
                addTemplatePreds(keyMap);
            }

            autoNumberTasks();
            afterMoveTask(newTask);
            markTaskDirty(scheduleTask);
            return newTask;
        }

        /**
         * Recurrsively copies the template task information, and its children and puts it into the schedule
         * @param taskBeingCopied
         * @param dropTask
         * @param dropAsChild
         * @param dropAfterScheduleTask
         * @param keyMap Must be an instance of an array. The first call, you can just pass in an empty array
         * @returns {*}
         */
        function addTemplateTask(newTask, dropTask, dropAsChild, dropAfterScheduleTask, keyMap, templateTaskChildren) {

            newTask.OldKey = newTask.TaskKey;
            //addded this to get any to do's on the moved task
            newTask.FromTaskKey = newTask.TaskKey;
            newTask.TaskKey = dsTasks.getNextKey();

            if (dropAsChild && dropTask) {
                newTask.SummaryTaskKey = dropTask.TaskKey;
            } else {
                if (dropTask) {
                    newTask.SummaryTaskKey = dropTask.SummaryTaskKey;
                } else {
                    newTask.SummaryTaskKey = 0;
                }
            }
            newTask.ProjectKey = project.ProjectKey;
            newTask.rowstate = 'I';
            newTask.batch = dsTasks.options.batch;
            newTask.DisplayOrder = null;
            newTask.ProjectOrder = null;
            newTask.PercComp = 0;
            newTask.ActStart = null;
            newTask.ActComplete = null;
            newTask.HasTransactions = 0;
            newTask.ActualHours = 0;

            if(newTask.TrackBudget === 1){
                //verify that no parents or children are tracking budget already

                //get the parent task since getParent won't work for this task yet
                var parentTask = newTask.SummaryTaskKey > 0 ? findTaskByKey(newTask.SummaryTaskKey) : 0;

                if (checkTrackBudgetState(newTask, 1, parentTask)) {
                    //this task can track budget, cascade MoneyTask = 1 to all parent tasks above it and BTK down
                    setTrackBudgetFlags(newTask, 1, parentTask);
                } else {
                    newTask.TrackBudget = 0;
                }
            }
           

            if (keyMap) keyMap.push({OldKey: newTask.OldKey, NewKey: newTask.TaskKey});

            if (!project.AutoIDTask) {
                var testCount = 0;
                while (!isTaskIDUnique(newTask.TaskID) && testCount < 20) {
                    newTask.TaskID += '_';
                    testCount++;
                }
            }

            //Add taskUsers
            //if the newTask and dropTask had the same TaskKey, then we're copying a task from the schedule and not bringing over a task from the template
            var taskUsers = (dropTask && (newTask.OldKey === dropTask.TaskKey)) ? dropTask.taskUsers : getTemplateTaskUsers(newTask.OldKey) || newTask.taskUsers;
            if (taskUsers) {
                if (!Array.isArray(taskUsers)) taskUsers = [taskUsers];
                var newTaskUser;
                var copyItem;
                for (var i = 0; i < taskUsers.length; i++) {
                    copyItem = taskUsers[i];
                    newTaskUser = {};
                    newTaskUser.TaskUserKey = dsTaskUsers.getNextKey();
                    newTaskUser.TaskKey = newTask.TaskKey;
                    newTaskUser.ProjectKey = newTask.ProjectKey;
                    newTaskUser.UserKey = copyItem.UserKey;
                    newTaskUser.UserName = copyItem.UserName;
                    newTaskUser.Initials = copyItem.Initials;
                    newTaskUser.ServiceKey = copyItem.ServiceKey;
                    newTaskUser.ServiceCode = copyItem.ServiceCode;
                    newTaskUser.ServiceDescription = copyItem.ServiceDescription;
                    newTaskUser.Hours = copyItem.Hours;
                    newTaskUser.Comments = copyItem.Comments;
                    newTaskUser.RecalcRequired = copyItem.RecalcRequired;

                    dsTaskUsers.addItem(newTaskUser);

                    dsTaskUsers.itemUpdated('I');
                }
            }

            var children = templateTaskChildren ? templateTaskChildren : newTask.children;
            children = children || [];
            var grandChildren;
            var insertAfter = newTask;
            for (var i = 0; i < children.length; i++) {
                grandChildren = dsTasks.source.includes(children[i]) ? DataManager.getChildren(dsTasks, children[i], true) : children[i].children;
                insertAfter = addTemplateTask(children[i], insertAfter, (i === 0), true, keyMap, grandChildren);
            }
            return newTask;
        }

        /**
         * Gets the task assignments from the template task in order to copy them to the schedule
         * @param templateTaskKey
         * @returns {*|Array|{}}
         */
        function getTemplateTaskUsers(templateTaskKey) {

            var dsTemplateTaskUsers = parms.selectedTemplate.dsTaskUsers;
            if (!dsTemplateTaskUsers) return;

            return dsTemplateTaskUsers.findByKey(templateTaskKey, 'TaskKey');
        }

        /**
         * Adds in the template prdecessors using the keyMap that was created in addTemplateTask
         * @param keyMap
         */
        function addTemplatePreds(keyMap) {

            if (!keyMap || keyMap.length === 0){
                var allPreds = getTemplatePreds();
                for (var i = 0; i < allPreds.length; i++) {
                    allPreds[i].NewTaskKey = null;
                    allPreds[i].NewPredKey = null;
                    allPreds[i].CopyCrossProject = false;
                }
                return;
            }

            var oldKey;
            var newKey;
            var preds;
            var predKey;
            var items;

            for (var i = 0; i < keyMap.length; i++) {
                oldKey = keyMap[i].OldKey;
                newKey = keyMap[i].NewKey;
                preds = getTemplatePreds(oldKey);
                if(preds) {
                    if (!Array.isArray(preds)) {
                        preds = [preds];
                    }
                    for (var j = 0; j < preds.length; j++) {
                        if (preds[j].CrossProject === 1) {
                            preds[j].CopyCrossProject = true;
                            preds[j].NewTaskKey = newKey;
                        } else {
                            predKey = preds[j].PredecessorKey;
                            items = keyMap.filter(function (item) {
                                return item.OldKey === predKey;
                            });
                            if (items && items.length > 0) {
                                preds[j].NewPredKey = items[0].NewKey;
                                preds[j].NewTaskKey = newKey;
                            }
                        }
                    }
                }
            }

            //only copy over the predecessors that have NewPredKey set
            var dsTemplatePreds = parms.selectedTemplate.dsPreds;
            if(dsTemplatePreds) {
                var templatePred;
                var task;
                var linkToThisTask;
                var newKeys;
                for (var i = 0; i < dsTemplatePreds.source.length; i++) {
                    templatePred = dsTemplatePreds.source[i];
                    if (templatePred.NewPredKey) {
                        task = findTaskByKey(templatePred.NewTaskKey);
                        linkToThisTask = findTaskByKey(templatePred.NewPredKey);
                        linkToTask(task, linkToThisTask, u.obn(templatePred.Lag), true);

                        newKeys = [];
                        preds = lookupPredecessors(task.TaskKey);
                        for (var j = 0; j < preds.length; j++) {
                            newKeys.push(u.obn(preds[j].PredecessorKey).toString());
                        }
                        task.AssignedPredsKeys = newKeys.join(',');
                    } else if (templatePred.CopyCrossProject) {
                        task = findTaskByKey(templatePred.NewTaskKey);
                        linkToTask(task, templatePred, u.obn(templatePred.Lag), true);

                        newKeys = [];
                        preds = lookupPredecessors(task.TaskKey);
                        for (var j = 0; j < preds.length; j++) {
                            newKeys.push(u.obn(preds[j].PredecessorKey).toString());
                        }
                        task.AssignedPredsKeys = newKeys.join(',');
                    }
                }
            }
        }

        /**
         * Gets predecessors of the template task
         * @param templateTaskKey Optional, if null, return all preds
         * @returns {*|Array|{}}
         */
        function getTemplatePreds(templateTaskKey) {

            var dsTemplatePreds = parms.selectedTemplate.dsPreds;
            if (!dsTemplatePreds) return;

            if (!templateTaskKey) return dsTemplatePreds.source;

            return dsTemplatePreds.findByKey(templateTaskKey, 'TaskKey');
        }

        /**
         * Formats the predecessor values for the select list control
         */
        function formatPredValues() {

            //force a refresh so that the HDD is built before the format function is called below
            // dsPreds.refresh(true, false);
            // dsTasks.refresh(true, false);

            //u.formatSelectListValues(dsTasks.source, 'predecessors', 'PredecessorKey', 'ProjectOrder', 'AssignedPredsKeys', 'AssignedPreds', ',', ',');
        }

        /**
         * This will set the "PredecessorKey" to the newly inserted TaskKey.
         * The HDD will automatically set the "TaskKey" because that's how it's linked to the schedule DS, but it doesn't know about "PredecessorKey"
         * @param taskKeyMap
         */
        function commitPredKeys(taskKeyMap) {
            var keyMap = buildKeyMap(taskKeyMap);

            var list = dsPreds.source.filter(function (item) {
                return item.PredecessorKey < 0 || item.TaskKey < 0;
            });

            var oldKey;
            var oldPredKey;
            var key;
            var newPredKey;
            var item;
            for (var i = 0; i < list.length; i++) {
                item = list[i];
                oldPredKey = 0;
                if(item.PredecessorKey < 0){
                    oldPredKey = item.PredecessorKey;
                    newPredKey = keyMap[oldPredKey];
                    if (newPredKey) {
                        item.PredecessorKey = newPredKey;
                    }
                }

                if(item.TaskKey < 0){
                    oldKey = item.TaskKey;
                    key = keyMap[oldKey];
                    if (key) {
                        item.TaskKey = key;
                    }
                }

                if (oldPredKey < 0) {
                    //find the task that this predecessor is on and replace it's negative key in the AssignedPredsKeys
                    var task = findTaskByKey(item.TaskKey);
                    if(!task) task = findTaskByKey(oldKey);
                    if (task) {
                        var selectedPredKeys = task.AssignedPredsKeys ? task.AssignedPredsKeys.split(',') : [];
                        for (var j = 0; j < selectedPredKeys.length; j++) {
                            if(selectedPredKeys[j] === u.obs(oldPredKey)){
                                selectedPredKeys[j]=newPredKey.toString();
                            }
                        }
                        task.AssignedPredsKeys = selectedPredKeys.join(',');
                    }
                }
            }
        }

        //copied from DataManager...
        function buildKeyMap(keys) {
            var keyMap = {};
            if (keys) {
                for (var i = 0; i < keys.length; i++)keyMap[keys[i].OldKey] = keys[i].NewKey;
            }
            return keyMap;
        }

        /**
         * Returns true if the testID is not used by any other task
         * @param testID
         * @returns {boolean}
         */
        function isTaskIDUnique(testID) {
            var task = dsTasks.findByKey(testID, 'TaskID');
            return task ? false : true;
        }

        /**
         * Takes an array of selected tasks and returns a hierarchical structure with the parents
         * @param selectedTasks Array of selected tasks
         * @param tasksOnly Boolean if true, don't include the assignments with the data
         * @param taskFilter object {entity: UserKey|ServiceKey, key: key}
         * @param taggedTaskUserKeys Array of TaskKeys that were currently selected by the user. We will try to mark them as tagged again if they're still in the list
         * @returns {Array} tasks and their parents
         */
        function getSelectedTaskAssignmentsWithParents(selectedTasks, tasksOnly, taskFilter, taggedTaskUserKeys) {

            var tasks = selectedTasks.sortOn('ProjectOrder:dni'); //add them in reverse order to get their parents

            var taskKeysAdded = {};
            var returnTasks = [];
            var taskKey;
            var parent;
            var taskUsers;
            var taskItem;
            var parentTaskItem;
            var childTask;
            var foundFilter;
            var taskAdded;
            var allTagged = false;

            if(taskFilter && taskFilter.allUnassigned){
                taskFilter.entity = '';
            }

            if(taskFilter && taskFilter.allTagged){
                allTagged = true;
            }

            for (var i = 0; i < tasks.length; i++) {
                taskItem = {
                    TempKey: $.guid++, TaskID: tasks[i].TaskID, TaskName: tasks[i].TaskName, SummaryTaskKey: taskKeysAdded[tasks[i].SummaryTaskKey] ? taskKeysAdded[tasks[i].SummaryTaskKey] : null, DisplayOrder: tasks[i].DisplayOrder,
                    TaskLevel: tasks[i].TaskLevel, TaskKey: tasks[i].TaskKey, tagged: allTagged
                };
                taskAdded = false;

                taskKey = tasks[i].TaskKey;
                if (!taskKeysAdded[taskKey]) {
                    if (taskFilter) {
                        //if a filter is passed in, only include the task if it has the user or service from the filter assigned to it
                        taskUsers = dsTaskUsers.findByKey(tasks[i].TaskKey, 'TaskKey');
                        if (taskUsers) {
                            if (!Array.isArray(taskUsers)) {
                                taskUsers = [taskUsers];
                            }
                        } else {
                            taskUsers = [];

                            if(taskFilter.allUnassigned && tasks[i].TaskType === 2) {
                                returnTasks.unshift(taskItem);
                                taskAdded = true;
                            }
                        }

                        foundFilter = false;
                        for (var j = 0; j < taskUsers.length; j++) {
                            if (foundFilter) break;
                            switch (taskFilter.entity.toLowerCase()) {
                                case 'userkey':
                                    if (taskUsers[j].UserKey === taskFilter.key && u.obn(taskUsers[j].PercComp) < 100) {
                                        returnTasks.unshift(taskItem);
                                        foundFilter = true;
                                        taskAdded = true;
                                    }
                                    break;
                                case 'servicekey':
                                    if (taskUsers[j].ServiceKey === taskFilter.key && u.obn(taskUsers[j].PercComp) < 100) {
                                        returnTasks.unshift(taskItem);
                                        foundFilter = true;
                                        taskAdded = true;
                                    }
                                    break;
                            }
                        }
                    } else {
                        returnTasks.unshift(taskItem);
                        taskAdded = true;
                    }

                    if (!tasksOnly) {
                        taskUsers = dsTaskUsers.findByKey(tasks[i].TaskKey, 'TaskKey');
                        if (taskUsers && !Array.isArray(taskUsers)) taskUsers = [taskUsers];

                        if (taskUsers) {
                            var includeTaskUser;
                            for (var j = 0; j < taskUsers.length; j++) {
                                //initialize the flag (if we're not using a filter, or not assigning using the "All Tasks" option, then all taskUsers should be included
                                includeTaskUser = (!taskFilter || !taskFilter.assignAllTasks);

                                if(!includeTaskUser){
                                    //if it wasn't yet set to true above, then see if this task user should be included based on the filter
                                    if(taskFilter.entity){
                                        switch (taskFilter.entity.toLowerCase()) {
                                            case 'userkey':
                                                if (taskUsers[j].UserKey === taskFilter.key) {
                                                    includeTaskUser = true;
                                                }
                                                break;
                                            case 'servicekey':
                                                if (taskUsers[j].ServiceKey === taskFilter.key) {
                                                    includeTaskUser = true;
                                                }
                                                break;
                                        }
                                    }
                                }

                                if(includeTaskUser) {
                                    if(u.obn(taskUsers[j].PercComp) < 100){
                                        //put "SummaryTaskKey" onto the taskUser records so that it fits into the HDD. These will be ignored upon save
                                        taskUsers[j].SummaryTaskKey = taskItem.TempKey;
                                        if (!taskUsers[j].TempKey) taskUsers[j].TempKey = $.guid++;
                                        taskUsers[j].DisplayOrder = j + 1;
                                        taskUsers[j].TaskLevel = taskItem.TaskLevel + 1;
                                        taskUsers[j].ProjectKey = project.ProjectKey;
                                        taskUsers[j].tagged = allTagged || isTagged(taggedTaskUserKeys, taskUsers[j].TaskUserKey);
                                        returnTasks.splice(1, 0, taskUsers[j]);
                                    }
                                }
                            }
                        }
                    }
                    if (taskAdded) taskKeysAdded[taskKey] = taskItem.TempKey;
                }

                if (taskAdded) {
                    //if the task was added, get all of its parents
                    childTask = taskItem;
                    taskKey = u.obn(tasks[i].SummaryTaskKey);
                    while (taskKey > 0) {
                        parent = findTaskByKey(taskKey);
                        if (!parent) break;

                        if (taskKeysAdded[parent.TaskKey]) {
                            //if the parent is already added by a task below, we'll need to move it up to where we are at now
                            for (var k = 0; k < returnTasks.length; k++) {
                                if (returnTasks[k].TaskKey === parent.TaskKey) {
                                    returnTasks.splice(0, 0, returnTasks.splice(k, 1)[0])
                                }
                            }
                        } else {
                            parentTaskItem = {
                                TempKey: $.guid++, TaskName: parent.TaskName, SummaryTaskKey: taskKeysAdded[parent.SummaryTaskKey] ? taskKeysAdded[parent.SummaryTaskKey] : null, DisplayOrder: parent.DisplayOrder, TaskLevel: parent.TaskLevel,
                                TaskKey: parent.TaskKey, tagged: allTagged
                            };
                            if (childTask) childTask.SummaryTaskKey = parentTaskItem.TempKey;
                            returnTasks.unshift(parentTaskItem);
                            taskKeysAdded[parent.TaskKey] = parentTaskItem.TempKey;
                            childTask = parentTaskItem;
                        }
                        taskKey = u.obn(parent.SummaryTaskKey);
                    }
                    childTask = null;
                }
            }

            for (var i = 0; i < returnTasks.length; i++) {
                returnTasks[i].SortOrder = i + 1;
            }

            return returnTasks;
        }

        /**
         * Returns true if the TaskUserKey is in the array of taggedTaskUserKeys
         * @param taggedTaskUserKeys
         * @param taskUserKey
         * @returns {*}
         */
        function isTagged(taggedTaskUserKeys, taskUserKey){
            if(!taggedTaskUserKeys) return false;
            return taggedTaskUserKeys.some(function(item){return u.obs(item) === u.obs(taskUserKey) });
        }

        /**
         * Called to clear out all of the internal arrays of modified items
         * @param taskKeyMap
         */
        function commitModifiedItems(taskKeyMap) {
            deletedPreds = [];
            deletedTaskUsers = [];
            removeProjectUsers = [];
            eventLog = [];
            commitPredKeys(taskKeyMap);
            formatPredValues();
            setOriginalValues();
        }

        /**
         * Returns the internal arrays of modified items for use when saving the schedule
         * @returns {{deletedPredecessors: Array, deletedTaskUsers: Array, addedAssignments: Array.<*>, deletedAssignments: Array}}
         */
        function getModifiedItems() {
            var assignmentChanges = getAssignmentChanges();
            return {deletedPredecessors: deletedPreds, deletedTaskUsers: deletedTaskUsers, addedAssignments: assignmentChanges.addedUsers, deletedAssignments: assignmentChanges.deletedUsers, removeProjectUsers: removeProjectUsers, eventLog: eventLog};
        }

        /**
         * Marks the task dirty
         * @param task
         */
        function markTaskDirty(task) {
            if (!task) return;
            if (task.rowstate && task.rowstate.indexOf('D') >= 0) return; //if it's already marked for delete, no need to add "U". Plus, some grid and datasource code filter specificially for "D", not "DU".
            dsTasks.itemUpdated('U', task, false);
        }

        function logEvent(action, msg) {
            eventLog.push({action: action, msg: msg});
        }

        this.hasGeneralErrors = hasGeneralErrors;
        this.getGeneralErrors = getGeneralErrors;
        this.hasGeneralNotifications = hasGeneralNotifications;
        this.getGeneralNotifications = getGeneralNotifications;
        this.invalidateSchedule = invalidateSchedule;
        this.scheduleIsInvalid = scheduleIsInvalid;
        this.validateSchedule = validateSchedule;
        this.validateStructure = validateStructure;
        this.findTaskByKey = findTaskByKey;
        this.getParentTask = getParentTask;
        this.loadPredecessors = loadPredecessors;
        this.isPredecessor = isPredecessor;
        this.hasPredecessors = hasPredecessors;
        this.initBlockoutDays = initBlockoutDays;
        this.isWorkDay = isWorkDay;
        this.isProjectWorkDay = projectWorkDay;
        this.findNextWorkDate = findNextWorkDate;
        this.calcNextDate = calcNextDate;
        this.calcDuration = calcDuration;
        this.setTrackBudgetFlags = setTrackBudgetFlags;
        this.checkTrackBudgetState = checkTrackBudgetState;
        this.pushDownActuals = pushDownActuals;
        this.addTask = addTask;
        this.lagUpdated = lagUpdated;
        this.linkToTask = linkToTask;
        this.addCrossProjectPredInformation = addCrossProjectPredInformation;
        this.unlinkFromTask = unlinkFromTask;
        this.deleteTasks = deleteTasks;
        this.isMoveTaskAllowed = isMoveTaskAllowed;
        this.isDropAsChildAllowed = isDropAsChildAllowed;
        this.removeTaskAssignedUsers = removeTaskAssignedUsers;
        this.afterMoveTask = afterMoveTask;
        this.copyTask = copyTask;
        this.autoNumberTasks = autoNumberTasks;
        this.setTaskNumbers = setTaskNumbers;
        this.initBuckets = initBuckets;
        this.calcColumns = calcColumns;
        this.calcBucketHours = calcBucketHours;
        this.calcBucketTotal = calcBucketTotal;
        this.validateTaskBuckets = validateTaskBuckets;
        this.validateAllBuckets = validateAllBuckets;
        this.refreshBlockoutDays = refreshBlockoutDays;
        this.linkTasks = linkTasks;
        this.unlinkTasks = unlinkTasks;
        this.commitModifiedItems = commitModifiedItems;
        this.getModifiedItems = getModifiedItems;
        this.logEvent = logEvent;
        this.markTaskDirty = markTaskDirty;
        this.rollupAssignedUsersToTask = rollupAssignedUsersToTask;
        this.taskUserDefaults = taskUserDefaults;
        this.removeTaskUser = removeTaskUser;
        this.reassign = reassign;
        this.reassignCheck = reassignCheck;
        this.setConstraint = setConstraint;
        this.dropTemplateTask = dropTemplateTask;
        this.addTaskPredecessor = addTaskPredecessor;
        this.getPreds = getPreds;
        this.formatPredValues = formatPredValues;
        this.getSelectedTaskAssignmentsWithParents = getSelectedTaskAssignmentsWithParents;
        this.removePredecessorFromSource = removePredecessorFromSource;
    };

    //region ParmsDefinition

    var ScheduleParm = function (dsTasks, dsPreds, dsTaskUsers, project, blockoutDays, prefs, assignments, options) {

        this.dsTasks = dsTasks;
        this.dsPreds = dsPreds;
        this.dsTaskUsers = dsTaskUsers;
        this.project = project || {};
        this.blockoutDays = blockoutDays || {};
        this.prefs = prefs || {};
        this.assignments = assignments || [];
        this.selectedTemplate = {dsTaskUsers: [], dsPreds: []};
        this.options = options || {};

        this.setOriginalValues = function () {

            var item;
            var list = dsTasks.source;
            for (var i = 0; i < dsTasks.source.length; i++) {
                item = dsTasks.source[i];
                item.original = {PlanDuration: item.PlanDuration, PercComp: item.PercComp};
            }

            for (var i = 0; i < dsTaskUsers.source.length; i++) {
                item = dsTaskUsers.source[i];
                item.original = {
                    UserKey: item.UserKey, UserName: item.UserName, ServiceKey: item.ServiceKey, ServiceDescription: item.ServiceDescription,
                    Hours: item.Hours, APlanStart: item.APlanStart, APlanComplete: item.APlanComplete, PercComp: item.PercComp
                };
            }
        }

        this.setOriginalValues(); //execute it the first time an instance is created

        this.scheduleMgr = new ScheduleMgr(this);

        this.constraintList = System.localize([
            {k: Enum.ProjectScheduleConstraintTypes.AS_SOON_AS_POSSIBLE, l: (this.project.ScheduleDirection === Enum.ProjectScheduleDirection.DIRECTION_FWD) ? "|26675|As Soon As Possible|" : "|26676|As Late As Possible|"},
            {k: Enum.ProjectScheduleConstraintTypes.MUST_START_ON, l: "|26691|Must Start On|"},
            {k: Enum.ProjectScheduleConstraintTypes.MUST_FINISH_ON, l: "|26692|Must Finish On|"},
            {k: Enum.ProjectScheduleConstraintTypes.FINISH_NO_EARLIER_THAN, l: "|26693|Finish No Earlier Than|"},
            {k: Enum.ProjectScheduleConstraintTypes.FINISH_NO_LATER_THAN, l: "|26694|Finish No Later Than|"},
            {k: Enum.ProjectScheduleConstraintTypes.START_NO_EARLIER_THAN, l: "|26695|Start No Earlier Than|"},
            {k: Enum.ProjectScheduleConstraintTypes.START_NO_LATER_THAN, l: "|26696|Start No Later Than|"}]);

        this.constraintListWithDuration = this.constraintList.concat(System.localize([{k: Enum.ProjectScheduleConstraintTypes.DURATION_OF_PROJECT, l: "|26697|Duration Of Project|"}]));

        //return true if dsTasks has not been set, meaning this is just for one task
        this.isTaskOnly = function () {
            return !this.dsTasks;
        };

        this.findTaskByKey = function (keyValue, keyField) {
            if (this.isTaskOnly()) return null;
            return this.scheduleMgr.findTaskByKey(keyValue, keyField);
        };

        this.refreshBlockoutDays = function (blockoutDays) {
            this.blockoutDays = blockoutDays;
            this.scheduleMgr.refreshBlockoutDays();
        };

        this.setSelectedTemplate = function (dsTaskUsers, dsPreds) {
            this.selectedTemplate.dsTaskUsers = dsTaskUsers;
            this.selectedTemplate.dsPreds = dsPreds;
        }
    };
    window.ScheduleParm = ScheduleParm;
    //endregion

    var ScheduleResponse = function (item, field) {

        this.item = item; //the row being updated
        this.field = field; //if set, this is the field that was just changed
        this.errors = [];
        this.generalErrors = [];
        this.notifications = [];

        this.addError = function (msg, entity, entityKey) {
            this.errors.push(new ScheduleValidationError(msg, entity || this.field, entityKey));
        };

        this.hasErrors = function () {
            return this.errors.length > 0 || this.generalErrors.length > 0;
        };

        this.hasEntityErrors = function (entity) {
            entity = entity || this.field;

            return this.errors.some(function (item) {
                return item.entity === entity;
            });
        };

        this.addDSErrors = function (ds) {

            //clear all row errors, then add the ones from the errors array
            ds.clearError(null, item);

            for (var i = 0; i < this.errors.length; i++) {
                ds.addError(System.localize(this.errors[i].msg), this.field, item);
            }

            if (this.generalErrors.length > 0) {
                Notify.error(this.generalErrors.join('\n'));
            }
        };

        this.addNotification = function (msg) {
            this.notifications.push({msg: msg});
        };

        this.hasNotifications = function () {
            return this.notifications.length > 0;
        };

        this.getNotifications = function () {

            var rtn = "";
            System.localize(this.notifications, 'msg');
            this.notifications.forEach(function (item) {
                if (rtn.length > 0) rtn += '\n';
                rtn += item.msg;
            });
            return rtn;
        };
    };
    window.ScheduleResponse = ScheduleResponse;

    var ScheduleValidationError = function (msg, entity, entityKey) {
        this.msg = msg;
        this.entity = entity;
        this.entityKey = entityKey;
    };
    window.ScheduleValidationError = ScheduleValidationError;


})(jQuery);